using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using Microsoft.Build.Framework;
using System.Threading;
using Microsoft.Build.BuildEngine;
using System.IO;

namespace Editor.Editor
{
    public partial class BuildContentDialog : Form, ILogger
    {
        XNAProjectWrapper project;
        System.Windows.Forms.Timer timer;
        bool closed = false;

        public BuildContentDialog(XNAProjectWrapper project)
        {
            this.project = project;
            InitializeComponent();
            timer = new System.Windows.Forms.Timer();
            timer.Interval = 500;
            timer.Tick += new EventHandler(timer_Tick);
            timer.Start();
            FormClosing += new FormClosingEventHandler(BuildContentDialog_FormClosing);
        }

        void BuildContentDialog_FormClosing(object sender, FormClosingEventArgs e)
        {
            timer.Stop();
            timer.Dispose();
            closed = true;
            project.Engine.UnregisterAllLoggers();
        }

        void timer_Tick(object sender, EventArgs e)
        {
            UpdateTextBox();
        }


        private void BuildContentDialog_Shown(object sender, EventArgs e)
        {
            Thread t = new Thread(new ThreadStart(Build));
            t.SetApartmentState(ApartmentState.STA);
            project.Engine.RegisterLogger(this);

            t.Start();
        }

        private void Build()
        {
            project.Project.Build("BuildNestedProjects");
        }

        StringBuilder sb = new StringBuilder();

        List<BuildWarningEventArgs> warnings = new List<BuildWarningEventArgs>();
        List<BuildErrorEventArgs> errors = new List<BuildErrorEventArgs>();
        private void UpdateTextBox()
        {
            
            textBox1.Text = sb.ToString();
            if (textBox1.Text.Length > 0)
            {
                int index = textBox1.Text.Trim().LastIndexOf(Environment.NewLine);
                if (index == -1)
                    return;
                index += Environment.NewLine.Length;
                if (index < textBox1.Text.Length)
                {
                    textBox1.SelectionStart = index;
                    textBox1.ScrollToCaret();
                }
            }
        }

        #region ILogger Members

        public void Initialize(IEventSource eventSource)
        {
            eventSource.ErrorRaised += new BuildErrorEventHandler(eventSource_ErrorRaised);
            eventSource.WarningRaised += new BuildWarningEventHandler(eventSource_WarningRaised);
            eventSource.MessageRaised += new BuildMessageEventHandler(eventSource_MessageRaised);
            eventSource.BuildFinished += new BuildFinishedEventHandler(eventSource_BuildFinished);
            eventSource.BuildStarted += new BuildStartedEventHandler(eventSource_BuildStarted);
        }

        void eventSource_BuildStarted(object sender, BuildStartedEventArgs e)
        {
            sb.AppendLine(e.Message);
        }


        void eventSource_BuildFinished(object sender, BuildFinishedEventArgs e)
        {
            sb.AppendLine(e.Message);
            if (errors.Count == 0 && warnings.Count == 0 && !closed)
            {
                Invoke(new MethodInvoker(Close));
            }
            else
            {

                Invoke(new MethodInvoker(delegate 
                    {
                        timer.Stop();
                        UpdateTextBox();
                        progressBar1.Style = ProgressBarStyle.Blocks;
                        button1.Enabled = true;
                        foreach (var error in errors)
                        {
                            DataGridViewRow r = dataGridView1.Rows[dataGridView1.Rows.Add(new object[] { "Error", error.File, error.Message })]; 
                            r.DefaultCellStyle.BackColor = Color.Red;
                           
                        }
                        foreach (var warning in warnings)
                        {
                            dataGridView1.Rows[dataGridView1.Rows.Add(new object[] { "Warning", warning.File,  warning.Message })].DefaultCellStyle.BackColor = Color.Yellow;
                        }
                        dataGridView1.AutoResizeRows();  
                    }));
            }
        }

        void eventSource_MessageRaised(object sender, BuildMessageEventArgs e)
        {

            //sb.AppendLine(e.Message);
            if(e.Importance == MessageImportance.High )
                sb.AppendLine(string.Format("    {0}", e.Message));
        }

        void eventSource_WarningRaised(object sender, BuildWarningEventArgs e)
        {
            sb.AppendLine(string.Format("  {0}", e.Message));
            warnings.Add(e);
            
        }

        void eventSource_ErrorRaised(object sender, BuildErrorEventArgs e)
        {
            sb.AppendLine(string.Format("{0}", e.Message));
            errors.Add(e);
        }


        string parameters = "";
        public string Parameters
        {
            get
            {
                return parameters;
            }
            set
            {
                parameters = value;
            }
        }

        public void Shutdown()
        {
        }
        LoggerVerbosity verbosity = LoggerVerbosity.Minimal;
        public LoggerVerbosity Verbosity
        {
            get
            {
                return verbosity;
            }
            set
            {
                verbosity = value;
            }
        }

        #endregion

        private void button1_Click(object sender, EventArgs e)
        {
            this.Close();
        }
    }
}