using System;
using System.Collections.Generic;
using System.Text;
using GameEngine.Level;
using Editor.Data;
using GameEngine.Xml;
using GameEngine.Helpers;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using System.Linq;
namespace Editor
{
    public enum SelectionMode { Select2D, Select3D, SelectPoints };

    public class SelectionChangedEventArgs : EventArgs
    {
        private List<SceneNode> selectedNodes;
        public List<SceneNode> SelectedNodes { get { return this.selectedNodes; } }

        public SelectionChangedEventArgs(List<SceneNode> selectedNodes)
        {
            this.selectedNodes = selectedNodes;
        }
    }
    public delegate void SelectionChangedHandler(object sender, SelectionChangedEventArgs e);

    public class SelectionHandler
    {
        public event SelectionChangedHandler SelectionChanged;

        public void OnSelectionChanged(SelectionChangedEventArgs e)
        {
            if (SelectionChanged != null)
                SelectionChanged(this, e);
        }

        private List<ObjectNode> selected3D = new List<ObjectNode>();
        private List<PhysicsPolygon> selected2D = new List<PhysicsPolygon>();
        private List<SelectionPoint> selectedPoints = new List<SelectionPoint>();

        private List<SceneNode> oldSelectedNodes = new List<SceneNode>();
        public List<SceneNode> OldSelectedNodes { get { return oldSelectedNodes; } }

        private List<SelectionPoint> oldSelectedPoints = new List<SelectionPoint>();
        public List<SelectionPoint> OldSelectedPoints { get { return oldSelectedPoints; } }

        private bool dirtied = false;
        private bool Dirtied
        {
            get { return this.dirtied; }
            set { this.dirtied = value; }
        }

        private Vector3 selectionCenter = Vector3.Zero;

        public Vector3 SelectionCenter
        {
            get 
            {
                if (Dirtied)
                    Recalculate();
                return selectionCenter;
            }

            set { selectionCenter = value; }
        }

        private Vector3 selectionExtent = Vector3.Zero;

        public Vector3 SelectionExtent
        {
            get
            {
                if (Dirtied)
                    Recalculate();
                return selectionExtent;
            }
            set { selectionExtent = value; }
        }

        private Quaternion selectionRotation;

        public Quaternion SelectionRotation { get{ return selectionRotation; } set { selectionRotation = value; } }

        private OrientedBoundingBox selectionOBB = new OrientedBoundingBox();

        private SelectionMode selectionMode = SelectionMode.Select3D;

        public SelectionMode SelectionMode
        {
            get { return selectionMode; }
            set { selectionMode = value; }
        }

        public int Count
        {
            get
            {
                switch (selectionMode)
                {
                    case SelectionMode.Select2D:
                        return selected2D.Count;
                    case SelectionMode.Select3D:
                        return selected3D.Count;
                    case SelectionMode.SelectPoints:
                        return selectedPoints.Count;
                    default:
                        break;
                }
                return 0;
            }
        }

        // Must be called before starting a selection
        public void BeginSelection()
        {
            ClearSelection();
        }

        public bool HasAncestorInSelection(ObjectNode node)
        {
            ObjectNode p = node.Parent as ObjectNode;
            if (p != null)
            {
                if (selected3D.Contains(p))
                {
                    return true;
                }
                else
                {
                    return HasAncestorInSelection(p);
                }
            }
            else
            {
                return false;
            }
        }

        public void EndSelection()
        {
            // Cleanup selection, removing children whose ancestors are already in the selection:
            switch (selectionMode)
            {
                case SelectionMode.Select2D:
                    throw new Exception("PhysicsPolygon not handled in selection");
                    // break;
                case SelectionMode.Select3D:
                    for (int i = selected3D.Count - 1; i >= 0; i--)
                    {
                        if(HasAncestorInSelection(selected3D[i]))
                        {
                            RemoveFromSelection(selected3D[i]);
                        }
                    }
                    break;
                default:
                    break;
            }


            switch (selectionMode)
            {
                case SelectionMode.SelectPoints:
                    List<SelectionPoint> selectedPoints = new List<SelectionPoint>();
                    foreach (SelectionPoint point in this.selectedPoints)
                        selectedPoints.Add(point);
                    if (oldSelectedPoints != selectedPoints)
                    {
                        oldSelectedPoints = selectedPoints;
                        // XXX Need event firing
                    }
                    break;
                case SelectionMode.Select2D:
                    /*foreach (PhysicsPolygon node in this.selected2D)
                        selectedNodes.Add((SceneNode)node);*/
                    throw new Exception("PhysicsPolygon cannot be cast to SceneNode...");
                case SelectionMode.Select3D:
                default:
                    List<SceneNode> selectedNodes = new List<SceneNode>();
                    foreach (ObjectNode node in this.selected3D)
                        selectedNodes.Add((SceneNode)node);
                    if (oldSelectedNodes != selectedNodes)
                    {
                        OnSelectionChanged(new SelectionChangedEventArgs(selectedNodes));
                        oldSelectedNodes = selectedNodes;
                    }
                    break;
            }
        }

        public void AddToSelection(ObjectNode gameObj)
        {
            selected3D.Add(gameObj);
            Dirtied = true;
        }

        public void AddToSelection(PhysicsPolygon polyObj)
        {
            selected2D.Add(polyObj);
            Dirtied = true;
        }

        public void AddToSelection(PolygonNode polyObj, int index)
        {
            selectedPoints.Add( new SelectionPoint(polyObj, index));
        }

        public bool Contains(ObjectNode gameObj)
        {
            return selected3D.Contains(gameObj);
        }

        public bool Contains(PhysicsPolygon polyObj)
        {
            return selected2D.Contains(polyObj);
        }

        public bool Contains(PolygonNode polyObj, int index)
        {
            SelectionPoint tmpPoint = new SelectionPoint(polyObj, index);
            for (int i = 0; i < selectedPoints.Count; i++)
            {
                if (tmpPoint.Equals(selectedPoints[i]))
                    return true;
            }
            return false;
        }

        public void RemoveFromSelection(ObjectNode gameObj)
        {
            selected3D.Remove(gameObj);
            Dirtied = true;
        }

        public void RemoveFromSelection(PhysicsPolygon polyObj)
        {
            selected2D.Remove(polyObj);
            Dirtied = true;
        }

        public void RemoveFromSelection(PolygonNode polyObj, int index)
        {
            SelectionPoint tmpPoint = new SelectionPoint(polyObj, index);
            for (int i = 0; i < selectedPoints.Count; i++)
            {
                if(tmpPoint.Equals(selectedPoints[i]))
                {
                    selectedPoints.RemoveAt(i);
                    return;
                }
            }
        }

        public void ClearSelection()
        {
            selected2D.Clear();
            selected3D.Clear();
            selectedPoints.Clear();
            Dirtied = true;
        }

        public ObjectNode GetFirstSelected3D()
        {
            if (selected3D.Count > 0)
            {
                return selected3D[0];
            }
            else
            {
                return null;
            }
        }

        public PhysicsPolygon GetFirstSelected2D()
        {
            if (selected2D.Count > 0)
            {
                return selected2D[0];
            }
            else
            {
                return null;
            }
        }

        public ObjectNode[] Get3DSelection()
        {
            return selected3D.ToArray();
        }

        public PhysicsPolygon[] Get2DSelection()
        {
            return selected2D.ToArray();
        }

        public SelectionPoint[] GetPointSelection()
        {
            return selectedPoints.ToArray();
        }

        // Test whether a ray intersects the current selection
        public bool Intersects(Ray ray)
        {
            switch (selectionMode)
            {
                case SelectionMode.Select2D:
                    break;
                case SelectionMode.Select3D:
                    if (Count > 0)
                    {
                        if (Count == 1)
                        {
                            ObjectNode hitNode = selected3D[0] as ObjectNode;
                            if (hitNode != null)
                                return hitNode.OrientedBoundingBox.Intersects(ray);
                            else
                                return false;
                        }
                        selectionOBB.Center = selectionCenter;
                        selectionOBB.Extent = selectionExtent;
                        selectionOBB.Rotation = Matrix.CreateFromQuaternion(selectionRotation);
                        return selectionOBB.Intersects(ray);
                    }
                    break;
                default:
                    break;
            }

            return false;
        }


        public void Recalculate()
        {

            switch (selectionMode)
            {
                case SelectionMode.Select2D:
                    break;
                case SelectionMode.Select3D:
                    Vector3 minPoint = Vector3.Zero;
                    Vector3 maxPoint = Vector3.Zero;

                    ObjectNode firstModel = selected3D.FirstOrDefault(x => x is ObjectNode) as ObjectNode;

                    if (selected3D.Count > 1 && firstModel != null)
                    {
                        Vector3 maxProj = new Vector3();

                        minPoint = firstModel.World.Translation;
                        maxPoint = minPoint;

                        for (int i = 0; i < selected3D.Count; i++)
                        {
                            ObjectNode modelNode = selected3D[i];

                            if (modelNode is ModelNode)
                            {
                                maxProj = modelNode.OrientedBoundingBox.MaxAAExtent();

                                maxPoint.X = Math.Max(modelNode.OrientedBoundingBox.Center.X + maxProj.X, maxPoint.X);
                                maxPoint.Y = Math.Max(modelNode.OrientedBoundingBox.Center.Y + maxProj.Y, maxPoint.Y);
                                maxPoint.Z = Math.Max(modelNode.OrientedBoundingBox.Center.Z + maxProj.Z, maxPoint.Z);

                                minPoint.X = Math.Min(modelNode.OrientedBoundingBox.Center.X - maxProj.X, minPoint.X);
                                minPoint.Y = Math.Min(modelNode.OrientedBoundingBox.Center.Y - maxProj.Y, minPoint.Y);
                                minPoint.Z = Math.Min(modelNode.OrientedBoundingBox.Center.Z - maxProj.Z, minPoint.Z);
                            }
                            else
                            {
                                maxProj = modelNode.MaxAAExtent();

                                maxPoint.X = Math.Max(modelNode.World.Translation.X + maxProj.X, maxPoint.X);
                                maxPoint.Y = Math.Max(modelNode.World.Translation.Y + maxProj.Y, maxPoint.Y);
                                maxPoint.Z = Math.Max(modelNode.World.Translation.Z + maxProj.Z, maxPoint.Z);

                                minPoint.X = Math.Min(modelNode.World.Translation.X - maxProj.X, minPoint.X);
                                minPoint.Y = Math.Min(modelNode.World.Translation.Y - maxProj.Y, minPoint.Y);
                                minPoint.Z = Math.Min(modelNode.World.Translation.Z - maxProj.Z, minPoint.Z);
                            }
                        }

                        selectionExtent = 0.5f * (maxPoint - minPoint);
                        selectionCenter = minPoint + selectionExtent;
                        selectionRotation = Quaternion.Identity;
                    }
                    else
                    {
                        if (firstModel != null)
                        {
                            if (firstModel is ModelNode)
                            {
                                selectionExtent = firstModel.OrientedBoundingBox.Extent;
                                selectionCenter = firstModel.WorldPosition;
                                selectionRotation = Quaternion.CreateFromRotationMatrix(firstModel.OrientedBoundingBox.Rotation);
                            }
                            else
                            {
                                selectionExtent = firstModel.MaxAAExtent();
                                selectionCenter = firstModel.WorldPosition;
                                selectionRotation = Quaternion.Identity;
                            }

                        }
                        else
                        {
                            selectionExtent = Vector3.One;
                            selectionCenter = Vector3.Zero;
                            selectionRotation = Quaternion.Identity;
                        }
                    }
                    //selectionExtent = Vector3.Subtract(maxPoint, minPoint) * 0.5f;
                    //selectionCenter = minPoint + selectionExtent;
                    break;
                case SelectionMode.SelectPoints:
                    Vector3 newCenter = Vector3.Zero;
                    if(selectedPoints.Count > 0) 
                    {
                        for (int i = 0; i < selectedPoints.Count; i++)
                        {
                            newCenter += selectedPoints[i].WorldPosition;
                        }
                        newCenter /= selectedPoints.Count;
                    }
                    SelectionCenter = newCenter;
                    SelectionRotation = Quaternion.Identity;
                    SelectionExtent = Vector3.One;
                    break;
                default:
                    break;
            }

            Dirtied = false;
        }
    }

    public class SelectionPoint
    {
        private PolygonNode container;
        public PolygonNode Container { get { return container; } set { container = value; } }
        private int index;
        public int Index { get { return index; } set { index = value; } }

        public Vector2 Value
        {
            get { return (container != null && index >= 0 && index < container.Points.Count) ? container.Points[index] : Vector2.Zero; }
            set { container.Points[index] = value; }
        }

        public Vector3 Value3D
        {
            get { return new Vector3(Value.X, 0, Value.Y); }
            set { Value = new Vector2(value.X, value.Z); }
        }

        public Vector3 WorldPosition
        {
            get { return container == null ? Vector3.Zero :  Value3D + container.WorldPosition; }
            set { Value3D = value - container.WorldPosition; }
        }

        public SelectionPoint(PolygonNode container, int index)
        {
            this.container = container;
            this.index = index;
        }

        public bool Equals(SelectionPoint other)
        {
            if (other == null)
                return false;

            return this.container == other.container && this.index == other.index;
        }
    }
}
