using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
using GameEngine.Level;
using Editor.Editor;
using GameEngine.ScreenSystem;
using GameEngine.Helpers;
using Microsoft.Xna.Framework.Input;
using GameEngine;
using GameEngine.Xml;
using GameEngine.Cameras;

namespace Editor.Tools
{
    public class MoveToolWorkingPlaneChangedEventArgs : EventArgs
    {
        private OrientationPlane oldGrid, newGrid;
        public OrientationPlane OldGrid { get { return this.oldGrid; } }
        public OrientationPlane NewGrid { get { return this.newGrid; } }

        public MoveToolWorkingPlaneChangedEventArgs(OrientationPlane oldGrid, OrientationPlane newGrid)
        {
            this.oldGrid = oldGrid;
            this.newGrid = newGrid;
        }
    }

    public delegate void MoveToolWorkingPlaneChangedHandler(object sender, MoveToolWorkingPlaneChangedEventArgs e);

    public enum MoveToolState { Initial, Moving };

    class MoveTool : Tool
    {
        private SpriteFont screenFont;

        private MoveToolState state = MoveToolState.Initial;
        private OrientationPlane workingPlane = OrientationPlane.XZ;
        public OrientationPlane WorkingPlane { get { return workingPlane; } }
        public event MoveToolWorkingPlaneChangedHandler WorkingPlaneChanged;

        private Vector3? newMousePosition;
        private List<Vector3> initialPositions = new List<Vector3>();
        private List<Vector3> lastSavedPositions = new List<Vector3>();
        private Vector3 initialSelectionPosition;
        private Vector3 lastSavedSelectionPosition;
        private Vector3? lastSavedMousePosition;
        private Plane movePlane;

        private bool gridInitialized = false;

        private Grid moveGrid;
        public bool HelperGridVisible { get { return moveGrid.Visible; } }
        public event HelperGridVisibilityChangedHandler HelperGridVisibilityChanged;
        private Effect primitivesEffect;

        private bool oneAxisMode = false;

        ObjectNode[] selection;

        public MoveTool(OrthoOrbitCamera camera, ToolHandler handler)
            : base(camera,handler)
        {
            this.Name = "MoveTool";
            this.moveGrid = new Grid(16, 16, 10f, Color.White);
            this.moveGrid.Visible = false;
        }

        public override void LoadContent(Microsoft.Xna.Framework.Content.ContentManager content)
        {
            screenFont = content.Load<SpriteFont>("Fonts/Arial");
            primitivesEffect = content.Load<Effect>("primitives");
        }

        public override void HandleInput(GameTime gameTime)
        {

            if (InputState.KeyReleased(Keys.Enter))
            {
                moveGrid.Visible = !moveGrid.Visible;
                UpdateMoveGridRotation();
            }

            if (InputState.IsKeyDown(Keys.LeftControl) || InputState.IsKeyDown(Keys.RightControl))
            {
                oneAxisMode = true;
            }
            else
            {
                oneAxisMode = false;
            }


            EnableControls(enabledControls);

            if (toolHandler.Selection.Count > 0)
            {
                switch (state)
                {
                    case MoveToolState.Initial:
                        CheckPlaneChange();

                        if (InputState.MousePressed(MouseButton.Left))
                        {
                            // Two things can happen - either we clicked on an existing collection and have to work with that
                            // or we click elsewhere and have to let the selectionTool take over
                            if (toolHandler.Selection.Intersects(camera.GetPickingRay(Cursor.Position)))
                            {
                                selection = toolHandler.Selection.Get3DSelection();
                                UpdateMovePlane(toolHandler.Selection.SelectionCenter);
                                lastSavedMousePosition = GetMovePlaneIntersection();

                                if (lastSavedMousePosition != null)
                                {
                                    initialPositions.Clear();
                                    lastSavedPositions.Clear();

                                    for (int i = 0; i < selection.Length; i++)
                                    {
                                        initialPositions.Add(selection[i].WorldPosition);
                                        lastSavedPositions.Add(selection[i].WorldPosition);
                                    }

                                    initialSelectionPosition = lastSavedSelectionPosition = toolHandler.Selection.SelectionCenter;

                                    state = MoveToolState.Moving;
                                }
                            }
                            else
                            {
                                toolHandler.SwitchTemporarilyToSelection(gameTime);
                            }
                        }
                        break;
                    case MoveToolState.Moving:
                        newMousePosition = GetMovePlaneIntersection();
                        if (newMousePosition != null)
                        {
                            Vector3 delta = newMousePosition.Value - lastSavedMousePosition.Value;

                            // crtl snaps to one axis
                            if (oneAxisMode)
                            {
                                switch (workingPlane)
                                {
                                    case OrientationPlane.XY:
                                        if (Math.Abs(delta.X) > Math.Abs(delta.Y))
                                            delta.Y = 0;
                                        else
                                            delta.X = 0;
                                        break;
                                    case OrientationPlane.XZ:
                                        if (Math.Abs(delta.X) > Math.Abs(delta.Z))
                                            delta.Z = 0;
                                        else
                                            delta.X = 0;
                                        break;
                                    case OrientationPlane.YZ:
                                        if (Math.Abs(delta.Y) > Math.Abs(delta.Z))
                                            delta.Z = 0;
                                        else
                                            delta.Y = 0;
                                        break;
                                    default:
                                        break;
                                }
                                oneAxisMode = true;
                            }
                            else
                            {
                                oneAxisMode = false;
                            }

                            for (int i = 0; i < selection.Length; i++)
                            {
                                Vector3 newPosition = lastSavedPositions[i] + delta;

                                // Shift removes snapping
                                if (!(InputState.IsKeyDown(Keys.LeftShift) || InputState.IsKeyDown(Keys.RightShift)))
                                {
                                    newPosition = SnapPosition(newPosition);
                                }

                                selection[i].WorldPosition = newPosition;
                            }

                            toolHandler.Selection.SelectionCenter = lastSavedSelectionPosition + selection[0].WorldPosition - lastSavedPositions[0];
                        }

                        if (CheckPlaneChange())
                        {
                            if (toolHandler.Selection.Count > 0)
                            {
                                for (int i = 0; i < selection.Length; i++)
                                {
                                    lastSavedPositions[i] = selection[i].WorldPosition;
                                }
                                lastSavedSelectionPosition = toolHandler.Selection.SelectionCenter;

                                UpdateMovePlane(toolHandler.Selection.SelectionCenter);
                            }
                            if (newMousePosition != null)
                                lastSavedMousePosition = newMousePosition;
                        }

                        if (InputState.MouseReleased(MouseButton.Left))
                        {
                            state = MoveToolState.Initial;
                            UpdateMovePlane(toolHandler.Selection.SelectionCenter);
                        }
                        break;
                    default:
                        break;
                }
            }
            else
            {
                CheckPlaneChange();
                if (InputState.MousePressed(MouseButton.Left))
                {
                    toolHandler.SwitchTemporarilyToSelection(gameTime);
                }
            }
        }

        public override void Update(Microsoft.Xna.Framework.GameTime gameTime)
        {
            moveGrid.Update();
        }

        public override void Draw(DrawInfo drawInfo)
        {
            if (!gridInitialized)
            {
                moveGrid.Initialize(drawInfo.GraphicsDevice, primitivesEffect);
                gridInitialized = true;
            }

            moveGrid.Draw(drawInfo);
        }

        public override void Draw2D(Microsoft.Xna.Framework.Graphics.SpriteBatch spriteBatch, DrawInfo drawInfo)
        {
            string oneAxis = oneAxisMode ? " (One-axis mode)" : "";
            spriteBatch.DrawString(screenFont, "Moving in plane: " + workingPlane.ToString() + oneAxis, new Vector2(30, 30), Color.White);
            string helpgridEnabled = moveGrid.Visible ? "on" : "off";
            spriteBatch.DrawString(screenFont, "Helping grid: " + helpgridEnabled + ", press enter to toggle", new Vector2(30, 50), Color.White);
        }

        public override void ToolStart()
        {
        }

        public override void ToolEnd()
        {
        }

        private Vector3? GetMovePlaneIntersection()
        {
            Ray cursorRay = camera.GetPickingRay(Cursor.Position);
            Nullable<float> intersectDistance = cursorRay.Intersects(movePlane);
            if (intersectDistance != null)
            {
                return cursorRay.Position + cursorRay.Direction * intersectDistance.Value;
            }
            else
            {
                return null;
            }
        }

        public virtual void OnWorkingPlaneChanged(MoveToolWorkingPlaneChangedEventArgs e)
        {
            if (WorkingPlaneChanged != null)
                WorkingPlaneChanged(this, e);
        }

        public bool SetWorkingPlane(OrientationPlane thePlane)
        {
            OrientationPlane initialPlane = workingPlane;

            workingPlane = thePlane;
            // We may well need to send the event anyway - controls may need to update!
            OnWorkingPlaneChanged(new MoveToolWorkingPlaneChangedEventArgs(initialPlane, thePlane));

            if (initialPlane != workingPlane)
            {
                UpdateMoveGridRotation();
                return true;
            }
            else
            {
                return false;
            }
        }

        public virtual void OnHelperGridVisibilityChanged(HelperGridVisibilityChangedEventArgs e)
        {
            if (HelperGridVisibilityChanged != null)
                HelperGridVisibilityChanged(this, e);
        }

        public void ToggleHelperGrid()
        {
            moveGrid.Visible = !moveGrid.Visible;
            OnHelperGridVisibilityChanged(new HelperGridVisibilityChangedEventArgs(moveGrid.Visible));
        }

        public void SetHelperGridVisibility(bool newValue)
        {
            moveGrid.Visible = newValue;
            OnHelperGridVisibilityChanged(new HelperGridVisibilityChangedEventArgs(newValue));
        }

        private bool CheckPlaneChange()
        {
            if (InputState.KeyPressed(Keys.Z))
            {
                return SetWorkingPlane(OrientationPlane.XY);
            }
            else if (InputState.KeyPressed(Keys.X))
            {
                return SetWorkingPlane(OrientationPlane.YZ);
            }
            else if (InputState.KeyPressed(Keys.C) || InputState.KeyPressed(Keys.Y))
            {
                return SetWorkingPlane(OrientationPlane.XZ);
            }
            return false;
        }

        private void UpdateMovePlane(Vector3 position)
        {
            switch (workingPlane)
            {
                case OrientationPlane.XY:
                    movePlane = new Plane(position, position + Vector3.UnitX, position + Vector3.UnitY);
                    break;
                case OrientationPlane.XZ:
                    movePlane = new Plane(position, position + Vector3.UnitX, position + Vector3.UnitZ);
                    break;
                case OrientationPlane.YZ:
                    movePlane = new Plane(position, position + Vector3.UnitY, position + Vector3.UnitZ);
                    break;
                default:
                    break;
            }

            moveGrid.Position = position;
            moveGrid.Position.X = Cursor.XyGrid.Snap(moveGrid.Position.X);
            moveGrid.Position.Y = Cursor.XyGrid.Snap(moveGrid.Position.Y);
            moveGrid.Position.Z = Cursor.XzGrid.Snap(moveGrid.Position.Z);
        }

        private void UpdateMoveGridRotation()
        {
            switch (workingPlane)
            {
                case OrientationPlane.XY:
                    moveGrid.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitX, MathHelper.PiOver2);
                    break;
                case OrientationPlane.XZ:
                    moveGrid.Rotation = Quaternion.Identity;
                    break;
                case OrientationPlane.YZ:
                    moveGrid.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitZ, MathHelper.PiOver2);
                    break;
                default:
                    break;
            }
        }


        private Vector3 SnapPosition(Vector3 input)
        {

            Vector3 position = input;

            switch (workingPlane)
            {
                case OrientationPlane.XY:
                    position.X = Cursor.XyGrid.Snap(position.X);
                    position.Y = Cursor.XyGrid.Snap(position.Y);
                    break;
                case OrientationPlane.XZ:
                    position.X = Cursor.XzGrid.Snap(position.X);
                    position.Z = Cursor.XzGrid.Snap(position.Z);
                    break;
                case OrientationPlane.YZ:
                    position.Y = Cursor.XyGrid.Snap(position.Y);
                    position.Z = Cursor.XzGrid.Snap(position.Z);
                    break;
                default:
                    break;
            }

            return position;
        }

    }
}
