﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.Level;
using GameEngine.ScreenSystem;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework;
using GameEngine.Helpers;
using GameEngine.Xml;
using Microsoft.Xna.Framework.Graphics;
using GameEngine.Cameras;

namespace Editor.Tools
{
    public enum PolyModToolMode { SelectAndMove, AddAndRemove, Moving }


    class PolyModifierTool : Tool
    {
        private SelectionMode lastSelectionMode;

        private bool tempToolChange = false;

        private SpriteFont screenFont;
        private Effect primitivesEffect;

        private SelectionHandler selection;

        private SelectionPoint[] selectedPoints;

        private List<Vector2> initialPositions = new List<Vector2>();
        private Vector2 startingPosition;

        private PolyModToolMode state = PolyModToolMode.SelectAndMove;

        private SelectionPoint closestPoint = new SelectionPoint(null, 0);
        public SelectionPoint ClosestPoint { get { return closestPoint; } }
        private SelectionPoint secondPoint = new SelectionPoint(null, 0);
        public SelectionPoint SecondPoint { get { return secondPoint; } }

        private List<PolygonNode> selectableObjects2D;

        Vector2? currentPosition = null;

        private VertexPositionColor[] vertices = new VertexPositionColor[6];

        private bool snap = true;

        public PolyModifierTool(OrthoOrbitCamera camera, ToolHandler handler)
            : base(camera, handler)
        {
            this.Name = "PolygonModifierTool";
            selection = handler.Selection;

            vertices[0].Color = vertices[1].Color = Color.Yellow;
            vertices[2].Color = vertices[3].Color = Color.Green;
            vertices[4].Color = vertices[5].Color = Color.White;
        }

        public void SetObjectCollections(List<PolygonNode> sel2DCollection)
        {
            this.selectableObjects2D = sel2DCollection;
        }

        public override void LoadContent(Microsoft.Xna.Framework.Content.ContentManager content)
        {
            screenFont = content.Load<SpriteFont>("Fonts/Arial");
            primitivesEffect = content.Load<Effect>("primitives");

        }

        public override void HandleInput(Microsoft.Xna.Framework.GameTime gameTime)
        {
            EnableControls(enabledControls);

            if (InputState.IsKeyDown(Keys.LeftShift) || InputState.IsKeyDown(Keys.RightShift))
            {
                snap = false;
            }
            else
            {
                snap = true;
            }

            // Switch state with enter
            if (InputState.KeyReleased(Keys.Enter))
            {
                switch (state)
                {
                    case PolyModToolMode.SelectAndMove:
                        state = PolyModToolMode.AddAndRemove;
                        break;
                    case PolyModToolMode.AddAndRemove:
                        state = PolyModToolMode.SelectAndMove;
                        closestPoint.Container = secondPoint.Container = null;
                        break;
                    case PolyModToolMode.Moving:
                    default:
                        break;
                }
            }

            switch (state)
	        {
		        case PolyModToolMode.SelectAndMove:
                    if (selection.Count > 0)
                    {
                        if (InputState.KeyReleased(Keys.Escape))
                        {
                            selection.ClearSelection();
                        }
                        if (InputState.MousePressed(MouseButton.Left))
                        {
                            Vector2? cursorPosition = Cursor.GetXZPlanePosition();
                            if(cursorPosition != null) {
                                selectedPoints = selection.GetPointSelection();

                                Vector2 hitPoint = Vector2.Zero;
                                bool hit = false;
                                for (int i = 0; i < selectedPoints.Length; i++)
                                {
                                    if (Math.Abs(selectedPoints[i].WorldPosition.X - cursorPosition.Value.X) <= 5 && Math.Abs(selectedPoints[i].WorldPosition.Z - cursorPosition.Value.Y) <= 5)
                                    {
                                        hit = true;
                                        hitPoint.X = selectedPoints[i].WorldPosition.X;
                                        hitPoint.Y = selectedPoints[i].WorldPosition.Z;
                                        break;
                                    }
                                }
                                if (hit)
                                {
                                    startingPosition = hitPoint;
                                    initialPositions.Clear();
                                    for (int i = 0; i < selectedPoints.Length; i++)
                                    {
                                        initialPositions.Add(selectedPoints[i].Value);
                                    }
                                    state = PolyModToolMode.Moving;
                                }
                                else
                                {
                                    tempToolChange = true;
                                    toolHandler.SwitchTemporarilyToSelection(gameTime);
                                }
                            }
                        }
                    }
                    else
                    {
                        if (InputState.MousePressed(MouseButton.Left))
                        {
                            tempToolChange = true;
                            toolHandler.SwitchTemporarilyToSelection(gameTime);
                        }
                    }
                    break;
                case PolyModToolMode.AddAndRemove:
                    if (InputState.MouseReleased(MouseButton.Left))
                    {
                        if (closestPoint.Container != null)
                        {
                            int insertAt = Math.Min(closestPoint.Index, secondPoint.Index);
                            // Check for insertion at end of array
                            if (insertAt == 0 && Math.Max(closestPoint.Index, secondPoint.Index) == closestPoint.Container.Points.Count - 1)
                            {
                                insertAt = closestPoint.Container.Points.Count;
                            }
                            else
                            {
                                insertAt++;
                            }

                            Vector2 newPoint = currentPosition.Value;
                            newPoint.X -= closestPoint.Container.WorldPosition.X;
                            newPoint.Y -= closestPoint.Container.WorldPosition.Z;

                            closestPoint.Container.AddPoint(insertAt, newPoint);
                        }
                    }
                    if (InputState.MouseReleased(MouseButton.Right))
                    {
                        if (closestPoint.Container != null)
                        {
                            if (closestPoint.Container.Points.Count > 3)
                            {
                                closestPoint.Container.RemovePoint(closestPoint.Index);
                            }
                        }
                    }
                    if (InputState.KeyReleased(Keys.R))
                    {
                        if (closestPoint.Container != null)
                        {
                            closestPoint.Container.ChangeWinding();
                        }
                    }
                    break;
                case PolyModToolMode.Moving:
                    if (InputState.MouseReleased(MouseButton.Left))
                    {
                        state = PolyModToolMode.SelectAndMove;
                    }
                    break;
                default:
                    break;
	        }
        }

        public override void Update(GameTime gameTime)
        {
            currentPosition = Cursor.GetXZPlanePosition();
            if (snap && currentPosition != null)
            {
                Vector2 snapped = new Vector2(Cursor.XzGrid.Snap(currentPosition.Value.X), Cursor.XzGrid.Snap(currentPosition.Value.Y));
                currentPosition = new Nullable<Vector2>(snapped);
            }

            switch (state)
            {
                case PolyModToolMode.SelectAndMove:
                    break;
                case PolyModToolMode.AddAndRemove:
                    // reset stuff
                    closestPoint.Container = null;
                    secondPoint.Container = null;

                    if (currentPosition != null)
                    {
                        selectedPoints = selection.GetPointSelection();
                        float closestDist = float.MaxValue;
                        float closestSecondaryDist = float.MaxValue;


                        foreach (PolygonNode polygon in selectableObjects2D)
                        {
                            Vector2 polygonPosition = new Vector2(polygon.WorldPosition.X, polygon.WorldPosition.Y);

                            int lastIndex = polygon.Points.Count - 1;

                            for (int i = 0; i < polygon.Points.Count; i++)
                            {
                                float dist, secondaryDist;

                                Vector2 startPoint = polygon.Points[lastIndex] + polygonPosition;
                                Vector2 endPoint = polygon.Points[i] + polygonPosition;

                                Vector2 lineSegment = endPoint - startPoint;

                                Vector2 lineNormal = Vector2.Normalize(lineSegment);

                                Vector2 startToMouse = currentPosition.Value - startPoint;
                                Vector2 endToMouse = currentPosition.Value - endPoint;

                                float startDist = Math.Abs(Vector2.Dot(lineNormal, startToMouse));
                                float endDist = Math.Abs(Vector2.Dot(lineNormal, endToMouse));

                                float segmentLength = Math.Abs(Vector2.Dot(lineNormal, lineSegment));

                                float projectedDist = startDist + endDist - segmentLength;
                               
                                if (projectedDist - 0.0001f > 0)
                                {
                                    if (startDist > endDist)
                                    {
                                        dist = Vector2.Subtract(currentPosition.Value, endPoint).LengthSquared();
                                    }
                                    else
                                    {
                                        dist = startToMouse.LengthSquared();
                                    }

                                    secondaryDist = Math.Min(startDist, endDist);
                                }
                                else
                                {
                                    // Distance equals the length of the perpendicular
                                    dist = Vector2.Subtract(startToMouse, startDist * lineNormal).LengthSquared();
                                    secondaryDist = Math.Min(startDist, endDist);
                                }


                                bool closest = false;

                                if (dist - 0.0001f < closestDist)
                                {
                                    // If two distances are the same (happens around "corners", use the secondary distance to choose which to use)
                                    if (Math.Abs(closestDist - dist) < 0.0001)
                                    {
                                        if (secondaryDist < closestSecondaryDist)
                                        {
                                            closest = true;
                                        }
                                    }
                                    else
                                    {
                                        closest = true;
                                    }
                                }

                                if(closest) {
                                    closestPoint.Container = secondPoint.Container = polygon;
                                    if (startDist > endDist)
                                    {
                                        closestPoint.Index = i;
                                        secondPoint.Index = lastIndex;
                                    }
                                    else
                                    {
                                        closestPoint.Index = lastIndex;
                                        secondPoint.Index = i;
                                    }

                                    closestDist = dist;
                                    closestSecondaryDist = secondaryDist;
                                }

                                lastIndex = i;
                            }
                        }
                    }
                    break;
                case PolyModToolMode.Moving:
                    if (currentPosition != null)
                    {
                        Vector2 movement = currentPosition.Value - startingPosition;

                        for (int i = 0; i < selectedPoints.Length; i++)
                        {
                            selectedPoints[i].Value = initialPositions[i] + movement;
                            selectedPoints[i].Container.Dirty();
                        }
                    }
                    break;
                default:
                    break;
            }
        }

        public override void Draw2D(Microsoft.Xna.Framework.Graphics.SpriteBatch spriteBatch, GameEngine.DrawInfo drawInfo)
        {
            base.Draw2D(spriteBatch, drawInfo);

            spriteBatch.DrawString(screenFont, "Current mode: " + state.ToString() + ", press enter to change", new Vector2(20, 20), Color.White);
            if (state == PolyModToolMode.AddAndRemove)
            {
                spriteBatch.DrawString(screenFont, "Press R the change the winding of the current polygon", new Vector2(20, 40), Color.White);
            }
            if (currentPosition != null)
            spriteBatch.DrawString(screenFont, "Current position: " + currentPosition == null ? "null" : currentPosition.Value.ToString(), new Vector2(20, 60), Color.White);
        }

        public override void Draw(GameEngine.DrawInfo drawInfo)
        {
            if (state == PolyModToolMode.AddAndRemove)
            {
                if (closestPoint.Container != null && secondPoint.Container != null)
                {
                    vertices[0].Position = new Vector3(currentPosition.Value.X, 0, currentPosition.Value.Y);
                    vertices[1].Position = closestPoint.WorldPosition;
                    vertices[2].Position = new Vector3(currentPosition.Value.X, 0, currentPosition.Value.Y);
                    vertices[3].Position = secondPoint.WorldPosition;

                    Vector2 startPoint;
                    Vector2 endPoint;

                    if ((closestPoint.Index + 1) % closestPoint.Container.Points.Count == secondPoint.Index )
                    {
                        startPoint = closestPoint.Value;
                        endPoint = secondPoint.Value;
                    }
                    else
                    {
                        startPoint = secondPoint.Value;
                        endPoint = closestPoint.Value;
                    }
                    Vector2 segmentVector = endPoint - startPoint;
                    Vector3 segmentNormal = new Vector3(-segmentVector.Y, 0, segmentVector.X);
                    segmentNormal.Normalize();

                    vertices[4].Position = closestPoint.WorldPosition + 0.5f * (secondPoint.WorldPosition - closestPoint.WorldPosition);
                    vertices[5].Position = vertices[4].Position + 25 * segmentNormal;

                    primitivesEffect.Parameters["WorldViewProj"].SetValue(drawInfo.Camera.View * drawInfo.Camera.Projection);
                    primitivesEffect.Parameters["Color"].SetValue(Color.White.ToVector4());
                    primitivesEffect.Begin();

                    foreach (EffectPass pass in primitivesEffect.CurrentTechnique.Passes)
                    {
                        pass.Begin();
                        drawInfo.GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineList, vertices, 0, 3);
                        pass.End();
                    }
                    primitivesEffect.End();
                }
            }
        }

        public override void ToolStart()
        {
            base.ToolStart();
            if (tempToolChange)
            {
                tempToolChange = false;
            }
            else
            {
                lastSelectionMode = selection.SelectionMode;
                selection.SelectionMode = SelectionMode.SelectPoints;
            }
        }

        public override void ToolEnd()
        {
            base.ToolEnd();
            if (! tempToolChange)
            {
                selection.SelectionMode = lastSelectionMode;
            }
        }
    }
}
