using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using GameEngine.ScreenSystem;
using Microsoft.Xna.Framework.Graphics;
using GameEngine.Level;
using GameEngine.Helpers;
using Microsoft.Xna.Framework.Input;
using Editor.Data;
using GameEngine;
using GameEngine.Cameras;
using GameEngine.Xml;
using System.Reflection;

namespace Editor.Tools
{
    public enum PhysicsPolygonDrawerState { Initial, Drawing };

    public class PolygonTool : Tool
    {
        public class PolygonEventArgs : EventArgs { public PhysicsPolygon Polygon; }

        public static List<Type> PolygonTypes { get; private set; }

        private List<Vector2> colledtedPoints = new List<Vector2>();
        private PhysicsPolygonDrawerState state = PhysicsPolygonDrawerState.Initial;
        private VertexPositionColor[] vertices = new VertexPositionColor[32];
        private Plane xzPlane = new Plane(new Vector3(0, 0, 0), new Vector3(1, 0, 0), new Vector3(0, 0, 1));
        private VertexDeclaration decl;
        static PolygonTool()
        {
            PolygonTypes = new List<Type>();
            PolygonTypes.AddRange(NodeReflector.GetNodeTypes(typeof(PolygonNode)));
            if (PolygonTypes.Contains(typeof(PolygonWallNode)))
                SelectedType = typeof(PolygonWallNode);
        }
        public static event EventHandler SelectedPolygonTypeChanged;
        private static int selectedTypeIndex = 0;
        public static int SelectedTypeIndex 
        {
            get
            {
                return selectedTypeIndex;
            }
            set
            {
                if (selectedTypeIndex == value)
                    return;

                selectedTypeIndex = value;

                if(SelectedPolygonTypeChanged != null)
                    SelectedPolygonTypeChanged(null, null);

            }
        }
        public static Type SelectedType
        {
            get { return PolygonTypes[SelectedTypeIndex]; }
            set
            {
                for (int i = 0; i < PolygonTypes.Count; i++)
                {
                    if (PolygonTypes[i] == value)
                    {
                        SelectedTypeIndex = i;
                        return;
                    }
                }
                throw new Exception("Type is not PolygonNode or derived from PolygonNode");
            }
        }

        public static void CycleSelectedPolygonType()
        {
            SelectedTypeIndex = (SelectedTypeIndex + 1) % PolygonTypes.Count;
        }

        private Vector2? targetPoint;

        private Effect primitivesEffect;
        private SpriteFont screenFont;

        private float snapToEndDistance = 2f;

        private bool snap = true;

        public event EventHandler<PolygonEventArgs> PolygonCreated;

        public PolygonTool(OrthoOrbitCamera camera, ToolHandler handler)
            : base(camera, handler)
        {
            this.Name = "PolygonTool";
        }

        public override void LoadContent(Microsoft.Xna.Framework.Content.ContentManager content)
        {
            primitivesEffect = content.Load<Effect>("primitives");
            screenFont = content.Load<SpriteFont>("Fonts/Arial");
            
        }


        public override void HandleInput(Microsoft.Xna.Framework.GameTime gameTime)
        {
            if (InputState.IsKeyDown(Keys.LeftShift) || InputState.IsKeyDown(Keys.RightShift))
            {
                snap = false;
            }
            else
            {
                snap = true;
            }

            if (InputState.KeyReleased(Keys.T))
            {
                PolygonTool.CycleSelectedPolygonType();
            }

            switch (state)
            {
                case PhysicsPolygonDrawerState.Initial:
                    // Use all basic controls
                    EnableControls(Controls.All);
                    if (InputState.MouseReleased(MouseButton.Left))
                    {
                        if (targetPoint != null)
                        {
                            vertices[colledtedPoints.Count] = new VertexPositionColor(new Vector3(targetPoint.Value.X, 0, targetPoint.Value.Y), Color.White);
                            colledtedPoints.Add(targetPoint.Value);
                            state = PhysicsPolygonDrawerState.Drawing;
                        }
                    }
                    break;
                case PhysicsPolygonDrawerState.Drawing:
                    // Use all basic controls except CameraToggle (backspace)
                    EnableControls(Controls.CameraToggle, true);
                    if (InputState.MouseReleased(MouseButton.Left))
                    {
                        if (targetPoint != null)
                        {
                            if (colledtedPoints.Count > 2)
                            {
                                Vector2 startDist = targetPoint.Value - colledtedPoints[0];
                                if (startDist.Length() < snapToEndDistance)
                                {
                                    System.Windows.Forms.DialogResult askUser = System.Windows.Forms.MessageBox.Show("Do you want to close this polygon?", "Finish Polygon?", System.Windows.Forms.MessageBoxButtons.YesNo);
                                    if (askUser == System.Windows.Forms.DialogResult.Yes)
                                    {
                                        FinishDrawing();
                                    }
                                    else
                                    {
                                        AddPoint();
                                    }
                                }
                                else
                                {
                                    AddPoint();
                                }
                            }
                            else
                            {
                                AddPoint();
                            }
                        }
                    }
                    if (InputState.KeyReleased(Keys.Back))
                    {
                        colledtedPoints.RemoveAt(colledtedPoints.Count - 1);
                        if (colledtedPoints.Count == 0)
                        {
                            state = PhysicsPolygonDrawerState.Initial;
                        }
                    }
                    if (InputState.KeyReleased(Keys.Escape))
                    {
                        Reset();
                    }
                    


                    if (InputState.MouseReleased(MouseButton.Right))
                    {
                        FinishDrawing();
                    }
                    break;
                default:
                    break;
            }
        }

        public override void Update(GameTime gameTime)
        {
            targetPoint = Cursor.GetXZPlanePosition(snap);
        }


        public override void Draw(DrawInfo drawInfo)
        {
            // Draw the temporary lines and stuff... Need help for this.



            int count = targetPoint != null ? colledtedPoints.Count + 1 : colledtedPoints.Count;
            
            if(count > 1) 
            {
                if (decl == null)
                {
                    decl = new VertexDeclaration(drawInfo.GraphicsDevice, VertexPositionColor.VertexElements);
                }
                drawInfo.GraphicsDevice.VertexDeclaration = decl;

                if(targetPoint != null) 
                {
                    vertices[colledtedPoints.Count].Color = Color.Red;
                    vertices[colledtedPoints.Count].Position = new Vector3(targetPoint.Value.X, 0, targetPoint.Value.Y);
                }

                primitivesEffect.Parameters["WorldViewProj"].SetValue(drawInfo.Camera.View * drawInfo.Camera.Projection);
                primitivesEffect.Parameters["Color"].SetValue(Color.White.ToVector4());
                primitivesEffect.Begin();

                foreach (EffectPass pass in primitivesEffect.CurrentTechnique.Passes)
                {
                    pass.Begin();
                    drawInfo.GraphicsDevice.DrawUserPrimitives<VertexPositionColor>(PrimitiveType.LineStrip, vertices, 0, count - 1);
                    pass.End();
                }
                primitivesEffect.End();

            }

        }

        public override void Draw2D(Microsoft.Xna.Framework.Graphics.SpriteBatch spriteBatch, DrawInfo drawInfo)
        {
            Vector2 drawPosition = new Vector2(5, 5);

            spriteBatch.DrawString(screenFont, SelectedType.Name, drawPosition, Color.White);
            drawPosition.Y += 20;
            if (targetPoint != null)
            {
                spriteBatch.DrawString(screenFont, "Target Point : " + targetPoint.ToString(), drawPosition, Color.White);
            }
            else
            {
                spriteBatch.DrawString(screenFont, "Target Point : null", drawPosition, Color.White);
            }


            drawPosition.Y += 20;

            for (int i = 0; i < colledtedPoints.Count; i++)
            {
                spriteBatch.DrawString(screenFont, "Point " + i.ToString() + ": " + colledtedPoints[i].ToString(), drawPosition, Color.White);
                drawPosition.Y += 20;
            }
        }

        void CheckArraySize()
        {
            if (vertices.Length < colledtedPoints.Count + 2) // One extra point added in next pass and one additional used for the targetPoint
            {
                VertexPositionColor[] newVertices = new VertexPositionColor[vertices.Length * 2];
                for (int i = 0; i < vertices.Length; i++)
                {
                    newVertices[i] = vertices[i];
                }
                vertices = newVertices;
            }
        }

        private void Reset()
        {
            colledtedPoints.Clear();
            state = PhysicsPolygonDrawerState.Initial;
        }

        private void FinishDrawing()
        {
            // Do we tell the user that he's an idiot if he tries to make a polygon of <= 2 points?
            if (colledtedPoints.Count > 2)
            {
                PhysicsPolygon newPolygon = new PhysicsPolygon(colledtedPoints.ToArray());
                // Need to set effect here since the polygon is created at runtime and won't have LoadContent called
                newPolygon.SetEffect(primitivesEffect);

                if(PolygonCreated != null) {
                    PolygonEventArgs args = new PolygonEventArgs();
                    args.Polygon = newPolygon;
                    PolygonCreated(this, args);
                }
            }
            colledtedPoints = new List<Vector2>();
            state = PhysicsPolygonDrawerState.Initial;
        }

        private void AddPoint()
        {
            CheckArraySize();
            vertices[colledtedPoints.Count] = new VertexPositionColor(new Vector3(targetPoint.Value.X, 0, targetPoint.Value.Y), Color.White);
            colledtedPoints.Add(targetPoint.Value);
        }

        public override void ToolStart()
        {
        }

        public override void ToolEnd()
        {
            Reset();
        }
    }
}
