using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using GameEngine.ScreenSystem;
using GameEngine.Cameras;
using GameEngine.Helpers;
using Editor.Editor;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Graphics;
using GameEngine;
using GameEngine.Xml;

namespace Editor.Tools
{
    public class RotateToolRotationAxisChangedEventArgs : EventArgs
    {
        private Axis oldAxis, newAxis;
        public Axis OldAxis { get { return this.oldAxis; } }
        public Axis NewAxis { get { return this.newAxis; } }

        public RotateToolRotationAxisChangedEventArgs(Axis oldAxis, Axis newAxis)
        {
            this.oldAxis = oldAxis;
            this.newAxis = newAxis;
        }
    }

    public delegate void RotateToolRotationAxisChangedHandler(object sender, RotateToolRotationAxisChangedEventArgs e);

    public class RotateTool : Tool
    {
        public float RotateSpeed = 0.005f;

        private bool working;
        private Vector2 startingPoint;

        private Quaternion selectionRotation;
        private List<Quaternion> startingRotations = new List<Quaternion>();
        private List<Vector3> startingPositions = new List<Vector3>();

        private Axis rotationAxis = Axis.Y;
        public Axis RotationAxis { get { return rotationAxis; } }
        public event RotateToolRotationAxisChangedHandler RotationAxisChanged; 

        private float snapDistance = MathHelper.Pi / 8;
        private float halfSnapDistance = MathHelper.Pi / 16;

        private Grid rotationGrid;
        public bool HelperGridVisible { get { return rotationGrid.Visible; } }
        public event HelperGridVisibilityChangedHandler HelperGridVisibilityChanged;
        private Effect primitivesEffect;
        private bool gridInitialized = false;

        private SpriteFont screenFont;


        ObjectNode[] selection;

        public RotateTool(OrthoOrbitCamera camera, ToolHandler handler)
            : base(camera, handler)
        {
            this.Name = "RotateTool";
            working = false;

            this.rotationGrid = new Grid(16, 16, 10f, Color.White);
            this.rotationGrid.Visible = false;

        }

        public override void LoadContent(Microsoft.Xna.Framework.Content.ContentManager content)
        {
            screenFont = content.Load<SpriteFont>("Fonts/Arial");
            primitivesEffect = content.Load<Effect>("primitives");            
        }

        public override void Update(Microsoft.Xna.Framework.GameTime gameTime)
        {
            rotationGrid.Update();
        }

        public override void Draw(DrawInfo drawInfo)
        {
            if (!gridInitialized)
            {
                rotationGrid.Initialize(drawInfo.GraphicsDevice, primitivesEffect);
                gridInitialized = true;
            }

            rotationGrid.Draw(drawInfo);            
        }

        public override void Draw2D(Microsoft.Xna.Framework.Graphics.SpriteBatch spriteBatch, DrawInfo drawInfo)
        {
            spriteBatch.DrawString(screenFont, "Rotating around: " + rotationAxis.ToString(), new Vector2(30, 30), Color.White);
            string helpgridEnabled = rotationGrid.Visible ? "on" : "off";
            spriteBatch.DrawString(screenFont, "Helping grid: " + helpgridEnabled + ", press enter to toggle", new Vector2(30, 50), Color.White);            
        }

        public override void ToolStart()
        {
            
        }

        public override void ToolEnd()
        {
            
        }

        public virtual void OnRotationAxisChanged(RotateToolRotationAxisChangedEventArgs e)
        {
            if (RotationAxisChanged != null)
                RotationAxisChanged(this, e);
        }

        public void SetRotationAxis(Axis newAxis)
        {
            Axis oldAxis = rotationAxis;
            rotationAxis = newAxis;
            ChangeGrid();
            OnRotationAxisChanged(new RotateToolRotationAxisChangedEventArgs(oldAxis, newAxis));
        }

        public virtual void OnHelperGridVisibilityChanged(HelperGridVisibilityChangedEventArgs e)
        {
            if (HelperGridVisibilityChanged != null)
                HelperGridVisibilityChanged(this, e);
        }

        public void ToggleHelperGrid()
        {
            rotationGrid.Visible = !rotationGrid.Visible;
            ChangeGrid();
            OnHelperGridVisibilityChanged(new HelperGridVisibilityChangedEventArgs(rotationGrid.Visible));
        }

        public void SetHelperGridVisibility(bool newValue)
        {
            rotationGrid.Visible = newValue;
            ChangeGrid();
            OnHelperGridVisibilityChanged(new HelperGridVisibilityChangedEventArgs(newValue));
        }

        override public void HandleInput(GameTime gameTime)
        {
            EnableControls(enabledControls);

            if (InputState.KeyPressed(Keys.X))
                SetRotationAxis(Axis.X);
            if (InputState.KeyPressed(Keys.Y) || InputState.KeyPressed(Keys.C))
                SetRotationAxis(Axis.Y);
            if (InputState.KeyPressed(Keys.Z))
                SetRotationAxis(Axis.Z);

            if (InputState.KeyReleased(Keys.Enter))
                ToggleHelperGrid();

            if (toolHandler.Selection.Count > 0)
            {
                if (!working)
                {
                    if (InputState.MousePressed(MouseButton.Left))
                    {
                        if (toolHandler.Selection.Intersects(camera.GetPickingRay(Cursor.Position)))
                        {
                            working = true;

                            selection = toolHandler.Selection.Get3DSelection();

                            startingRotations.Clear();
                            startingPositions.Clear();
                            for (int i = 0; i < selection.Length; i++)
                            {
                                startingRotations.Add(selection[i].Rotation);
                                startingPositions.Add(selection[i].WorldPosition);
                            }

                            startingPoint = Cursor.Position;

                            selectionRotation = toolHandler.Selection.SelectionRotation;

                            ChangeGrid();
                        }
                        else
                        {
                            toolHandler.SwitchTemporarilyToSelection(gameTime);
                        }
                    }
                }
                else
                {
                    if (InputState.MouseReleased(MouseButton.Left))
                    {
                        working = false;
                        ChangeGrid();
                    }

                    float delta = (startingPoint.X - Cursor.Position.X) * RotateSpeed;

                    if (!(InputState.IsKeyDown(Keys.LeftShift) || InputState.IsKeyDown(Keys.RightShift)))
                    {
                        delta = SnapAngle(delta);
                    }

                    switch (rotationAxis)
                    {
                        case Axis.X:
                            RotateSelection(Quaternion.CreateFromAxisAngle(Vector3.UnitX, -delta));
                            break;
                        case Axis.Y:
                            RotateSelection(Quaternion.CreateFromAxisAngle(Vector3.UnitY, -delta));
                            break;
                        case Axis.Z:
                            RotateSelection(Quaternion.CreateFromAxisAngle(Vector3.UnitZ, -delta));
                            break;
                        default:
                            break;
                    }
                }
            }
            else
            {
                if(InputState.MousePressed(MouseButton.Left))
                    toolHandler.SwitchTemporarilyToSelection(gameTime);
            }
        }

        public void RotateSelection(Quaternion rotation)
        {
            if (selection.Length == 0)
            {
                return;
            }
            if (selection.Length == 1)
            {
                selection[0].Rotation = Quaternion.Concatenate(startingRotations[0], rotation);
            }
            else
            {
                for (int i = 0; i < selection.Length; i++)
                {
                    selection[i].Rotation = Quaternion.Concatenate(startingRotations[i], rotation);

                    
                    // Rotate offset from center of selection
                    Vector3 offset = startingPositions[i] - toolHandler.Selection.SelectionCenter;
                    selection[i].WorldPosition = startingPositions[i] + Vector3.Transform(offset, rotation) - offset;
                }

                toolHandler.Selection.SelectionRotation = Quaternion.Concatenate(selectionRotation, rotation);
            }
        }

        public float SnapAngle(float input)
        {
            float rest = input % snapDistance;

            float result = input - rest;

            if (Math.Abs(rest) > halfSnapDistance)
                result += Math.Sign(rest) * snapDistance;

            return result;
        }

        public void ChangeGrid()
        {
            switch (rotationAxis)
            {
                case Axis.X:
                    rotationGrid.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitZ, MathHelper.PiOver2);
                    break;
                case Axis.Y:
                    rotationGrid.Rotation = Quaternion.Identity;
                    break;
                case Axis.Z:
                    rotationGrid.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitX, MathHelper.PiOver2);
                    break;
                default:
                    break;
            }

            if (toolHandler.Selection.Count > 0)
            {
                rotationGrid.Position = toolHandler.Selection.SelectionCenter;
            }
        }

    }
}
