using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using GameEngine.ScreenSystem;
using GameEngine.Helpers;
using Microsoft.Xna.Framework.Graphics;
using GameEngine.Level;
using GameEngine;
using GameEngine.Xml;
using Editor;
using Microsoft.Xna.Framework.Input;
using GameEngine.Cameras;

namespace Editor.Tools
{
    class SelectionTool : Tool
    {
        public enum SelectionToolState { Initial, DraggingBox };

        Vector2 initialScreenPosition;
        Ray topLeftRay, topRightRay, bottomLeftRay, bottomRightRay;
        Vector2 currentScreenPosition;

        SelectionToolState state;
        public SelectionToolState State { get { return state; } }

        private SelectionHandler selection;

        private string returnTool = "";
        public string ReturnTool { get { return returnTool;} set { returnTool = value; } }

        Effect primitivesEffect;

        Texture2D dotTexture;

        Vector2 topLeft, bottomRight;

        SelectionFrustum selectionFrustum;

        float frostumDepth = 10000f;

        SpriteFont screenFont;

        int width, height;

        bool append = false;

        private List<ObjectNode> selectableObjects3D;
        private List<PolygonNode> selectableObjects2D;

        Ray selectionRay;

        public SelectionTool(OrthoOrbitCamera camera, ToolHandler handler)
            : base(camera, handler)
        {
            this.Name = "SelectionTool";
            this.selection = handler.Selection;
        }

        public void SetObjectCollections(List<ObjectNode> sel3DCollection, List<PolygonNode> sel2DCollection)
        {
            this.selectableObjects3D = sel3DCollection;
            this.selectableObjects2D = sel2DCollection;
        }

        public override void LoadContent(Microsoft.Xna.Framework.Content.ContentManager content)
        {
            dotTexture = content.Load<Texture2D>("Bitmaps/dot");
            primitivesEffect = content.Load<Effect>("primitives");

            screenFont = content.Load<SpriteFont>("Fonts/Arial");
        }

        public override void HandleInput(GameTime gameTime)
        {
            EnableControls(enabledControls);

            switch (state)
            {
                case SelectionToolState.Initial:
                    if (InputState.MousePressed(MouseButton.Left))
                    {
                        StartSelection();
                    }
                    break;
                case SelectionToolState.DraggingBox:
                    if (InputState.MouseReleased(MouseButton.Left))
                    {
                        EndSelection();
                    }

                    if (append && !(InputState.IsKeyDown(Keys.LeftControl) || InputState.IsKeyDown(Keys.RightControl)))
                    {
                        append = false;
                    }
                    break;
                default:
                    break;
            }

        }

        public override void Update(Microsoft.Xna.Framework.GameTime gameTime)
        {
            if (state == SelectionToolState.DraggingBox)
            {
                currentScreenPosition = Cursor.Position;
                selectionRay = camera.GetPickingRay(currentScreenPosition);

                topLeft = new Vector2(Math.Min(initialScreenPosition.X, currentScreenPosition.X), Math.Min(initialScreenPosition.Y, currentScreenPosition.Y));
                bottomRight = new Vector2(Math.Max(initialScreenPosition.X, currentScreenPosition.X), Math.Max(initialScreenPosition.Y, currentScreenPosition.Y));
                width = (int)(bottomRight.X - topLeft.X);
                height = (int)(bottomRight.Y - topLeft.Y);

                selection.BeginSelection();

                switch (selection.SelectionMode)
                {
                    case SelectionMode.Select2D:
                        break;
                    case SelectionMode.Select3D:
                        Handle3DSelection();
                        break;
                    case SelectionMode.SelectPoints:
                        HandlePointSelection();
                        break;
                    default:
                        break;
                }


            }
        }

        private void Handle3DSelection()
        {
            if (append)
            {
                for (int i = 0; i < selection.OldSelectedNodes.Count; i++)
                {
                    ObjectNode node = selection.OldSelectedNodes[i] as ObjectNode;
                    if (node != null)
                    {
                        selection.AddToSelection(node);
                    }
                }
            }

            if (width >= 1 && height >= 1)
            {
                UpdateSelectionFrustum();

                for (int i = 0; i < selectableObjects3D.Count; i++)
                {

                    if (selectionFrustum.Intersects(selectableObjects3D[i].OrientedBoundingBox))
                    {
                        if (append)
                        {
                            if (selection.OldSelectedNodes.Contains(selectableObjects3D[i]))
                            {
                                selection.RemoveFromSelection(selectableObjects3D[i]);
                            }
                            else
                            {
                                selection.AddToSelection(selectableObjects3D[i]);
                            }
                        }
                        else
                        {
                            selection.AddToSelection(selectableObjects3D[i] as ObjectNode);
                        }
                    }
                }
            }
            else
            {
                ObjectNode closestModel = null;
                // If we only had one object in last selection try not to select that object again.
                // Instead select any object behind it
                if (selection.OldSelectedNodes.Count == 1)
                {
                    ObjectNode lastSelected = selection.OldSelectedNodes[0] as ObjectNode;
                    if (lastSelected != null)
                    {
                        float testMin, testMax;
                        if (lastSelected.OrientedBoundingBox.Intersects(selectionRay, out testMin, out testMax))
                        {
                            closestModel = GetModelByRay(selectionRay, testMin);
                        }
                    }
                }
                if (closestModel == null)
                    closestModel = GetModelByRay(selectionRay, float.MinValue);
                if (closestModel != null)
                {
                    if (append)
                    {
                        if (selection.OldSelectedNodes.Contains(closestModel))
                        {
                            selection.RemoveFromSelection(closestModel);
                        }
                        else
                        {
                            selection.AddToSelection(closestModel as ObjectNode);
                        }
                    }
                    else
                    {
                        selection.AddToSelection(closestModel as ObjectNode);
                    }
                }
            }
        }

        private void HandlePointSelection()
        {
            if (append)
            {
                for (int i = 0; i < selection.OldSelectedPoints.Count; i++)
                {
                    SelectionPoint point = selection.OldSelectedPoints[i];
                    selection.AddToSelection(point.Container, point.Index);
                }
            }

            if (width > 0 && height > 0)
            {
                UpdateSelectionFrustum();
                for (int i = 0; i < selectableObjects2D.Count; i++)
                {
                    for (int j = 0; j < selectableObjects2D[i].Points.Count; j++)
                    {
                        if (selectionFrustum.Contains(new Vector3(selectableObjects2D[i].Points[j].X, 0, selectableObjects2D[i].Points[j].Y) + selectableObjects2D[i].WorldPosition))
                        {
                            if (append)
                            {
                                SelectionPoint thisPoint = new SelectionPoint(selectableObjects2D[i], j);
                                if(selection.OldSelectedPoints.Contains(thisPoint)) 
                                {
                                    selection.RemoveFromSelection(thisPoint.Container, thisPoint.Index);
                                }
                                else
                                {
                                    selection.AddToSelection(thisPoint.Container, thisPoint.Index);
                                }
                            }
                            else
                            {
                                selection.AddToSelection(selectableObjects2D[i], j);
                            }
                        }
                    }
                }
            }
            else
            {
                Vector2? cursorXZPos = Cursor.GetXZPlanePosition();
                if(cursorXZPos != null) {
                    for (int i = 0; i < selectableObjects2D.Count; i++)
                    {
                        PolygonNode polyNode = selectableObjects2D[i];
                        for (int j = 0; j < polyNode.Points.Count; j++)
                        {
                            Vector2 pointWorldPosition = new Vector2(polyNode.WorldPosition.X + polyNode.Points[j].X, polyNode.WorldPosition.Z + polyNode.Points[j].Y);
                            if (Math.Abs(cursorXZPos.Value.X - pointWorldPosition.X) <= 5f && Math.Abs(cursorXZPos.Value.Y - pointWorldPosition.Y) <= 5f)
                            {
                                if (append)
                                {
                                    SelectionPoint thisPoint = new SelectionPoint(polyNode, j);
                                    if (selection.OldSelectedPoints.Contains(thisPoint))
                                    {
                                        selection.RemoveFromSelection(thisPoint.Container, thisPoint.Index);
                                    }
                                    else
                                    {
                                        selection.AddToSelection(thisPoint.Container, thisPoint.Index);
                                    }
                                }
                                else
                                {
                                    selection.AddToSelection(polyNode, j);
                                }
                            }
                        }
                    }
                }
            }
        }

        public void UpdateSelectionFrustum()
        {
            topLeftRay = camera.GetPickingRay(topLeft);
            topRightRay = camera.GetPickingRay(topLeft + new Vector2(width, 0));
            bottomLeftRay = camera.GetPickingRay(topLeft + new Vector2(0, height));
            bottomRightRay = camera.GetPickingRay(bottomRight);

            Vector3 farTopLeft = topLeftRay.Position + frostumDepth * topLeftRay.Direction;
            Vector3 farBottomLeft = bottomLeftRay.Position + frostumDepth * bottomLeftRay.Direction;
            Vector3 farBottomRight = bottomRightRay.Position + frostumDepth * bottomRightRay.Direction;

            selectionFrustum = new SelectionFrustum(topRightRay.Position, topLeftRay.Position, bottomRightRay.Position, farBottomLeft, farBottomRight, farTopLeft);
        }

        private ObjectNode GetModelByRay(Ray ray, float minDistance)
        {
            ObjectNode closestModel = null;
            float testMin, testMax;
            float closestMin = float.MaxValue;

            for (int i = 0; i < selectableObjects3D.Count; i++)
            {
                if (selectableObjects3D[i].OrientedBoundingBox.Intersects(ray, out testMin, out testMax))
                {
                    if (testMin < closestMin && testMin > minDistance)
                    {
                        closestMin = testMin;
                        closestModel = selectableObjects3D[i];
                    }
                }
            }

            return closestModel;
        }

        public override void Draw(DrawInfo drawInfo)
        {
        }

        public override void Draw2D(Microsoft.Xna.Framework.Graphics.SpriteBatch spriteBatch, DrawInfo drawInfo)
        {
            spriteBatch.DrawString(screenFont, "SelectionMode: " + selection.SelectionMode.ToString(), new Vector2(20, 20), Color.White);
            spriteBatch.DrawString(screenFont, "Selection contains " + selection.Count.ToString() + " items", new Vector2(20, 40), Color.White);
            switch (state)
            {
                case SelectionToolState.Initial:
                    break;
                case SelectionToolState.DraggingBox:
                    spriteBatch.Draw(dotTexture, new Rectangle((int)topLeft.X, (int)topLeft.Y, width, 1), Color.White);
                    spriteBatch.Draw(dotTexture, new Rectangle((int)topLeft.X, (int)topLeft.Y, 1, height), Color.White);
                    spriteBatch.Draw(dotTexture, new Rectangle((int)topLeft.X, (int)bottomRight.Y, width, 1), Color.White);
                    spriteBatch.Draw(dotTexture, new Rectangle((int)bottomRight.X, (int)topLeft.Y, 1, height), Color.White);

                    break;
                default:
                    break;
            }

        }

        public override void ToolStart()
        {
        }

        public override void ToolEnd()
        {
        }

        private void StartSelection()
        {
            initialScreenPosition = Cursor.Position;
            append = (InputState.IsKeyDown(Keys.LeftControl) || InputState.IsKeyDown(Keys.RightControl));
            state = SelectionToolState.DraggingBox;
        }

        private void EndSelection()
        {
            selection.EndSelection();

            state = SelectionToolState.Initial;
            if (returnTool != "")
            {
                toolHandler.SelectTool(returnTool);
                returnTool = "";
            }
        }
    }
}

