using System;
using System.Collections.Generic;
using System.Text;
using GameEngine.Animation;
using Microsoft.Xna.Framework.Content.Pipeline.Graphics;
using Microsoft.Xna.Framework.Content.Pipeline;
using Microsoft.Xna.Framework;

namespace GameContentPipeline
{
    public class AnimationDataFactory
    {
        private const int MaxBones = 59;

        public static AnimationData GetAnimationData(NodeContent input, ContentProcessorContext context)
        {
            AnimationData data = new AnimationData();

            ValidateMesh(input, context, null);

            BoneContent skeleton = MeshHelper.FindSkeleton(input);

            if (skeleton == null)
            {
                // Vi antager at der stadig er rigid animations...
                data.Type = AnimationType.Rigid;
                
                

                foreach (var anim in input.Animations)
                {
                    foreach (var channel in anim.Value.Channels)
                    {
                        //AnimationChannel c;
                        
                        // Convert the keyframe data.
                        foreach (AnimationKeyframe keyframe in channel.Value)
                        {
//                            Keyframe kf = new Keyframe(
//                            keyframes.Add(new Keyframe(boneIndex, keyframe.Time, keyframe.Transform));
                        }


                    }
                }

            }
            else
            {
                data.Type = AnimationType.Skinnned;

                FlattenTransforms(input, skeleton);

                List<BoneContent> bones = (List<BoneContent>)MeshHelper.FlattenSkeleton(skeleton);

                if (bones.Count > MaxBones)
                {
                    throw new InvalidContentException(string.Format(
                        "Skeleton has {0} bones, but the maximum supported is {1}.",
                        bones.Count, MaxBones));
                }

                List<Matrix> bindPose = new List<Matrix>();
                List<Matrix> inverseBindPose = new List<Matrix>();
                List<int> skeletonHierarchy = new List<int>();

                foreach (BoneContent bone in bones)
                {
                    bindPose.Add(bone.Transform);
                    inverseBindPose.Add(Matrix.Invert(bone.AbsoluteTransform));
                    skeletonHierarchy.Add(bones.IndexOf(bone.Parent as BoneContent));
                }

                // Convert animation data to our runtime format.
                Dictionary<string, AnimationTrack> animationTracks;
                animationTracks = ProcessAnimations(skeleton.Animations, bones);
                data.AminationTracks = animationTracks;
            }


            
            return data;
        }

        /// <summary>
        /// Bakes unwanted transforms into the model geometry, so everything ends up in the same coordinate system.
        /// </summary>
        private static void FlattenTransforms(NodeContent node, BoneContent skeleton)
        {
            foreach (NodeContent child in node.Children)
            {
                // Don't process the skeleton, because that is special.
                if (child == skeleton)
                    continue;

                // Bake the local transform into the actual geometry.
                MeshHelper.TransformScene(child, child.Transform);

                // Having baked it, we can now set the local
                // coordinate system back to identity.
                child.Transform = Matrix.Identity;

                // Recurse.
                FlattenTransforms(child, skeleton);
            }
        }

        /// <summary>
        /// Makes sure this mesh contains the kind of data we know how to animate.
        /// </summary>
        private static bool ValidateMesh(NodeContent node, ContentProcessorContext context, string parentBoneName)
        {
            MeshContent mesh = node as MeshContent;

            if (mesh != null)
            {
                // Validate the mesh.
                if (parentBoneName != null)
                {
                    context.Logger.LogWarning(null, null,
                        "Mesh {0} is a child of bone {1}. SkinnedModelProcessor " +
                        "does not correctly handle meshes that are children of bones.",
                        mesh.Name, parentBoneName);
                }

                /*if (!MeshHasSkinning(mesh))
                {
                    context.Logger.LogWarning(null, null,
                        "Mesh {0} has no skinning information, so it has been deleted.",
                        mesh.Name);

                    mesh.Parent.Children.Remove(mesh);
                    return;
                }*/
            }
            else if (node is BoneContent)
            {
                // If this is a bone, remember that we are now looking inside it.
                parentBoneName = node.Name;
            }
            else
            {
                return false;
            }
            bool ok = true;
            // Recurse (iterating over a copy of the child collection,
            // because validating children may delete some of them).
            foreach (NodeContent child in new List<NodeContent>(node.Children))
            {
                ok = ok && ValidateMesh(child, context, parentBoneName);
            }
            return ok;
        }

        /// <summary>
        /// Checks whether a mesh contains skininng information.
        /// </summary>
        private static bool MeshHasSkinning(MeshContent mesh)
        {
            foreach (GeometryContent geometry in mesh.Geometry)
            {
                if (!geometry.Vertices.Channels.Contains(VertexChannelNames.Weights()))
                    return false;
            }

            return true;
        }

        /// <summary>
        /// Converts an intermediate format content pipeline AnimationContentDictionary
        /// object to our runtime AnimationClip format.
        /// </summary>
        private static Dictionary<string, AnimationTrack> ProcessAnimations
            (AnimationContentDictionary animations, List<BoneContent> bones)
        {
            // Build up a table mapping bone names to indices.
            Dictionary<string, int> boneMap = new Dictionary<string, int>();

            for (int i = 0; i < bones.Count; i++)
            {
                string boneName = bones[i].Name;

                if (!string.IsNullOrEmpty(boneName))
                    boneMap.Add(boneName, i);
            }

            // Convert each animation in turn.
            Dictionary<string, AnimationTrack> animationTracks;
            animationTracks = new Dictionary<string, AnimationTrack>();

            foreach (KeyValuePair<string, AnimationContent> animation in animations)
            {
                AnimationTrack processed = ProcessAnimation(animation.Value, boneMap);
                animationTracks.Add(animation.Key, processed);
            }

            if (animationTracks.Count == 0)
            {
                throw new InvalidContentException("Input file does not contain any animations.");
            }

            return animationTracks;
        }

        /// <summary>
        /// Converts an intermediate format content pipeline AnimationContent
        /// object to our runtime AnimationClip format.
        /// </summary>
        private static AnimationTrack ProcessAnimation(AnimationContent animation, Dictionary<string, int> boneMap)
        {
            List<Keyframe> keyframes = new List<Keyframe>();

            // For each input animation channel.
            foreach (KeyValuePair<string, AnimationChannel> channel in animation.Channels)
            {
                // Look up what bone this channel is controlling.
                int boneIndex;

                if (!boneMap.TryGetValue(channel.Key, out boneIndex))
                {
                    throw new InvalidContentException(string.Format(
                        "Found animation for bone '{0}', " +
                        "which is not part of the skeleton.", channel.Key));
                }

                // Convert the keyframe data.
                foreach (AnimationKeyframe keyframe in channel.Value)
                {
                    keyframes.Add(new Keyframe(boneIndex, keyframe.Time, keyframe.Transform));
                }
            }

            // Sort the merged keyframes by time.
            keyframes.Sort(CompareKeyframeTimes);

            if (keyframes.Count == 0)
                throw new InvalidContentException("Animation has no keyframes.");

            if (animation.Duration <= TimeSpan.Zero)
                throw new InvalidContentException("Animation has a zero duration.");

            return new AnimationTrack(animation.Duration, keyframes);
        }

        private static int CompareKeyframeTimes(Keyframe a, Keyframe b)
        {
            return a.Time.CompareTo(b.Time);
        }
    }
}
