﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace GameEngine.Cameras
{
    public class ChaseCamera : Camera
    {
        public Vector3 ChasePosition { get; set; } // Position of object being chased.
        public Vector3 ChaseDirection { get; set; } // Direction the chased object is facing.
        public Vector3 DesiredPositionOffset { get; set; } // Desired camera position in the chased object's coordinate system.
        public Vector3 LookAtOffset { get; set; } // Look at point in the chased object's coordinate system.
        public Vector3 Velocity { get; set; } // Velocity of camera.
        public float Stiffness { get; set; } // Physics coefficient which controls the influence of the camera's position over the spring force. The stiffer
                                             // the spring, the closer it will stay to the chased object.
        public float Damping { get; set; } // Physics coefficient which approximates internal friction of the spring. Sufficient damping will prevent the spring from oscillating infinitely.
        public float Mass { get; set; } // Mass of the camera body. Heaver objects require stiffer springs with less damping to move at the same rate as lighter objects.

        private Vector3 desiredPosition;
        private Vector3 lookAt;
        public Vector3 LookAt { get { UpdateWorldPositions(); return lookAt; } } // Look at point in world space.
        public Vector3 DesiredPosition { get { UpdateWorldPositions(); return desiredPosition; } } // Desired camera position in world space.

        public ChaseCamera(float fieldOfView, Vector3 position, Vector3 target, Vector3 up, Viewport viewPortp, float nearPlane, float farPlane, 
            Vector3 chasePosition, Vector3 chaseDirection)
            : base(fieldOfView, position, target, up, viewPortp, nearPlane, farPlane)
        {
            this.ChasePosition = chasePosition;
            this.ChaseDirection = chaseDirection;

            this.DesiredPositionOffset = new Vector3(0, 2f, 2f);
            this.LookAtOffset = new Vector3(0, 2.8f, 0);
            this.Stiffness = 1800.0f;
            this.Damping = 600.0f;
            this.Mass = 50.0f;
        }

        public override void UpdateCamera(GameTime gameTime)
        {
            if (gameTime == null) throw new ArgumentNullException("gameTime");

            UpdateWorldPositions();

            float elapsed = gameTime.ElapsedSeconds();

            // Calculate spring force
            Vector3 stretch = Position - desiredPosition;
            Vector3 force = -Stiffness * stretch - Damping * Velocity;

            // Apply acceleration
            Vector3 acceleration = force / Mass;
            Velocity += acceleration * elapsed;

            // Apply velocity
            Position += Velocity * elapsed;

            UpdateMatrices();            
        }

        private void UpdateWorldPositions()
        {
            Matrix transform = Matrix.Identity;
            transform.Forward = ChaseDirection;
            transform.Up = Up;
            transform.Right = Vector3.Cross(Up, ChaseDirection);

            desiredPosition = ChasePosition + Vector3.TransformNormal(DesiredPositionOffset, transform);
            lookAt = ChasePosition + Vector3.TransformNormal(LookAtOffset, transform);
        }

        private void UpdateMatrices()
        {
            view = Matrix.CreateLookAt(this.Position, this.LookAt, this.Up);
            projection = Matrix.CreatePerspectiveFieldOfView(this.FieldOfView, this.Viewport.AspectRatio, this.NearPlane, this.FarPlane);
            frustrum = new BoundingFrustum(view * projection);

        }

        public void Reset()
        {
            UpdateWorldPositions();

            Velocity = Vector3.Zero;

            Position = desiredPosition;

            UpdateMatrices();
        }

    }
}
