using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace GameEngine.Cameras
{
    public enum CameraMode
    {
        Orthographic,
        Perspective
    }

    public class OrthoOrbitCamera : Camera
    {
        private CameraMode mode;
        public CameraMode Mode { get { return mode; } }

        public OrthoOrbitCamera(CameraMode mode) : base()
        {
            this.mode = mode;
        }

        public OrthoOrbitCamera
            (float fieldOfView, Vector3 position, Vector3 target, Vector3 up, Viewport viewPortp, float nearPlane, float farPlane, CameraMode mode) 
            : base(fieldOfView, position, target, up, viewPortp, nearPlane, farPlane)
        {
            this.mode = mode;
        }

        public void SetMode(CameraMode mode)
        {
            this.mode = mode;
        }
                

        public void PushInDirection(Vector3 direction)
        {
            Vector3 camDirection;
            Vector3 side;
            Vector3 camUp;
            Vector3 movement;
            camDirection = Target - Position;
            camDirection.Normalize();

            side = Vector3.Cross(camDirection, Up);
            side.Normalize();

            camUp = Vector3.Cross(side, camDirection);
            movement = camDirection * direction.Z + side * direction.X + camUp * direction.Y;
                        
            Position += movement;
            Target += movement;
        }

        public void Zoom(float f)
        {
            Vector3 camDirection;
            camDirection = Target - Position;
            camDirection.Normalize();
            Position += camDirection * f;
        }

        public override void UpdateCamera(GameTime gameTime)
        {
            switch (mode)
            {
                case CameraMode.Orthographic:
                    Matrix.CreateLookAt(ref Position, ref Target, ref Up, out view);
                    float zDiff = Position.Y - Target.Y;
                    Matrix.CreateOrthographic(Viewport.Width * 0.01f * zDiff, Viewport.Height * 0.01f * zDiff, nearPlane, farPlane, out projection);
                    break;
                case CameraMode.Perspective:
                    Matrix.CreateLookAt(ref Position, ref Target, ref Up, out view);
                    Matrix.CreatePerspectiveFieldOfView(FieldOfView, viewPort.AspectRatio, nearPlane, farPlane, out projection);
                    break;
                default:
                    break;
            }
            frustrum = new BoundingFrustum(View * Projection);
        }

        /// <summary>
        /// Orbit around the specified target point.
        /// </summary>
        /// <param name="horizontalOffSet">The value to move horizontally</param>
        /// <param name="verticalOffSet">The value to move vertically</param>
        /// <param name="orbitTarget">The center point to orbit around</param>
        public void Orbit(float horizontalOffSet, float verticalOffSet, Vector3 orbitTarget)
        {
            // Get a vector from current camera position to the specified orbit target
            Vector3 targetVector = Vector3.Subtract(Position, orbitTarget);

            // Calculate the spherical coordinates
            float r = targetVector.Length();
            float theta = (float)Math.Atan2(targetVector.X, targetVector.Z);
            float phi = (float)Math.Acos(targetVector.Y / r);

            // Apply the offset values to the spherical coordinates
            theta += horizontalOffSet;
            // Avoid annoying behavoir around top and bottom positions
            phi = (verticalOffSet > 0 && phi < 3.1f) ? phi + verticalOffSet : phi;
            phi = (verticalOffSet < 0 && phi > 0.1f) ? phi + verticalOffSet : phi;

            // Calculate new position vector based on the moved coordinates and apply it.
            this.Position += Vector3.Subtract(new Vector3(
                (float)(r * Math.Sin(theta) * Math.Sin(phi)),
                (float)(r * Math.Cos(phi)),
                (float)(r * Math.Cos(theta) * Math.Sin(phi))
                ), targetVector);    
        }

        /// <summary>
        /// Orbit around camera target.
        /// </summary>        
        public void Orbit(float horizontalOffSet, float verticalOffSet)
        {
            Orbit(horizontalOffSet, verticalOffSet, this.Target);
        }

        public static Vector3 Orbit(Vector3 input, Vector3 center, float horizontalOffSet, float verticalOffSet)
        {
            Vector3 diff = Vector3.Subtract(input, center);

            float r = diff.Length();
            float theta = (float)Math.Atan2(diff.X, diff.Z);
            float phi = (float)Math.Acos(diff.Y / r);

            phi = (verticalOffSet > 0 && phi < 3.1f) ? phi + verticalOffSet : phi;
            phi = (verticalOffSet < 0 && phi > 0.1f) ? phi + verticalOffSet : phi;

            theta += horizontalOffSet;

            Vector3 orbit = new Vector3(
                (float)(r * Math.Sin(theta) * Math.Sin(phi)),
                (float)(r * Math.Cos(phi)),
                (float)(r * Math.Cos(theta) * Math.Sin(phi))
                );

            return input + (orbit - diff);
        }

    }
}
