﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;

namespace GameEngine.Cameras
{
    public class ScriptableCamera : Camera
    {
        private Dictionary<string, object[]> tracks;
        private bool inMovement;
        private Vector3[] positions;
        private Vector3[] targets;
        private float speed;
        private int index;
        private float t;

        public ScriptableCamera(float fieldOfView, Vector3 position, Vector3 target, Vector3 up, Viewport viewPortp, float nearPlane, float farPlane)
            : base(fieldOfView, position, target, up, viewPortp, nearPlane, farPlane)
        {
            this.tracks = new Dictionary<string, object[]>();
            this.inMovement = false;
            this.speed = 0;
            this.index = 0;
            this.t = 0;
        }

        public override void UpdateCamera(GameTime gameTime)
        {
            if (inMovement)
            {
                t += gameTime.ElapsedMillis() * speed * 0.001f;
                ProgressCamera();
            }

            Matrix.CreateLookAt(ref Position, ref Target, ref Up, out view);
            Matrix.CreatePerspectiveFieldOfView(FieldOfView, viewPort.AspectRatio, nearPlane, farPlane, out projection);
        }

        private void ProgressCamera()
        {
            if (t >= 0.5f)
            {
                if (index >= positions.Length - 3)
                {
                    inMovement = false;
                    return;
                }
                else
                {
                    t = 0;
                    index++;
                }
            }
            else
            {
                this.Position = Interpolate(positions[index], positions[index + 1], positions[index + 2], t);
                this.Target = Interpolate(targets[index], targets[index + 1], targets[index + 2], t);
            }
        }

        /// <summary>
        /// Add a new camera track. The two array parameters must contain the same number of elements.
        /// </summary>
        /// <param name="id">An unique ID for the camera track.</param>
        /// <param name="positions">An array of camera positions, in world space.</param>
        /// <param name="targets">An array of camera targets, in world space.</param>
        public void AddTrack(string id, Vector3[] positions, Vector3[] targets)
        {
            tracks.Add(id, new object[] { positions, targets });
        }

        /// <summary>
        /// Start the camera track
        /// </summary>
        /// <param name="id">The id of the track</param>
        /// <param name="speed">How fast should it move?</param>
        /// <returns>True if it starts succesfully, otherwise false.</returns>
        public bool RunTrack(string id, float speed)
        {
            object[] track;
            if (tracks.TryGetValue(id, out track) == false)
            {
                return false;
            }

            this.positions = (Vector3[])track[0];
            this.targets = (Vector3[])track[1];

            if (positions.Length != targets.Length || positions.Length < 4)
            {
                return false;
            }

            this.inMovement = true;
            this.speed = speed;
            this.t = 0;
            this.index = 0;

            return true;
        }

        private Vector3 Interpolate(Vector3 p1, Vector3 p2, Vector3 p3, float t)
        {
            return Vector3.Lerp(Vector3.Lerp(p1, p2, t), Vector3.Lerp(p2, p3, t), t);
        }

        public void RemoveTrack(string id)
        {
            tracks.Remove(id);
        }

        public void Stop()
        {
            inMovement = false;
        }

        /// <summary>
        /// Check if the camera is running any tracks.
        /// </summary>
        public bool InMovement { get { return inMovement; } }
    }
}
