﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.Xml;
using GameEngine.LevelGeneration;
using Microsoft.Xna.Framework;

namespace GameEngine.Helpers
{
    public delegate void OnCollision(ObjectNode a, ObjectNode b);

    public class CollisionSimulator
    {
        public event OnCollision OnCollision;

        public List<DynamicCircleNode> dynamicNodes { get; set; }
        private LevelTile[,] levelTiles;
        int TILESIZE = 200;
        int HALFSIZE = 100;
        Vector2 topLeftOffset;
        Vector2 topRightOffset;
        Vector2 bottomLeftOffset;
        Vector2 bottomRightOffset;

        public CollisionSimulator(LevelTile[,] levelTiles)
        {
            dynamicNodes = new List<DynamicCircleNode>();
            this.levelTiles = levelTiles;
            SetTileSizes(TilesetManager.CurrentTileset.TileSize);
        }

        public CollisionSimulator(LevelTile[,] levelTiles, int tileSize)
        {
            dynamicNodes = new List<DynamicCircleNode>();
            this.levelTiles = levelTiles;
            SetTileSizes(tileSize);
        }

        private void SetTileSizes(int tileSize)
        {
            TILESIZE = tileSize;
            HALFSIZE = TILESIZE / 2;
            topLeftOffset = new Vector2(-HALFSIZE, -HALFSIZE);
            topRightOffset = new Vector2(HALFSIZE, -HALFSIZE);
            bottomLeftOffset = new Vector2(-HALFSIZE, HALFSIZE);
            bottomRightOffset = new Vector2(HALFSIZE, HALFSIZE);
        }

        List<ObjectNode> shootRayResultList = new List<ObjectNode>();
        public List<float> shootRayResultDistances = new List<float>();

        CollisionRay leftRay = new CollisionRay(Vector2.Zero, Vector2.UnitX);
        CollisionRay rightRay = new CollisionRay(Vector2.Zero, Vector2.UnitX);

        public DynamicCircleNode GetClosestTargetInPie(CollisionRay ray, float angle, float maxDistance, LevelTile tile)
        {

            List<DynamicCircleNode> results = GetTargetsInPie(ray, angle, maxDistance, tile);

            if (results.Count > 0)
            {
                return results[0];
            }
            else
            {
                return null;
            }
        }


        List<DynamicCircleNode> pieTargets = new List<DynamicCircleNode>();
        List<float> pieTargetDistances = new List<float>();

        public List<DynamicCircleNode> GetTargetsInPie(CollisionRay ray, float angle, float maxDistance, LevelTile tile)
        {
            if (tile == null)
                return null;
            Vector2 originalDirection = ray.Direction;
            Vector2 leftNormal, rightNormal;
            ray.Rotate((float)MathHelper.PiOver2 - angle);
            leftNormal = ray.Direction;
            ray.Rotate(2 * angle - (float)Math.PI);
            rightNormal = ray.Direction;
            ray.Direction = originalDirection;

            pieTargets.Clear();
            pieTargetDistances.Clear();

            GetTargetsInPieRecursive(ray, leftNormal, rightNormal, maxDistance, tile);

            for (int i = pieTargets.Count - 1; i >= 0; i--)
            {
                BoundingCircle colCircle = pieTargets[i].GetCollisionCircle();
                float closestWallDist = GetClosestPolygonCollisionDistance(new CollisionRay(ray.Position, ray.Position - colCircle.Center, true), tile);
                if (closestWallDist < pieTargetDistances[i] + colCircle.Radius)
                    pieTargets.Remove(pieTargets[i]);
            }

            return pieTargets;
        }

        private void GetTargetsInPieRecursive(CollisionRay ray, Vector2 leftNormal, Vector2 rightNormal, float maxDistance, LevelTile tile)
        {
            if (tile == null)
                return;

            int startIndex = shootRayResultList.Count;

            for (int i = 0; i < tile.SceneNode.ChildNodes.Count; i++)
            {
                DynamicCircleNode dynNode = tile.SceneNode.ChildNodes[i] as DynamicCircleNode;
                if (dynNode != null)
                {
                    BoundingCircle colCircle = dynNode.GetCollisionCircle();
                    Vector2 startToCenter = colCircle.Center - ray.Position;
                    float rayDist = Vector2.Dot(ray.Direction, startToCenter);
                    if (
                        rayDist > 0 && rayDist < maxDistance &&
                        Vector2.Dot(leftNormal, startToCenter) > 0 &&
                        Vector2.Dot(rightNormal, startToCenter) > 0
                    )
                    {
                        MergeToPieTargets(dynNode, rayDist, startIndex);
                    }
                }
            }

            bool checkLeft, checkRight, checkTop, checkBottom;
            checkLeft = checkRight = checkTop = checkBottom = false;

            float colDist;

            leftRay.Position = rightRay.Position = ray.Position;
            leftRay.Direction = new Vector2(-leftNormal.Y, leftNormal.X);
            rightRay.Direction = new Vector2(rightNormal.Y, -rightNormal.X);

            // Check for collision with the four walls of the tile
            // Left
            if (IntersectsLeft(leftRay, out colDist))
            {
                if (colDist < maxDistance)
                    checkLeft = true;
                if (IntersectsUp(rightRay, out colDist))
                {
                    if (colDist < maxDistance)
                        checkTop = true;
                }
                else
                    if (!checkLeft && IntersectsLeft(rightRay, out colDist) && colDist < maxDistance)
                        checkLeft = true;
            }
            // Up
            else if (IntersectsUp(leftRay, out colDist))
            {
                if (colDist < maxDistance)
                    checkTop = true;
                if (IntersectsRight(rightRay, out colDist))
                {
                    if (colDist < maxDistance)
                        checkRight = true;
                }
                else
                    if (!checkTop && IntersectsUp(rightRay, out colDist) && colDist < maxDistance)
                        checkTop = true;
            }
            // Right
            else if (IntersectsRight(leftRay, out colDist))
            {
                if (colDist < maxDistance)
                    checkRight = true;
                if (IntersectsDown(rightRay, out colDist))
                {
                    if (colDist < maxDistance)
                        checkBottom = true;
                }
                else
                    if (!checkTop && IntersectsRight(rightRay, out colDist) && colDist < maxDistance)
                        checkRight = true;
            }
            // Down
            else if (IntersectsDown(leftRay, out colDist))
            {
                if (colDist < maxDistance)
                    checkBottom = true;
                if (IntersectsLeft(rightRay, out colDist))
                {
                    if (colDist < maxDistance)
                        checkLeft = true;
                }
                else
                    if (!checkTop && IntersectsDown(rightRay, out colDist) && colDist < maxDistance)
                        checkBottom = true;
            }

            if (checkLeft)
                GetTargetsInPieRecursive(new CollisionRay(ray.Position + new Vector2(TILESIZE, 0), ray.Direction), leftNormal, rightNormal, maxDistance, tile.TileLeft);
            if (checkTop)
                GetTargetsInPieRecursive(new CollisionRay(ray.Position + new Vector2(0, TILESIZE), ray.Direction), leftNormal, rightNormal, maxDistance, tile.TileAbove);
            if (checkRight)
                GetTargetsInPieRecursive(new CollisionRay(ray.Position + new Vector2(-TILESIZE, 0), ray.Direction), leftNormal, rightNormal, maxDistance, tile.TileRight);
            if (checkBottom)
                GetTargetsInPieRecursive(new CollisionRay(ray.Position + new Vector2(0, -TILESIZE), ray.Direction), leftNormal, rightNormal, maxDistance, tile.TileBelow);
        }

        private int MergeToPieTargets(DynamicCircleNode node, float distance, int startIndex)
        {
            for (int i = startIndex; i < pieTargets.Count; i++)
            {
                if (distance <= pieTargetDistances[i])
                {
                    pieTargets.Insert(i, node);
                    pieTargetDistances.Insert(i, distance);
                    return i;
                }
            }
            pieTargets.Add(node);
            pieTargetDistances.Add(distance);
            return pieTargets.Count - 1;
        }


        private bool IntersectsLeft(CollisionRay ray, out float colDist)
        {
            if (ray.Position.X - 0.00001f < -HALFSIZE)
            {
                colDist = 0;
                return false;
            }
            return ray.Intersects(topLeftOffset, bottomLeftOffset, out colDist);
        }

        private bool IntersectsUp(CollisionRay ray, out float colDist)
        {
            if (ray.Position.Y - 0.00001f < -HALFSIZE)
            {
                colDist = 0;
                return false;
            }
            return ray.Intersects(topRightOffset, topLeftOffset, out colDist);
        }

        private bool IntersectsRight(CollisionRay ray, out float colDist)
        {
            if (ray.Position.X + 0.00001f > HALFSIZE)
            {
                colDist = 0;
                return false;
            }
            return ray.Intersects(bottomRightOffset, topRightOffset, out colDist);
        }

        private bool IntersectsDown(CollisionRay ray, out float colDist)
        {
            if (ray.Position.Y + 0.00001f > HALFSIZE)
            {
                colDist = 0;
                return false;
            }
            return ray.Intersects(bottomLeftOffset, bottomRightOffset, out colDist);
        }
        

        public List<ObjectNode> ShootRay(CollisionRay ray, LevelTile startTile, out Vector2 endPolyNormal)
        {

            shootRayResultList.Clear();
            shootRayResultDistances.Clear();
            ShootRayRecursive(ray, startTile, TILESIZE * 2, 0f, out endPolyNormal);

            return shootRayResultList;

        }

        private void ShootRayRecursive(CollisionRay ray, LevelTile tile, float maxDist, float offset, out Vector2 normal)
        {
            //float hitDist = 0;
            normal = Vector2.Zero;
            if (tile == null)
                return;

            int startIndex = shootRayResultList.Count;

            // First figure out if we hit a wall...
            PolygonNode endNode = null;
            float colDist;
            Vector2 tmpNormal;
            foreach (var poly in tile.SceneNode.Statics)
            {
                if (poly != null && !poly.ShootThrough && poly.Intersects(ray, out colDist, out tmpNormal))
                {
                    if (colDist < maxDist)
                    {
                        maxDist = colDist + offset;
                        endNode = poly;
                        normal = tmpNormal;
                    }
                }
            }
            
            for (int i = 0; i < tile.SceneNode.DynamicCount; i++)
            {
                DynamicCircleNode dynNode = tile.SceneNode[i];
                if (dynNode.PassThrough)
                    continue;

                if (ray.Intersects(dynNode.GetCollisionCircle(), out colDist))
                {
                    if (colDist < maxDist)
                        MergeNodeToShootRayResult(dynNode, colDist + offset, startIndex);
                }
            }
            if (endNode != null)
            {
                // Insert the wall that we hit and remove any nodes behind it before we return
                int lastIndex = MergeNodeToShootRayResult(endNode, maxDist, startIndex);
                if (lastIndex < shootRayResultList.Count - 1)
                    shootRayResultList.RemoveRange(lastIndex + 1, shootRayResultList.Count - lastIndex - 1);
                return;
            }

            // Check for collision with the four walls of the tile
            // Left
            if (IntersectsLeft(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                ShootRayRecursive(new CollisionRay(soFar + new Vector2(TILESIZE, 0), ray.Direction), tile.TileLeft, maxDist - colDist, colDist, out normal);
            }
            // Up
            if (IntersectsUp(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                ShootRayRecursive(new CollisionRay(soFar + new Vector2(0, TILESIZE), ray.Direction), tile.TileAbove, maxDist - colDist, colDist, out normal);
            }
            // Right
            if (IntersectsRight(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                ShootRayRecursive(new CollisionRay(soFar - new Vector2(TILESIZE, 0), ray.Direction), tile.TileRight, maxDist - colDist, colDist, out normal);
            }
            // Down
            if (IntersectsDown(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                ShootRayRecursive(new CollisionRay(soFar - new Vector2(0, TILESIZE), ray.Direction), tile.TileBelow, maxDist - colDist, colDist, out normal);
            }
            
        }

        private int MergeNodeToShootRayResult(ObjectNode node, float distFromRay, int startIndex)
        {
            for (int i = startIndex; i < shootRayResultList.Count; i++)
			{
                if (distFromRay <= shootRayResultDistances[i])
                {
                    shootRayResultList.Insert(i, node);
                    shootRayResultDistances.Insert(i, distFromRay);
                    return i;
                }
			}
            shootRayResultList.Add(node);
            shootRayResultDistances.Add(distFromRay);
            return shootRayResultList.Count - 1;
        }

        /*
        public Vector3 GetTileStartPosition3D(LevelTileNode tileNode, float radius)
        {
            Vector2 startPos2D = GetTileStartPosition(tileNode, radius);
            return new Vector3(startPos2D.X, 0, startPos2D.Y);
        }

        public Vector2 GetTileStartPosition(LevelTileNode tileNode, float radius)
        {
            Vector2 tmpResolve;

            int spawnSize = TILESIZE - (int)Math.Ceiling(2 * radius);
            int halfSize = spawnSize / 2;

            BoundingCircle testCircle = new BoundingCircle(Vector2.Zero, radius);
            for (int i = 0; i < 10; i++)
            {
                Vector2 testPosition = new Vector2(rand.Next(spawnSize) - halfSize, rand.Next(spawnSize) - halfSize);
                testCircle.Center = testPosition;
                bool useable = true;
                for (int j = 0; j < tileNode.DynamicCount; j++)
                {
                    if (testCircle.Intersects(tileNode[j].GetCollisionCircle(), out tmpResolve))
                    {
                        useable = false;
                        break;
                    }
                }
                for (int j = 0; j < tileNode.Statics.Length; j++)
                {
                    if (tileNode.Statics[j].Intersects(testCircle, out tmpResolve))
                    {
                        useable = false;
                        break;
                    }
                }
                if (useable)
                    return testPosition;
            }

            return Vector2.Zero;
        }
         */

        public void Update(GameTime gameTime, Point[] pointsOfInterest, int worldMaxCollideDist)
        {
            dynamicNodes.Clear();

            for (int i = 0; i < pointsOfInterest.Length; i++)
            {
                Point tileCoords;
                int x = Math.Max(pointsOfInterest[i].X - worldMaxCollideDist, 0);
                int ex = Math.Min(pointsOfInterest[i].X + worldMaxCollideDist, levelTiles.GetLength(0) - 1);

                int y = Math.Max(pointsOfInterest[i].Y - worldMaxCollideDist, 0);
                int ey = Math.Min(pointsOfInterest[i].Y + worldMaxCollideDist, levelTiles.GetLength(1) - 1);

                for (tileCoords.X = x; tileCoords.X <= ex; tileCoords.X++)
                {
                    for (tileCoords.Y = y; tileCoords.Y <= ey; tileCoords.Y++)
                    {
                        LevelTile tile = levelTiles[tileCoords.X, tileCoords.Y];
                        if (tile != null)
                        {
                            LevelTileNode container = tile.SceneNode as LevelTileNode;
                            for (int nodeIndex = 0; nodeIndex < container.DynamicCount; nodeIndex++)
                            {
                                CollideWithDynamics(container[nodeIndex]);
                                if (container[nodeIndex].PassThrough)
                                    continue;
                                dynamicNodes.Add(container[nodeIndex]);
                            }
                        }
                    }
                }
            }


            for (int i = 0; i < dynamicNodes.Count; i++)
            {
                CollideWithWorld(dynamicNodes[i]);
            }
        }

        private bool IsAbove(BoundingCircle circle)
        {
            return circle.Center.Y - circle.Radius < -HALFSIZE;
        }

        private bool IsBelow(BoundingCircle circle)
        {
            return circle.Center.Y + circle.Radius > HALFSIZE;
        }

        private bool IsToTheLeft(BoundingCircle circle)
        {
            return circle.Center.X - circle.Radius < -HALFSIZE;
        }

        private bool IsToTheRight(BoundingCircle circle)
        {
            return circle.Center.X + circle.Radius > HALFSIZE;
        }

        private void CollideWithWorld(DynamicCircleNode dynNode)
        {
            Vector2 resolve;

            BoundingCircle currentCircle = dynNode.GetCollisionCircle();

            // Collide with stuff in the local tile

            LevelTileNode container = dynNode.Parent as LevelTileNode;

            bool resolved = false;
            if (dynNode.LastMovement.LengthSquared() > currentCircle.Radius * currentCircle.Radius)
            {
                float lastMoveLength = dynNode.LastMovement.Length();
                Vector2 lastMoveDirection = dynNode.LastMovement / lastMoveLength;

                Vector2 startPoint = currentCircle.Center - dynNode.LastMovement;

                int stepCount = (int)(lastMoveLength / currentCircle.Radius);
                for (int i = 1; i <= stepCount; i++)
                {
                    BoundingCircle stepCircle = new BoundingCircle(startPoint + i * lastMoveDirection, currentCircle.Radius);
                    for (int j = 0; j < container.Statics.Length; j++)
                    {
                        PolygonNode other = container.Statics[j];
                        Vector2 tmpResolve;
                        if (other.Intersects(stepCircle, dynNode.LastMovement, out tmpResolve))
                        {
                            stepCircle.Center += tmpResolve;
                            OnCollision(dynNode, other);
                            resolved = true;
                        }
                    }
                    if (resolved)
                    {
                        currentCircle = dynNode.ResolveWallCollision(stepCircle.Center - currentCircle.Center);
                        break;
                    }
                }
            }
            if (!resolved)
            {
                for (int i = 0; i < container.Statics.Length; i++)
                {
                    PolygonNode other = container.Statics[i];
                    if (other.Intersects(currentCircle, dynNode.LastMovement, out resolve))
                    {
                        currentCircle = dynNode.ResolveWallCollision(resolve);
                        OnCollision(dynNode, other);
                    }
                }
            }

            LevelTile containingTile = (dynNode.Parent as LevelTileNode).Tile;

            // Collide with stuff in closeby tiles
            if (IsToTheLeft(currentCircle))
            {
                LevelTile leftTile = containingTile.TileLeft;

                // Collide with left tile
                CollideWithWorldInTile(leftTile, dynNode, new Vector2(TILESIZE, 0));

                if (IsAbove(currentCircle))
                {
                    // Collide with topleft tile
                    if (leftTile != null)
                        CollideWithWorldInTile(leftTile.TileAbove, dynNode, new Vector2(TILESIZE, TILESIZE));
                    // Collide with top tile
                    CollideWithWorldInTile(containingTile.TileAbove, dynNode, new Vector2(0, TILESIZE));
                }
                else if (IsBelow(currentCircle))
                {
                    // Check bottomleft tile
                    if (leftTile != null)
                        CollideWithWorldInTile(leftTile.TileBelow, dynNode, new Vector2(TILESIZE, -TILESIZE));
                    // Check bottom tile
                    CollideWithWorldInTile(containingTile.TileBelow, dynNode, new Vector2(0, -TILESIZE));
                }
            }
            else if (IsToTheRight(currentCircle))
            {
                LevelTile rightTile = containingTile.TileRight;

                // Check right tile
                CollideWithWorldInTile(rightTile, dynNode, new Vector2(-TILESIZE, 0));

                if (IsAbove(currentCircle))
                {
                    // Check top tile
                    CollideWithWorldInTile(containingTile.TileAbove, dynNode, new Vector2(0, TILESIZE));
                    // Check topright tile
                    if (rightTile != null)
                        CollideWithWorldInTile(rightTile.TileAbove, dynNode, new Vector2(-TILESIZE, TILESIZE));

                }
                else if (IsBelow(currentCircle))
                {
                    // Check bottom tile
                    CollideWithWorldInTile(containingTile.TileBelow, dynNode, new Vector2(0, -TILESIZE));
                    // Check bottomright tile
                    if (rightTile != null)
                        CollideWithWorldInTile(rightTile.TileBelow, dynNode, new Vector2(-TILESIZE, -TILESIZE));
                }
            }
            else
            {
                if (IsAbove(currentCircle))
                {
                    // Check top tile
                    CollideWithWorldInTile(containingTile.TileAbove, dynNode, new Vector2(0, TILESIZE));
                }
                else if (IsBelow(currentCircle))
                {
                    // Check bottom tile
                    CollideWithWorldInTile(containingTile.TileBelow, dynNode, new Vector2(0, -TILESIZE));
                }
            }
            dynNode.CommitCollision();
        }

        private void CollideWithWorldInTile(LevelTile tile, DynamicCircleNode dynNode, Vector2 circleOffset)
        {
            if (tile == null)
                return;
            LevelTileNode container = tile.SceneNode as LevelTileNode;

            if (container.IsEmpty)
                return;

            Vector2 resolve;
            BoundingCircle currentCircle = dynNode.GetCollisionCircle();
            currentCircle.Center += circleOffset;

            for (int i = 0; i < container.Statics.Length; i++)
            {
                if (container.Statics[i].Intersects(currentCircle, dynNode.LastMovement, out resolve))
                {
                    currentCircle = dynNode.ResolveWallCollision(resolve);
                    currentCircle.Center += circleOffset;
                    OnCollision(dynNode, container.Statics[i]);
                }
            }

        }

        private void CollideWithDynamics(DynamicCircleNode dynNode)
        {
            Vector2 resolve;
            BoundingCircle currentCircle = dynNode.GetCollisionCircle();

            LevelTileNode container = dynNode.Parent as LevelTileNode;
            LevelTile containingTile = container.Tile;

            // Collide with stuff in local tile
            for (int i = 0; i < container.DynamicCount; i++)
            {
                DynamicCircleNode other = container[i];
                // Don't check self
                if (other == dynNode)
                    continue;

                if (other.GetCollisionCircle().Intersects(currentCircle, out resolve))
                {
                    if (! (dynNode.PassThrough || other.PassThrough))
                    {
                        currentCircle = dynNode.ResolveCircleCollision(resolve * 0.5f);
                        other.ResolveCircleCollision(resolve * -0.5f);
                    }
                    OnCollision(dynNode, other);
                }
            }
            
            // Check collision with the tiles to the left of us
            if (IsToTheLeft(currentCircle))
            {
                LevelTile leftTile = containingTile.TileLeft;

                // Collide with left tile
                CollideWithDynamicsInTile(dynNode, leftTile, new Vector2(TILESIZE, 0));

                if (IsAbove(currentCircle))
                {
                    // Collide with topleft tile
                    if (leftTile != null)
                        CollideWithDynamicsInTile(dynNode, leftTile.TileAbove, new Vector2(TILESIZE, TILESIZE));
                    // Collide with top tile
                    CollideWithDynamicsInTile(dynNode, containingTile.TileAbove, new Vector2(0, TILESIZE));
                }
                else if (IsBelow(currentCircle))
                {
                    // Check bottomleft tile
                    if (leftTile != null)
                        CollideWithDynamicsInTile(dynNode, leftTile.TileBelow, new Vector2(TILESIZE, -TILESIZE));
                    // Check bottom tile
                    CollideWithDynamicsInTile(dynNode, containingTile.TileBelow, new Vector2(0, -TILESIZE));
                }
            }
            else if (IsToTheRight(currentCircle))
            {
                LevelTile rightTile = containingTile.TileRight;

                // Check right tile
                CollideWithDynamicsInTile(dynNode, rightTile, new Vector2(-TILESIZE, 0));

                if (IsAbove(currentCircle))
                {
                    // Check top tile
                    CollideWithDynamicsInTile(dynNode, containingTile.TileAbove, new Vector2(0, TILESIZE));
                    // Check topright tile
                    if (rightTile != null)
                        CollideWithDynamicsInTile(dynNode, rightTile.TileAbove, new Vector2(-TILESIZE, TILESIZE));

                }
                else if (IsBelow(currentCircle))
                {
                    // Check bottom tile
                    CollideWithDynamicsInTile(dynNode, containingTile.TileBelow, new Vector2(0, -TILESIZE));
                    // Check bottomright tile
                    if (rightTile != null)
                        CollideWithDynamicsInTile(dynNode, rightTile.TileBelow, new Vector2(-TILESIZE, -TILESIZE));
                }
            }
            else
            {
                if (IsAbove(currentCircle))
                {
                    // Check top tile
                    CollideWithDynamicsInTile(dynNode, containingTile.TileAbove, new Vector2(0, TILESIZE));
                }
                else if (IsAbove(currentCircle))
                {
                    // Check bottom tile
                    CollideWithDynamicsInTile(dynNode, containingTile.TileBelow, new Vector2(0, -TILESIZE));
                }
            }

            dynNode.CommitCollision();
        }

        public void CollideWithDynamicsInTile(DynamicCircleNode dynNode, LevelTile tile, Vector2 circleOffset)
        {
            if (tile == null)
                return;
            if ((tile.SceneNode as LevelTileNode).IsEmpty)
                return;

            Vector2 resolve;
            BoundingCircle currentCircle = dynNode.GetCollisionCircle();
            currentCircle.Center += circleOffset;

            LevelTileNode tileNode = tile.SceneNode as LevelTileNode;

            // Collide with stuff in local tile
            for (int i = 0; i < tileNode.DynamicCount; i++)
            {
                DynamicCircleNode other = tileNode[i];
                if (other.GetCollisionCircle().Intersects(currentCircle, out resolve))
                {
                    if (!(dynNode.PassThrough || other.PassThrough))
                    {
                        currentCircle = dynNode.ResolveCircleCollision(resolve * 0.5f);
                        currentCircle.Center += circleOffset;
                        other.ResolveCircleCollision(resolve * -0.5f);
                    }
                    OnCollision(dynNode, other);
                }
            }
        }

        public Vector2 GetClosestPolygonCollision(CollisionRay ray, LevelTile tile)
        {
            float dist = GetClosestPolygonCollisionDistance(ray, tile);
            return ray.Position + ray.Direction * dist;
        }

        public float GetClosestPolygonCollisionDistance(CollisionRay ray, LevelTile tile)
        {
            return GetClosestPolygonCollisionDistance(ray, tile, true);
        }

        public float GetClosestPolygonCollisionDistance(CollisionRay ray, LevelTile tile, bool checkNeighbours)
        {
            if (tile == null)
                return 0;

            float closestDist = TILESIZE * TILESIZE;
            float colDist;
            for (int i = 0; i < tile.SceneNode.Statics.Length; i++)
            {
                PolygonNode poly = tile.SceneNode.Statics[i] as PolygonNode;
                if (poly != null && poly.Intersects(ray, out colDist))
                {
                    if (colDist < closestDist)
                        closestDist = colDist;
                }
            }
            if (closestDist < TILESIZE * TILESIZE)
                return closestDist;

            if (!checkNeighbours)
                return 0;

            // Check for collision with the four walls of the tile
            // Left
            if (IntersectsLeft(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                return colDist + GetClosestPolygonCollisionDistance(new CollisionRay(soFar + new Vector2(TILESIZE, 0), ray.Direction), tile.TileLeft);
            }
            // Up
            if (IntersectsUp(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                return colDist + GetClosestPolygonCollisionDistance(new CollisionRay(soFar + new Vector2(0, TILESIZE), ray.Direction), tile.TileAbove);
            }
            // Right
            if (IntersectsRight(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                return colDist + GetClosestPolygonCollisionDistance(new CollisionRay(soFar - new Vector2(TILESIZE, 0), ray.Direction), tile.TileRight);
            }
            // Down
            if (IntersectsDown(ray, out colDist))
            {
                Vector2 soFar = ray.Position + ray.Direction * colDist;
                return colDist + GetClosestPolygonCollisionDistance(new CollisionRay(soFar - new Vector2(0, TILESIZE), ray.Direction), tile.TileBelow);
            }

            // We hould never get here:
            return 0;
        }
    }
}
