using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using GameEngine.Level;

namespace GameEngine.Helpers
{
    public class OrientedBoundingBox
    {
        public static readonly float Epsilon = 1.0e-6f;

        public Vector3 Center = new Vector3(0f);

        private Vector3 extent = Vector3.One;

        public Vector3 Extent
        {
            get { return extent; }
            set {
                extent.X = Math.Abs(value.X);
                extent.Y = Math.Abs(value.Y);
                extent.Z = Math.Abs(value.Z);
            }
        }

        private Matrix rotation = Matrix.Identity;

        public Matrix Rotation
        {
            get { return rotation; }
            set { rotation = value; }
        }

        private bool selected;

        public bool Selected
        {
            get { return selected; }
            set { selected = value; }
        }

        public Vector3 NearTopRight
        {
            get { return Center + Vector3.Transform(extent, Rotation); }
        }

        public Vector3 FarTopRight
        {
            get { return Center + Vector3.Transform(new Vector3(extent.X, extent.Y, -extent.Z), Rotation); }
        }

        public Vector3 NearBottomRight
        {
            get { return Center + Vector3.Transform(new Vector3(extent.X, -extent.Y, extent.Z), Rotation); }
        }

        public Vector3 FarBottomRight
        {
            get { return Center + Vector3.Transform(new Vector3(extent.X, -extent.Y, -extent.Z), Rotation); }
        }

        public Vector3 NearTopLeft
        {
            get { return Center + Vector3.Transform(new Vector3(-extent.X, extent.Y, extent.Z), Rotation); }
        }

        public Vector3 FarTopLeft
        {
            get { return Center + Vector3.Transform(new Vector3(-extent.X, extent.Y, -extent.Z), Rotation); }
        }

        public Vector3 NearBottomLeft
        {
            get { return Center + Vector3.Transform(new Vector3(-extent.X, -extent.Y, extent.Z), Rotation); }
        }

        public Vector3 FarBottomLeft
        {
            get { return Center + Vector3.Transform(new Vector3(-extent.X, -extent.Y, -extent.Z), Rotation); }
        }

        public OrientedBoundingBox()
        {
        }

        public OrientedBoundingBox(Vector3 center, Vector3 extent, Matrix rotation)
        {
            this.Center = center;
            this.extent = extent;
            this.rotation = rotation;
        }

        public bool Intersects(OrientedBoundingBox other)
        {
            Vector3 a = extent;
            Vector3 b = other.Extent;

            float aDist, bDist, cDist;

            bool parallelAxes = false;

            Matrix relativeRotation = rotation * Matrix.Transpose(other.Rotation);

            Matrix relativeAbsoluteRotation;
            relativeAbsoluteRotation.M11 = Math.Abs(relativeRotation.M11);
            relativeAbsoluteRotation.M12 = Math.Abs(relativeRotation.M12);
            relativeAbsoluteRotation.M13 = Math.Abs(relativeRotation.M13);
            relativeAbsoluteRotation.M21 = Math.Abs(relativeRotation.M21);
            relativeAbsoluteRotation.M22 = Math.Abs(relativeRotation.M22);
            relativeAbsoluteRotation.M23 = Math.Abs(relativeRotation.M23);
            relativeAbsoluteRotation.M31 = Math.Abs(relativeRotation.M31);
            relativeAbsoluteRotation.M32 = Math.Abs(relativeRotation.M32);
            relativeAbsoluteRotation.M33 = Math.Abs(relativeRotation.M33);

            // Check if any of the axes of the two boxes are paralel
            if (relativeAbsoluteRotation.M11 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M12 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M13 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M21 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M22 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M23 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M31 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M32 + Epsilon >= 1.0f) parallelAxes = true;
            if (relativeAbsoluteRotation.M33 + Epsilon >= 1.0f) parallelAxes = true;

            // relative translation
            Vector3 c = Vector3.Transform(other.Center - Center, Rotation);

            // Axis separation along A's first axis
            cDist = Math.Abs(c.X);
            aDist = a.X;
            bDist = b.X * relativeAbsoluteRotation.M11 + b.Y * relativeAbsoluteRotation.M12 + b.Z * relativeAbsoluteRotation.M13;
            if (cDist > aDist + bDist)
                return false;

            // Axis separation along A's second axis
            cDist = Math.Abs(c.Y);
            aDist = a.Y;
            bDist = b.X * relativeAbsoluteRotation.M21 + b.Y * relativeAbsoluteRotation.M22 + b.Z * relativeAbsoluteRotation.M23;
            if (cDist > aDist + bDist)
                return false;

            // Axis separation along A's third axis
            cDist = Math.Abs(c.Z);
            aDist = a.Z;
            bDist = b.X * relativeAbsoluteRotation.M31 + b.Y * relativeAbsoluteRotation.M32 + b.Z * relativeAbsoluteRotation.M33;
            if (cDist > aDist + bDist)
                return false;

            // Axis separation along B's first axis
            cDist = Math.Abs(c.X * relativeRotation.M11 + c.Y * relativeRotation.M21 + c.Z * relativeRotation.M31);
            aDist = a.X * relativeAbsoluteRotation.M11 + a.Y * relativeAbsoluteRotation.M21 + a.Z * relativeAbsoluteRotation.M31;
            bDist = b.X;
            if (cDist > aDist + bDist)
                return false;

            // Axis separation along B's second axis
            cDist = Math.Abs(c.X * relativeRotation.M12 + c.Y * relativeRotation.M22 + c.Z * relativeRotation.M32);
            aDist = a.X * relativeAbsoluteRotation.M12 + a.Y * relativeAbsoluteRotation.M22 + a.Z * relativeAbsoluteRotation.M32;
            bDist = b.Y;
            if (cDist > aDist + bDist)
                return false;

            // Axis separation along B's third axis
            cDist = Math.Abs(c.X * relativeRotation.M13 + c.Y * relativeRotation.M23 + c.Z * relativeRotation.M33);
            aDist = a.X * relativeAbsoluteRotation.M13 + a.Y * relativeAbsoluteRotation.M23 + a.Z * relativeAbsoluteRotation.M33;
            bDist = b.Z;
            if (cDist > aDist + bDist)
                return false;

            // if the two boxes have parallel axes, we're done, intersection
            if ( parallelAxes )
                return true;

            // separating axis A0 x B0
            cDist = Math.Abs(c.Z * relativeRotation.M21 - c.Y * relativeRotation.M31);
            aDist = a.Y * relativeAbsoluteRotation.M31 + a.Z * relativeAbsoluteRotation.M21;
            bDist = b.Y * relativeAbsoluteRotation.M13 + b.Z * relativeAbsoluteRotation.M12;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A0 x B1
            cDist = Math.Abs(c.Z * relativeRotation.M22 - c.Y * relativeRotation.M32);
            aDist = a.Y * relativeAbsoluteRotation.M32 + a.Z * relativeAbsoluteRotation.M22;
            bDist = b.X * relativeAbsoluteRotation.M13 + b.Z * relativeAbsoluteRotation.M11;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A0 x B2
            cDist = Math.Abs(c.Z * relativeRotation.M23 - c.Y * relativeRotation.M33);
            aDist = a.Y * relativeAbsoluteRotation.M33 + a.Z * relativeAbsoluteRotation.M23;
            bDist = b.X * relativeAbsoluteRotation.M12 + b.Y * relativeAbsoluteRotation.M11;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A1 x B0
            cDist = Math.Abs(c.X * relativeRotation.M31 - c.Z * relativeRotation.M11);
            aDist = a.X * relativeAbsoluteRotation.M31 + a.Z * relativeAbsoluteRotation.M11;
            bDist = b.Y * relativeAbsoluteRotation.M23 + b.Z * relativeAbsoluteRotation.M22;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A1 x B1
            cDist = Math.Abs(c.X * relativeRotation.M32 - c.Z * relativeRotation.M12);
            aDist = a.X * relativeAbsoluteRotation.M32 + a.Z * relativeAbsoluteRotation.M12;
            bDist = b.X * relativeAbsoluteRotation.M23 + b.Z * relativeAbsoluteRotation.M21;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A1 x B2
            cDist = Math.Abs(c.X * relativeRotation.M33 - c.Z * relativeRotation.M13);
            aDist = a.X * relativeAbsoluteRotation.M33 + a.Z * relativeAbsoluteRotation.M13;
            bDist = b.X * relativeAbsoluteRotation.M22 + b.Y * relativeAbsoluteRotation.M21;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A2 x B0
            cDist = Math.Abs(c.Y * relativeRotation.M11 - c.X * relativeRotation.M21);
            aDist = a.X * relativeAbsoluteRotation.M21 + a.Y * relativeAbsoluteRotation.M11;
            bDist = b.Y * relativeAbsoluteRotation.M33 + b.Z * relativeAbsoluteRotation.M32;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A2 x B1
            cDist = Math.Abs(c.Y * relativeRotation.M12 - c.X * relativeRotation.M22);
            aDist = a.X * relativeAbsoluteRotation.M22 + a.Y * relativeAbsoluteRotation.M12;
            bDist = b.X * relativeAbsoluteRotation.M33 + b.Z * relativeAbsoluteRotation.M31;
            if ( cDist > aDist + bDist )
                return false;

            // separating axis A2 x B2
            cDist = Math.Abs(c.Y * relativeRotation.M13 - c.X * relativeRotation.M23);
            aDist = a.X * relativeAbsoluteRotation.M23 + a.Y * relativeAbsoluteRotation.M13;
            bDist = b.X * relativeAbsoluteRotation.M32 + b.Y * relativeAbsoluteRotation.M31;
            if ( cDist > aDist + bDist )
                return false;

            // all tests failed, have intersection
            return true;

        }

        public bool Intersects(Ray ray)
        {
            float min, max;
            return Intersects(ray, out min, out max);
        }

        public bool Intersects(Ray ray, out float minT, out float maxS)
        {
            maxS = -float.MaxValue;
            minT = float.MaxValue;

            // compute difference vector
            Vector3 diff = Center - ray.Position;

            // for each axis do
            for (int i = 0; i < 3; ++i)
            {
                // get axis i
                Vector3 rotAxis;
                float extendAxis;
                switch (i)
                {
                    case 0:
                    default:
                        extendAxis = extent.X;
                        rotAxis = rotation.Right;
                        break;
                    case 1:
                        extendAxis = extent.Y;
                        rotAxis = rotation.Up;
                        break;
                    case 2:
                        extendAxis = extent.Z;
                        rotAxis = rotation.Backward;
                        break;
                }

                // project relative vector onto axis
                float e = Vector3.Dot(rotAxis, diff);
                float f = Vector3.Dot(ray.Direction, rotAxis);


                
                // ray is parallel to plane
                if ( Math.Abs(f) < Epsilon )
                {

                    // ray passes by box
                    if ( -e - extendAxis > 0.0f || -e + extendAxis > 0.0f )
                        return false;
                    continue;
                }

                float s = (e - extendAxis) / f;
                float t = (e + extendAxis) / f;

                // fix order
                if ( s > t )
                {
                    float temp = s;
                    s = t;
                    t = temp;
                }

                // adjust min and max values
                if ( s > maxS )
                    maxS = s;
                if ( t < minT )
                    minT = t;

                // check for intersection failure
                if ( minT < 0.0f || maxS > minT )
                    return false;
            }

            // done, have intersection
            return true;
        }

        public float Intersects(Plane plane)
        {
            Vector3 xNormal = Vector3.Transform(plane.Normal, Matrix.Transpose(Rotation));

            // maximum extent in direction of plane normal 
            float r = Math.Abs(extent.X * xNormal.X) 
                    + Math.Abs(extent.Y * xNormal.Y) 
                    + Math.Abs(extent.Z * xNormal.Z);

            // signed distance between box center and plane
            float d = Vector3.Dot(Vector3.Normalize(plane.Normal), Center) + plane.D;

            // return signed distance
            if (Math.Abs(d) < r)
                return 0.0f;
            else if (d < 0.0f)
                return d + r;
            else
                return d - r;
        }

        public Vector3 MaxAAExtent()
        {
            Matrix absRotation = Matrix.Transpose(rotation);
            absRotation.M11 = Math.Abs(absRotation.M11);
            absRotation.M12 = Math.Abs(absRotation.M12);
            absRotation.M13 = Math.Abs(absRotation.M13);
            absRotation.M21 = Math.Abs(absRotation.M21);
            absRotation.M22 = Math.Abs(absRotation.M22);
            absRotation.M23 = Math.Abs(absRotation.M23);
            absRotation.M31 = Math.Abs(absRotation.M31);
            absRotation.M32 = Math.Abs(absRotation.M32);
            absRotation.M33 = Math.Abs(absRotation.M33);


            Vector3 maxProj = new Vector3();
            maxProj.X = Vector3.Dot(extent, absRotation.Right);
            maxProj.Y = Vector3.Dot(extent, absRotation.Up);
            maxProj.Z = Vector3.Dot(extent, absRotation.Backward);

            return maxProj;
        }

    }
}