﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.Xml;

namespace GameEngine.LevelGeneration
{
    public enum Rotation
    {
        Up = 0,
        Right = 90,
        Down = 180,
        Left = 270
    }

    public enum EdgeState
    {
        None,
        Edge,
        Blocked
    }

    [Flags]
    public enum TileType : uint
    {
        // #2#
        // 1 4
        // #8#

        // ###
        // - #
        // ###
        End = 1,
        
        // ###
        // - -
        // ###
        Hallway = 5,
        
        // ###
        // - #
        // #|#
        Corner = 9,
        
        // ###
        // - -
        // #|#
        Tintersection = 13,
        
        // #|#
        // - -
        // #|#
        Crossing = 15
    }

    public class LevelTile
    {
        public SpecialTileNodeType SpecialType { get; set; }

        private LevelTileNode sceneNode;
        public LevelTileNode SceneNode { get { return this.sceneNode; } set { this.sceneNode = value; } }
        private Rotation rotation;
        public Rotation Rotation { get { return this.rotation; } }
        private TileType tileType;
        public TileType TileType { get { return this.tileType; } }

        private EdgeState topEdge;
        public EdgeState TopEdge { get { return this.topEdge; } set { this.topEdge = value; UpdateTypeAndRotation(); } }
        private EdgeState rightEdge;
        public EdgeState RightEdge { get { return this.rightEdge; } set { this.rightEdge = value; UpdateTypeAndRotation(); } }
        private EdgeState bottomEdge;
        public EdgeState BottomEdge { get { return this.bottomEdge; } set { this.bottomEdge = value; UpdateTypeAndRotation(); } }
        private EdgeState leftEdge;
        public EdgeState LeftEdge { get { return this.leftEdge; } set { this.leftEdge = value; UpdateTypeAndRotation(); } }

        private LevelTile tileAbove;
        public LevelTile TileAbove { get { return this.tileAbove; } set { this.tileAbove = value; } }
        private LevelTile tileRight;
        public LevelTile TileRight { get { return this.tileRight; } set { this.tileRight = value; } }
        private LevelTile tileBelow;
        public LevelTile TileBelow { get { return this.tileBelow; } set { this.tileBelow = value; } }
        private LevelTile tileLeft;
        public LevelTile TileLeft { get { return this.tileLeft; } set { this.tileLeft = value; } }

        private bool isPointOfInterest = false;
        public bool IsPointOfInterest { get { return this.isPointOfInterest; } set { this.isPointOfInterest = value; } }
        public int X { get; private set; }
        public int Y { get; private set; }


        public LevelTile(int x, int y)
        {
            this.rotation = Rotation.Up;
            this.topEdge = EdgeState.None;
            this.rightEdge = EdgeState.None;
            this.bottomEdge = EdgeState.None;
            this.leftEdge = EdgeState.None;
            this.X = x;
            this.Y = y;
            SpecialType = SpecialTileNodeType.None;
        }

        public void UpdateTypeAndRotation()
        {
            int numberOfEdges = 0;
            numberOfEdges += (topEdge == EdgeState.Edge) ? 1 : 0;
            numberOfEdges += (rightEdge == EdgeState.Edge) ? 1 : 0;
            numberOfEdges += (bottomEdge == EdgeState.Edge) ? 1 : 0;
            numberOfEdges += (leftEdge == EdgeState.Edge) ? 1 : 0;

            switch (numberOfEdges)
            {
                case 4:
                    this.tileType = TileType.Crossing;
                    // A crossing tile can be rotated in any direction and still have the edges connected
                    this.rotation = (Rotation)Enum.Parse(typeof(Rotation), (Util.Random.Next(4) * 90).ToString());
                    break;
                case 3:
                    this.tileType = TileType.Tintersection;
                    // T-intersections can only be rotated in one direction...
                    if (this.topEdge == EdgeState.Edge && this.rightEdge == EdgeState.Edge && this.bottomEdge == EdgeState.Edge)
                        this.rotation = Rotation.Left;
                    else if (this.rightEdge == EdgeState.Edge && this.bottomEdge == EdgeState.Edge && this.leftEdge == EdgeState.Edge)
                        this.rotation = Rotation.Up;
                    else if (this.bottomEdge == EdgeState.Edge && this.leftEdge == EdgeState.Edge && this.topEdge == EdgeState.Edge)
                        this.rotation = Rotation.Right;
                    else
                        this.rotation = Rotation.Down;
                    break;
                case 2:
                    if (this.topEdge == this.bottomEdge)
                    {
                        this.tileType = TileType.Hallway;
                        // A hallway can be rotated 180 degrees, depending on its orientation
                        if (this.topEdge == EdgeState.Edge)
                            this.rotation = (Util.Random.Next(2) == 1) ? Rotation.Right : Rotation.Left;
                        else
                            this.rotation = (Util.Random.Next(2) == 1) ? Rotation.Up : Rotation.Down;
                    }
                    else
                    {
                        this.tileType = TileType.Corner;
                        // A corner can only be rotated in one direction
                        if (this.topEdge == EdgeState.Edge && this.rightEdge == EdgeState.Edge)
                            this.rotation = Rotation.Down;
                        else if (this.rightEdge == EdgeState.Edge && this.bottomEdge == EdgeState.Edge)
                            this.rotation = Rotation.Left;
                        else if (this.bottomEdge == EdgeState.Edge && this.leftEdge == EdgeState.Edge)
                            this.rotation = Rotation.Up;
                        else
                            this.rotation = Rotation.Right;
                    }
                    break;
                case 1:
                    this.tileType = TileType.End;
                    // An end can only be rotated in one direction
                    if (this.topEdge == EdgeState.Edge)
                        this.rotation = Rotation.Right;
                    else if (this.rightEdge == EdgeState.Edge)
                        this.rotation = Rotation.Down;
                    else if (this.bottomEdge == EdgeState.Edge)
                        this.rotation = Rotation.Left;
                    else
                        this.rotation = Rotation.Up;
                    break;
                default:
                    // This means we don't actually have any edges at all, this should never happen, really
                    break;
            }
        }

        public void RandomizeTile(Dictionary<LevelTileNode, TileType> availableNodes)
        {
            List<LevelTileNode> appropriateNodes = new List<LevelTileNode>();
            foreach (KeyValuePair<LevelTileNode, TileType> item in availableNodes)
                if (item.Value == this.TileType)
                    appropriateNodes.Add(item.Key);
            if(appropriateNodes.Count > 0)
                this.sceneNode = appropriateNodes[Util.Random.Next(appropriateNodes.Count)];
        }
    }
}
