﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;

namespace GameEngine.Physics
{
    public class VerletSystem
    {
        public Dictionary<string, VerletParticle> Verlets { get; private set; }
        public List<VerletConstraint> Constraints { get; set; }
        
        private Vector3 gravity;
        private float timeStep, min, max;

        private const float GRAVITY_THRESHOLD = 1f;
        private const int ITERATION_COUNT = 5;

        public VerletSystem(float gravityAmount, float min, float max)
        {
            Verlets = new Dictionary<string, VerletParticle>();
            Constraints = new List<VerletConstraint>();

            this.gravity = new Vector3(0, -gravityAmount, 0);
            this.min = min;
            this.max = max;
        }

        public void AddParticle(string id, Vector3 position)
        {
            Verlets.Add(id, new VerletParticle(id, position));
        }
        public void AddParticle(string id, float x, float y, float z)
        {
            Verlets.Add(id, new VerletParticle(id, new Vector3(x, y, z)));
        }
        public void AddConstraint(string a, string b, float restLength, bool visible)
        {
            Constraints.Add(new VerletConstraint(a, b, restLength, visible));
        }

        public void Update(float timeStep)
        {
            this.timeStep = timeStep;

            AccumulateForces();
            Verlet();

            for (int i = 0; i < ITERATION_COUNT; i++)
            {
                SatisfyConstraints();
            }
        }

        public void Verlet()
        {
            foreach (VerletParticle p in Verlets.Values)
            {
                Vector3 temp = p.Position;
                p.Position += p.Position - p.OldPosition + p.Force * timeStep * timeStep;
                p.OldPosition = temp;
            }
        }

        public void ApplyForce(string id, Vector3 force)
        {
            Verlets[id].Force += force;
        }

        private void AccumulateForces()
        {
            foreach (VerletParticle p in Verlets.Values)
            {
                Vector3 delta = p.Position - p.OldPosition;

                if (p.Position.Y < GRAVITY_THRESHOLD)
                {
                    p.Force = gravity;
                    p.OldPosition = p.Position;
                }
                else
                {
                    p.Force += gravity;
                }
            }
        }

        private void SatisfyConstraints()
        {

            foreach (VerletParticle p in Verlets.Values)
            {
                p.Position.X = MathHelper.Clamp(p.Position.X, min, max);
                p.Position.Y = MathHelper.Clamp(p.Position.Y, min, max);
                p.Position.Z = MathHelper.Clamp(p.Position.Z, min, max);
            }

            for (int i = 0; i < Constraints.Count; i++)
            {
                Vector3 delta = Verlets[Constraints[i].B].Position - Verlets[Constraints[i].A].Position;
                float diff = (delta.Length() - Constraints[i].RestLength) / delta.Length();
                Verlets[Constraints[i].A].Position += delta * 0.5f * diff;
                Verlets[Constraints[i].B].Position -= delta * 0.5f * diff;
            }
        }

        /// <summary>
        /// Apply external force to the verlet system. The force vector is randomized.
        /// </summary>
        /// <param name="fromDirection">The direction of the force. Will be normalized.</param>
        /// <param name="forceMin">The minimum force value of the random force.</param>
        /// <param name="forceMax">The maximum force value of the random force.</param>
        /// <param name="targetPoints">Number of verlet particles to hit. Not randomized.</param>
        public void HitIt(Vector3 fromDirection, int forceMin, int forceMax, int targetPoints)
        {
            Vector3.Normalize(fromDirection);
            VerletParticle[] particles = Verlets.Values.ToArray();
            VerletParticle p;
            int force;

            for (int i = 0; i < targetPoints; i++)
            {
                p = particles[Util.Random.Next(particles.Length)];
                force = Util.Random.Next(forceMin, forceMax + 1);
                p.Force += fromDirection * force;
            }
        }
    }
}
