﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework;
using GameEngine.Xml;
using Microsoft.Xna.Framework.Graphics;
using NPack;

namespace GameEngine
{
    public static class Util
    {
        public static Random Random = new MersenneTwister();

        public static void NullDispose<T>(ref T t) where T : IDisposable
        {
            if (t != null)
            {
                t.Dispose();
                t = default(T);
            }
        }

        public static void Load<T>(this ContentManager content, string asset, out T resource)
        {
            resource = content.Load<T>(asset);
        }

        public static bool TryLoad<T>(this ContentManager content, string asset, out T resource)
        {
            try
            {
                resource = content.Load<T>(asset);
            }
            catch 
            {
                resource = default(T);
                return false;
            }
            return true;
        }

        public static Vector2 GetPerpendicular(this Vector2 v)
        {
            return new Vector2(-v.Y, v.X);
        }

        public static Vector2 GetNormalizedPerpendicular(this Vector2 v)
        {
            return Vector2.Normalize(new Vector2(-v.Y, v.X));
        }
        
        public static void UpdateRecursive(this NodeList nl, GameTime dt, DrawInfo di)
        {
            for (int i = nl.Count - 1; i >= 0; --i)
            {
                SceneNode node = nl[i];
                INodeUpdateable update = node as INodeUpdateable;
                if (update != null)
                    update.Update(dt, di);

                node.ChildNodes.UpdateRecursive(dt, di);
            }
        }

        public static void DrawRecursive(this NodeList nl, GameTime dt, DrawInfo drawInfo)
        {
            for (int i = 0; i < nl.Count; i++)
            {
                INodeRenderable update = nl[i] as INodeRenderable;
                if (update != null)
                    update.Draw(dt, drawInfo);

                nl[i].ChildNodes.DrawRecursive(dt, drawInfo);
            }

        }

        public static void LoadRecursive(this NodeList nl, ContentManager content, GraphicsDevice device )
        {
            for (int i = 0; i < nl.Count; i++)
            {
                INodeLoadable update = nl[i] as INodeLoadable;
                if (update != null)
                    update.LoadContent(content, device);

                nl[i].ChildNodes.LoadRecursive(content, device);
            }

        }

        public static void UpdateBoundingBoxes(this NodeList nl)
        {
            for (int i = 0; i < nl.Count; i++)
            {
                ObjectNode update = nl[i] as ObjectNode;
                if (update != null)
                    update.UpdateBoundingBox();

                nl[i].ChildNodes.UpdateBoundingBoxes();
            }

        }

        public static void UnloadRecursive(this NodeList nl)
        {
            for (int i = 0; i < nl.Count; i++)
            {
                INodeLoadable update = nl[i] as INodeLoadable;
                if (update != null)
                    update.UnloadContent();

                nl[i].ChildNodes.UnloadRecursive();
            }

        }

        public static float ElapsedSeconds(this GameTime dt)
        {
            return (float)dt.ElapsedGameTime.TotalSeconds;
        }
        
        public static float ElapsedMillis(this GameTime dt)
        {
            return (float)dt.ElapsedGameTime.TotalMilliseconds;
        }

        public static Vector2 ToVector2XZ(this Vector3 v)
        {
            return new Vector2(v.X, v.Z);
        }

    }
}
