﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.LevelGeneration;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
using GameEngine.Helpers;
using Microsoft.Xna.Framework.Content;

namespace GameEngine.Xml
{
    public class LevelTileNode : ObjectNode, INodeRenderable, INodeLoadable
    {
        public LevelTile Tile { get; set; }
        public bool IsVisible { get; set; }
        public float Size { get; set; }
        public float Height { get; set; }

        public bool IsEmpty { get; set; }

        public int SpawnCount { get; set; }
        
        List<Quad> quads = new List<Quad>();
        Effect effect;
        Texture2D BLACK;

        public PolygonNode[] Statics = new PolygonNode[0];
        List<DynamicCircleNode> Dynamics = new List<DynamicCircleNode>();
        public List<SpawnPointNode> SpawnPoints = new List<SpawnPointNode>();
        public Vector3 CameraOffset { get; private set; }

        private ObjectNode modelContainer = new ObjectNode("models");
        
        public void AddDynamicNode(DynamicCircleNode node)
        {
            Dynamics.Add(node);
            ChildNodes.Add(node);
        }

        public void RemoveDynamicNode(DynamicCircleNode node)
        {
            Dynamics.Remove(node);
            ChildNodes.Remove(node);
        }

        public void AddSpawnPointNode(SpawnPointNode node)
        {
            AddDynamicNode(node);
            SpawnPoints.Add(node);
        }

        public void RemoveSpawnPointNode(SpawnPointNode node)
        {
            SpawnPoints.Remove(node);
            RemoveDynamicNode(node);
        }

        public DynamicCircleNode this[int i]
        {
            get
            {
                return Dynamics[i];
            }
        }

        public bool ContainsDynamic(DynamicCircleNode node)
        {
            return Dynamics.Contains(node);
        }


        public int DynamicCount { get { return Dynamics.Count; } }

        public LevelTileNode(string name) : base(name)
        {
        }

        public void LoadContent(Microsoft.Xna.Framework.Content.ContentManager content, Microsoft.Xna.Framework.Graphics.GraphicsDevice graphicsDevice)
        {
            effect = content.Load<Effect>(EngineFX.RenderGBuffer);


            if (IsEmpty)
            {
                Vector3 wallHeight = new Vector3(0, Height, 0);
                float wallWidth = Size / 2f;
                quads.Add(Quad.GenerateWall(graphicsDevice, new Vector3(-wallWidth, 0, -wallWidth), new Vector3(-wallWidth, 0, wallWidth), new Vector3(0, Height, 0)));
                quads.Add(Quad.GenerateWall(graphicsDevice, new Vector3(-wallWidth, 0, wallWidth), new Vector3(wallWidth, 0, wallWidth), new Vector3(0, Height, 0)));
                quads.Add(Quad.GenerateWall(graphicsDevice, new Vector3(wallWidth, 0, wallWidth), new Vector3(wallWidth, 0, -wallWidth), new Vector3(0, Height, 0)));
                quads.Add(Quad.GenerateWall(graphicsDevice, new Vector3(wallWidth, 0, -wallWidth), new Vector3(-wallWidth, 0, -wallWidth), new Vector3(0, Height, 0)));
                quads.Add(Quad.GenerateXZ(graphicsDevice, Size, Height, Vector2.Zero, Vector2.Zero, Vector4.One));

                BLACK = new Texture2D(graphicsDevice, 1, 1);
                BLACK.SetData(new Color[] { Color.Black });
            }
            else
            {
                /*List<PolygonNode> staticPolys = new List<PolygonNode>();
                Load(ChildNodes, content, graphicsDevice, staticPolys);
                this.Statics = staticPolys.ToArray();
                for (int i = ChildNodes.Count - 1; i >= 0; i--)
                {
                    if (ChildNodes[i] is PolygonNode)
                        continue;
                    else 
                    {
                        var node = ChildNodes[i];
                        ChildNodes.Remove(node);
                        modelContainer.ChildNodes.Add(node);
                    }
                }
                switch (Tile.Rotation)
                {
                    case GameEngine.LevelGeneration.Rotation.Right:
                        modelContainer.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, MathHelper.PiOver2);
                        break;
                    case GameEngine.LevelGeneration.Rotation.Down:
                        modelContainer.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, MathHelper.Pi);
                        break;
                    case GameEngine.LevelGeneration.Rotation.Left:
                        modelContainer.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, MathHelper.PiOver2 * 3);
                        break;
                }
                ChildNodes.Add(modelContainer);*/
                
                switch (Tile.Rotation)
                {
                    case GameEngine.LevelGeneration.Rotation.Right:
                        modelContainer.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, -MathHelper.PiOver2);
                        break;
                    case GameEngine.LevelGeneration.Rotation.Down:
                        modelContainer.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, -MathHelper.Pi);
                        break;
                    case GameEngine.LevelGeneration.Rotation.Left:
                        modelContainer.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, -MathHelper.PiOver2 * 3);
                        break;
                }
                modelContainer.ChildNodes.AddRange(ChildNodes);
                ChildNodes.Clear();
                ChildNodes.Add(modelContainer);

                // update world
                Update(new GameTime(), new DrawInfo());
                ChildNodes.UpdateRecursive(new GameTime(), new DrawInfo());
                
                List<PolygonNode> statics = new List<PolygonNode>();
                SpawnPoints.Clear();
                PopulateLists(ChildNodes, statics, SpawnPoints);
                this.Statics = statics.ToArray();
                
                ChildNodes.LoadRecursive(content, graphicsDevice);

            }

            Position = new Vector3(Size * Tile.X, 0, Size * Tile.Y);
            Update(new GameTime(), new DrawInfo());
            ChildNodes.UpdateRecursive(new GameTime(), new DrawInfo());
            FreezeRecursively(this);
        }

        public void FreezeRecursively(ObjectNode n)
        {
            if (n == null || n is DynamicCircleNode)
                return;

            ModelNode m = n as ModelNode;
            if (m != null && m.HasAnimation)
                return;

            n.FreezeWorld = true;
            for (int i = n.ChildNodes.Count - 1; i >= 0; i--)
            {
                FreezeRecursively(n.ChildNodes[i] as ObjectNode);
            }
        }


        void PopulateLists(NodeList list, List<PolygonNode> staticNodes, List<SpawnPointNode> spawnPointNodes)
        {
            for (int i = list.Count - 1; i >= 0; i--)
            {
                var item = list[i];

                PolygonNode pnode = item as PolygonNode;
                if (pnode != null)
                {
                    for (int x = 0; x < pnode.Points.Count; x++)
                    {
                        Vector3 t = Vector3.Transform(new Vector3(pnode.Points[x].X, 0, pnode.Points[x].Y), pnode.World);
                        pnode.Points[x] = new Vector2(t.X, t.Z);
                    }

                    pnode.Dirty();
                    pnode.ResolveDirty();

                    if (pnode.Parent != this)
                    {
                        pnode.Parent.ChildNodes.Remove(pnode);
                        this.ChildNodes.Add(pnode);
                        pnode.Position = Vector3.Zero;
                        pnode.Scale = Vector3.One;
                        pnode.Rotation = Quaternion.Identity;
                    }

                    // TODO: Move polys to top node and bake relative positions
                    staticNodes.Add(pnode);
                }
                else
                {
                    SpawnPointNode sNode = item as SpawnPointNode;
                    if (sNode != null)
                    {
                        // Remove from parent
                        sNode.Parent.ChildNodes.Remove(sNode);
                        sNode.Position = Vector3.Transform(Vector3.Zero, sNode.World);
                        // And add below this one 
                        AddSpawnPointNode(sNode);
                    }

                    TileInfoNode tNode = item as TileInfoNode;
                    if (tNode != null)
                    {
                        if (tNode.CameraOffset != Vector3.Zero)
                            this.CameraOffset = tNode.CameraOffset;
                        else
                            this.CameraOffset = tNode.DefaultCameraOffset;
                    }
                }
                PopulateLists(item.ChildNodes, staticNodes, spawnPointNodes);
            }
        }

        private void Load(NodeList list, ContentManager content, GraphicsDevice device, List<PolygonNode> staticNodes)
        {
            for (int i = list.Count - 1; i >= 0; i--)
            {
                var item = list[i];

                PolygonNode pnode = item as PolygonNode;
                if (pnode != null)
                {
                    SceneNode n = pnode.Parent;
                    while (n != null)
                    {
                        ObjectNode on = n as ObjectNode;
                        if (on != null)
                        {

                        }

                    }


                    switch (Tile.Rotation)
                    {
                        case GameEngine.LevelGeneration.Rotation.Right:
                            RotatePoints(pnode.Points, Matrix.CreateRotationZ(MathHelper.PiOver2));
                            break;
                        case GameEngine.LevelGeneration.Rotation.Down:
                            RotatePoints(pnode.Points, Matrix.CreateRotationZ(MathHelper.Pi));
                            break;
                        case GameEngine.LevelGeneration.Rotation.Left:
                            RotatePoints(pnode.Points, Matrix.CreateRotationZ(MathHelper.PiOver2 * 3));
                            break;
                    }
                    pnode.Dirty();
                    pnode.ResolveDirty();

                    if (pnode.Parent != this)
                    {
                        pnode.Parent.ChildNodes.Remove(pnode);
                        this.ChildNodes.Add(pnode);
                    }

                    // TODO: Move polys to top node and bake relative positions
                    staticNodes.Add(pnode);

                }

                INodeLoadable loadable = item as INodeLoadable;
                if (loadable != null)
                {
                    loadable.LoadContent(content, device);
                }
                Load(item.ChildNodes, content, device, staticNodes);
            }
        }

        public void TranslatePoints(List<Vector2> points, Vector2 translation)
        {
            for (int i = 0; i < points.Count; i++)
            {
                points[i] += translation;
            }
        }

        public void RotatePoints(List<Vector2> points, Matrix rotation)
        {


            for (int i = 0; i < points.Count; i++)
            {
                points[i] = Vector2.Transform(points[i], rotation);
            }
        }



        public void UnloadContent()
        {
            foreach (var item in quads)
            {
                item.Dispose();
            }
            quads.Clear();
            Util.NullDispose(ref BLACK);
        }
        public override void Update(GameTime dt, DrawInfo drawInfo)
        {
            base.Update(dt, drawInfo);
        }


        public void Draw(Microsoft.Xna.Framework.GameTime dt, DrawInfo drawInfo)
        {
            if (IsEmpty)
            {
                effect.Parameters["World"].SetValue(World);
                effect.Parameters["Texture"].SetValue(BLACK);
                for (int i = 0; i < quads.Count; i++)
                {
                    quads[i].Draw(drawInfo.GraphicsDevice, effect);
                }
            }
            else
            {
                ChildNodes.DrawRecursive(dt, drawInfo);
            }
        }

        private void DrawWall(Quad wall, DrawInfo drawInfo)
        {
            if (Vector3.Dot(wall.Normal, (Position + wall.Center) - drawInfo.Camera.Position) < 0)
                effect.CurrentTechnique = effect.Techniques[1];
            else
                effect.CurrentTechnique = effect.Techniques[0];

            wall.Draw(drawInfo.GraphicsDevice, effect);
        }

        public List<Vector2> FindSpawnPositions(float radius)
        {
            List<Vector2> spawnPoint = new List<Vector2>();

            Vector2 dummy;
            Vector2 start = new Vector2(-TilesetManager.CurrentTileset.TileSize / 2f + radius, -TilesetManager.CurrentTileset.TileSize / 2f + radius);
            Vector2 end = new Vector2(TilesetManager.CurrentTileset.TileSize / 2f - radius, TilesetManager.CurrentTileset.TileSize / 2f - radius);

            BoundingCircle boundingCircle = new BoundingCircle(start, radius);

            for (float x = start.X; x <= end.X; x += radius)
            {
                for (float y = start.Y; y <= end.Y; y += radius)
                {
                    boundingCircle.Center = new Vector2(x, y);
                    bool ok = true;
                    for (int i = 0; i < Statics.Length; i++)
                    {
                        if (Statics[i].Intersects(boundingCircle, Vector2.Zero, out dummy))
                        {
                            ok = false;
                            break;
                        }
                    }
                    if (ok)
                        spawnPoint.Add(boundingCircle.Center);
                }
            }
            return spawnPoint;
        }

        public SpawnPointNode GetAvailableSpawnPoint(SpawnPointNode.SpawnType type) 
        {
            Vector2 dummy;
            for (int i = 0; i < SpawnPoints.Count; i++)
            {
                if (SpawnPoints[i].SpawnPointType == type)
                {
                    bool ok = true;
                    for (int j = 0; j < Dynamics.Count; j++)
                    {
                        if (Dynamics[j] == SpawnPoints[i])
                            continue;
                        if(SpawnPoints[i].GetCollisionCircle().Intersects(Dynamics[j].GetCollisionCircle(), out dummy))
                        {
                            ok = false;
                            break;
                        }
                    }
                    if (ok)
                        return SpawnPoints[i];
                }
            }

            return null;
        }

        public SpawnPointNode GetRandomSpawnPoint(SpawnPointNode.SpawnType type)
        {
            int typeCount = 0;
            for (int i = 0; i < SpawnPoints.Count; i++)
                if (SpawnPoints[i].SpawnPointType == type)
                    typeCount++;

            if (typeCount == 0)
                return null;

            typeCount = Util.Random.Next(typeCount);

            for (int i = 0; i < SpawnPoints.Count; i++)
            {
                if (SpawnPoints[i].SpawnPointType == type)
                    if (typeCount == 0)
                        return SpawnPoints[i];
                    else
                        typeCount--;
            }

            return null;
        }

        public SpawnPointNode GetGoodSpawnPoint(SpawnPointNode.SpawnType type)
        {
            SpawnPointNode result = GetAvailableSpawnPoint(type);
            if (result != null)
                return result;

            return GetRandomSpawnPoint(type);
        }


    }
}
