using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework;
using System.ComponentModel;
using GameEngine.Helpers;
using GameEngine.Animation;
using System.Linq;

namespace GameEngine.Xml
{
    public delegate void EffectConfigHandler(ModelNode n, ModelMesh mesh, int partIndex);

    public class ModelNode : ObjectNode, INodeLoadable, INodeRenderable
    {
        private bool hasAnimation;
        private Model model;
        private bool isVisible = true;
        private string asset;
        BoundingBox aabb;
        private Vector3 maxAAExtent;

        [Browsable(false)]
        public bool IsVisible { get { return isVisible; } set { isVisible = value; } }
        public string Asset { get { return asset; } set { asset = value; } }
        public bool HasAnimation { get { return hasAnimation; } }

        public double AnimationFactor { get; set; }

        [Browsable(false)]
        public string RunningAnimation { get { return AnimationPlayer.RunningAnimation; } }
        
        public AnimationPlayer AnimationPlayer;
        public StaticModelManipulator StaticManipulator;

        [Browsable(false)]
        public Model Model{ get { return model; } }
	
        private Vector3 bboxOffset = Vector3.One;
        private Vector3 initialExtent = Vector3.One;

        public static List<ModelNode> UpdatedAnimationBears = new List<ModelNode>();
        public Vector4 tintColor = Vector4.One;

        bool hitsFrustrum = false;
        BoundingBox bb;

        //public event EffectConfigHandler EffectConfig;

        public ModelNode(string name) : base(name)
        {
            AnimationFactor = 1f;
        }

        public override void PopulateXml(System.Xml.XmlElement element, System.Xml.XmlDocument document)
        {
            base.PopulateXml(element, document);
            if(!string.IsNullOrEmpty(asset))
                element.SetAttribute("Asset", asset);
        }
        public override void SetFromXml(System.Xml.XmlElement element)
        {
            base.SetFromXml(element);
            if (element.HasAttribute("Asset"))
            {
                asset = element.Attributes["Asset"].Value;
            }
        }

        public virtual void LoadContent(Microsoft.Xna.Framework.Content.ContentManager manager, GraphicsDevice graphicsDevice)
        {
            if (!manager.TryLoad<Model>(asset, out model))
                return;


            Dictionary<String, object> tag = model.Tag as Dictionary<String, object>;
            if (tag == null)
                return;

            if (tag.ContainsKey("AnimationData") && ((AnimationData)tag["AnimationData"]).Tracks.Count > 0)
            {
                AnimationPlayer = new AnimationPlayer(tag["AnimationData"] as AnimationData, model);

                this.hasAnimation = true;

                if (AnimationPlayer.Data.Tracks.Count == 1)
                {
                    StartAnimation(AnimationPlayer.Data.Tracks.Keys.First(), true);
                }
            }
            else
            {
                StaticManipulator = new StaticModelManipulator(model);
            }

            if (tag.ContainsKey("BoundingBox"))
            {
                aabb = (BoundingBox)tag["BoundingBox"];
                initialExtent = 0.5f * Vector3.Subtract(aabb.Max, aabb.Min);
                bboxOffset = aabb.Min + initialExtent;
                UpdateBoundingBox();
            }

            foreach (ModelMesh mesh in model.Meshes)
            {
                foreach (Effect e in mesh.Effects)
                {
                    if (hasAnimation)
                    {
                        if (AnimationPlayer.Data.Skinned && AnimationPlayer.Data.ModelBoneToSkinMap[mesh.ParentBone.Parent.Index] == -1)
                        {
                            e.CurrentTechnique = e.Techniques["Skinned"];
                        }
                        else
                        {
                            e.CurrentTechnique = e.Techniques["Technique1"];
                        }
                    }

                }
            }

        }

        public override void UpdateBoundingBox()
        {
            Vector3 bbScale, bbTranslation;
            Quaternion bbRotation;

            World.Decompose(out bbScale, out bbRotation, out bbTranslation);
            // Rotation is same rotation as the model
            orientedBoundingBox.Rotation = Matrix.CreateFromQuaternion(bbRotation);
            // The position is the object's position plus the original offset, rotated and scaled to fit the model
            orientedBoundingBox.Center = bbTranslation + Vector3.Transform(bboxOffset * bbScale, orientedBoundingBox.Rotation);
            // The extent of the bounding box is the initial extent scaled to the object's new size
            orientedBoundingBox.Extent = initialExtent * bbScale;

            maxAAExtent = orientedBoundingBox.MaxAAExtent();
        }

        public virtual void UnloadContent()
        {
            StaticManipulator = null;
            model = null;
            
        }

        public void StartAnimation(string name, bool loop)
        {
            if (hasAnimation)
            {
                AnimationPlayer.StartAnimation(name, loop);
            }
        }

        public override void Update(GameTime dt, DrawInfo drawInfo)
        {
            if(drawInfo.UpdateOBBs || float.IsNaN(maxAAExtent.X))
                UpdateBoundingBox();
            
            base.Update(dt, drawInfo);

            bb = new BoundingBox(-maxAAExtent + WorldPosition, maxAAExtent + WorldPosition);
            
            hitsFrustrum = false;
            if (drawInfo.AllCameras != null && isVisible)
            {
                for (int i = 0; i < drawInfo.AllCameras.Length; i++)
                {
                    if (drawInfo.AllCameras[i].Frustrum.Intersects(bb))
                    {
                        hitsFrustrum = true;
                        break;
                    }
                }
            }

            if (hasAnimation && isVisible && (drawInfo.AllCameras == null || hitsFrustrum) && !FreezeWorld)
            {
                UpdatedAnimationBears.Add(this);
                TimeSpan factored = TimeSpan.FromTicks((long)(dt.ElapsedGameTime.Ticks * AnimationFactor));
                AnimationPlayer.Update(factored, true);
            }

        }

        public void Draw(GameTime dt, DrawInfo di)
        {

            if (!isVisible || !di.Camera.Frustrum.Intersects(bb))
                return;

            di.Models.Add(this);
        }

        public override void MinAndMaxPoint(out Vector3 minPoint, out Vector3 maxPoint)
        {
            Vector3 obbExtent = orientedBoundingBox.MaxAAExtent();
            minPoint = orientedBoundingBox.Center - obbExtent;
            maxPoint = orientedBoundingBox.Center + obbExtent;
        }

        public void ConfigureEffect(ModelMesh mesh, int part)
        {
            if (mesh == Model.Meshes[0] && part == 2)
                mesh.MeshParts[part].Effect.Parameters["tintColor"].SetValue(tintColor);
        }

        public static int SortByAsset(ModelNode m1, ModelNode m2) { return m1.Asset.CompareTo(m2.asset); }

    }
}
