using System;
using System.Collections.Generic;
using System.Text;
using Microsoft.Xna.Framework;
using System.ComponentModel;
using GameEngine.Helpers;


namespace GameEngine.Xml
{
    public class ObjectNode : SceneNode, INodeUpdateable
    {
        public bool isActive = true;
        public Vector3 Position, Scale = Vector3.One;
        public Quaternion Rotation = Quaternion.Identity;
        
        public Matrix World;
        
        public ObjectNode(string name) : base(name) { }

        [Category("Object")]
        public virtual Vector3 InitialPosition { get { return Position; } set { Position = value; } }
        [Category("Object")]
        public virtual Vector3 InitialScale { get { return Scale; } set { Scale = value; } }
        [Category("Object")]
        public virtual Quaternion InitialRotation { get { return Rotation; } set { Rotation = value; } }

        protected OrientedBoundingBox orientedBoundingBox = new OrientedBoundingBox();

        [Browsable(false)]
        public virtual OrientedBoundingBox OrientedBoundingBox
        {
            get {
                UpdateBoundingBox();
                return orientedBoundingBox; 
            }
            set { orientedBoundingBox = value; }
        }


        [Browsable(false)]
        public bool IsActive 
        { 
            get 
            {
                return isActive; 
            } 
            set 
            { 
                isActive = value; 
            } 
        }

        public bool FreezeWorld = false;

        public virtual void Update(GameTime dt, DrawInfo drawInfo) 
        {
            if (!FreezeWorld)
            {
                World = Matrix.CreateScale(Scale) * Matrix.CreateFromQuaternion(Rotation) * Matrix.CreateTranslation(Position);
                ObjectNode p = Parent as ObjectNode;
                if (p != null)
                {
                    World *= p.World;
                }
            }
            /*for (int i = 0; i < updateNodes.Count; i++)
            {
                updateNodes[i].Update(dt, drawInfo);

            }*/
        }

        public virtual void UpdateBoundingBox()
        {
            Vector3 bbScale, bbTranslation;
            Quaternion bbRotation;

            World.Decompose(out bbScale, out bbRotation, out bbTranslation);
            // Rotation is same rotation as the model
            orientedBoundingBox.Rotation = Matrix.Identity;
            // The position is the object's position plus the original offset, rotated and scaled to fit the model
            orientedBoundingBox.Center = bbTranslation;
            // The extent of the bounding box is the initial extent scaled to the object's new size
            orientedBoundingBox.Extent =  MaxAAExtent();
        }

        public override void PopulateXml(System.Xml.XmlElement element, System.Xml.XmlDocument document)
        {
            base.PopulateXml(element, document);
            
            if(Position != Vector3.Zero)
                XmlHelper.SetAttribute(element, "Position", Position);
            if(Scale != Vector3.One)
                XmlHelper.SetAttribute(element, "Scale", Scale);
            if(Rotation != Quaternion.Identity)
                XmlHelper.SetAttribute(element, "Rotation", Rotation);
        }
        public override void SetFromXml(System.Xml.XmlElement element)
        {
            base.SetFromXml(element);

            if (element.HasAttribute("Position"))
                Position = XmlHelper.ParseVector3(element.Attributes["Position"]);

            if (element.HasAttribute("Scale"))
                Scale = XmlHelper.ParseVector3(element.Attributes["Scale"]);

            if (element.HasAttribute("Rotation"))
                Rotation = XmlHelper.ParseQuaternion(element.Attributes["Rotation"]);
        }

        List<INodeUpdateable> updateNodes = new List<INodeUpdateable>();
        [Browsable(false)]
        public ICollection<INodeUpdateable> UpdateNodes
        {
            get
            {
                return updateNodes;
            }
        }

        [Browsable(false)]
        public Vector3 WorldPosition
        {
            get { return World.Translation; }
            set
            {
                Vector3 result = value;
                ObjectNode p = Parent as ObjectNode;
                if (p != null)
                {
                    // reverse parent world rotation and scaling
                    Vector3 tmpScale, tmpTranslation;
                    Quaternion tmpRotation;
                    p.World.Decompose(out tmpScale, out tmpRotation, out tmpTranslation);
                    result = Vector3.Transform(result - tmpTranslation, Matrix.Transpose(Matrix.CreateFromQuaternion(tmpRotation)));
                    result /= tmpScale;
                }
                Position = result;
                World.Translation = value;
            }
        }

        public virtual Vector3 MaxAAExtent()
        {
            Vector3 tmpMinPoint, tmpMaxPoint;

            MinAndMaxPoint(out tmpMinPoint, out tmpMaxPoint);

            Vector3 maxExtent;

            maxExtent.X = Math.Max(Math.Abs(World.Translation.X - tmpMaxPoint.X), Math.Abs(World.Translation.X - tmpMinPoint.X));
            maxExtent.Y = Math.Max(Math.Abs(World.Translation.Y - tmpMaxPoint.Y), Math.Abs(World.Translation.Y - tmpMinPoint.Y));
            maxExtent.Z = Math.Max(Math.Abs(World.Translation.Z - tmpMaxPoint.Z), Math.Abs(World.Translation.Z - tmpMinPoint.Z));

            return maxExtent;
        }

        public virtual void MinAndMaxPoint(out Vector3 minPoint, out Vector3 maxPoint) 
        {
            minPoint = maxPoint = World.Translation;            

            Vector3 tmpMinPoint, tmpMaxPoint;

            bool changed = false;

            foreach (SceneNode child in ChildNodes)
            {
                ObjectNode objectChild = child as ObjectNode;
                if (objectChild != null)
                {
                    objectChild.MinAndMaxPoint(out tmpMinPoint, out tmpMaxPoint);
                    minPoint.X = Math.Min(tmpMinPoint.X, minPoint.X);
                    minPoint.Y = Math.Min(tmpMinPoint.Y, minPoint.Y);
                    minPoint.Z = Math.Min(tmpMinPoint.Z, minPoint.Z);

                    maxPoint.X = Math.Max(tmpMaxPoint.X, maxPoint.X);
                    maxPoint.Y = Math.Max(tmpMaxPoint.Y, maxPoint.Y);
                    maxPoint.Z = Math.Max(tmpMaxPoint.Z, maxPoint.Z);

                    changed = true;
                }
            }

            if (!changed)
            {
                minPoint.X -= 5;
                minPoint.Y -= 5;
                minPoint.Z -= 5;
                maxPoint.X += 5;
                maxPoint.Y += 5;
                maxPoint.Z += 5;
            }
        }

    }
}
