﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.LevelGeneration;
using System.ComponentModel;
using GameEngine.Designer;
using Microsoft.Xna.Framework;
using GameEngine.Helpers;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;

namespace GameEngine.Xml
{
    public enum SpecialTileNodeType
    {
        None,
        Start,
        End
    }

    public class TileInfoNode : ObjectNode, INodeLoadable, INodeRenderable
    {
        public const TileType DefaultTileType = TileType.End;
        public TileType TileType { get; set; }

        private bool dirtied, visible = true;
        private string floorTextureAsset;
        private Vector2 floorOffset, floorSize;
        private QuadNormal floor;
        private Texture2D texture;
        private bool hasTexture;
        private SpecialTileNodeType specialType;
        private GeometryDrawInfo gdi = new GeometryDrawInfo();

        public Vector3 DefaultCameraOffset = new Vector3(0, 260f, 100f);
        private Vector3 cameraOffset;

        private Vector4 floorColor = Vector4.One;
        private Effect fxGBuffer;
        private ContentManager content;
        private GraphicsDevice device;
        private TilesetNode tileSet;

        [Category("Tile")]
        [DisplayName("Texture")]
        [TypeConverter(typeof(TextureChooser))]
        public string FloorTexture
        {
            get { return floorTextureAsset; }
            set
            {
                if (floorTextureAsset == value)
                    return;

                floorTextureAsset = value;
                Dirty();
            }
        }

        [Category("Tile")]
        [DisplayName("Texture Offset")]
        public Vector2 FloorTextureOffset
        {
            get { return floorOffset; }
            set
            {
                if (floorOffset == value)
                    return;

                floorOffset = value;
                Dirty();
            }
        }

        [Category("Tile")]
        [DisplayName("Texture Size")]
        public Vector2 FloorTextureSize
        {
            get { return floorSize; }
            set
            {
                if (floorSize == value)
                    return;

                floorSize = value;
                Dirty();
            }
        }

        [Category("Tile")]
        [DisplayName("Floor Color")]
        public Vector4 FloorColor
        {
            get { return floorColor; }
            set
            {
                if (floorColor == value)
                    return;

                floorColor = value;
                Dirty();
            }
        }

        [Category("Tile")]
        [DisplayName("Special Type")]
        public SpecialTileNodeType SpecialType
        {
            get { return this.specialType; }
            set
            {
                if (specialType == value)
                    return;

                specialType = value;
                Dirty();
            }
        }

        [Category("Tile")]
        [DisplayName("Camera Offset")]
        public Vector3 CameraOffset
        {
            get { return this.cameraOffset; }
            set
            {
                if (cameraOffset == value)
                    return;

                cameraOffset = value;
                Dirty();
            }
        }


        public TileInfoNode(string name) : base(name)
        {
            this.TileType = DefaultTileType;
            this.CameraOffset = DefaultCameraOffset;
        }

        public override void PopulateXml(System.Xml.XmlElement element, System.Xml.XmlDocument document)
        {
            if (TileType != DefaultTileType)
                element.SetAttribute("TileType", TileType.ToString());

            if (!string.IsNullOrEmpty(FloorTexture))
                element.SetAttribute("FloorTexture", FloorTexture);

            if (FloorTextureOffset != Vector2.Zero)
                XmlHelper.SetAttribute(element, "FloorTextureOffset", FloorTextureOffset);

            if (FloorTextureSize != Vector2.Zero)
                XmlHelper.SetAttribute(element, "FloorTextureSize", FloorTextureSize);

            if (FloorColor != Vector4.One)
                XmlHelper.SetAttribute(element, "FloorColor", FloorColor);

            if (SpecialType != SpecialTileNodeType.None)
                element.SetAttribute("SpecialType", SpecialType.ToString());

            if (CameraOffset != Vector3.One)
                XmlHelper.SetAttribute(element, "CameraOffset", CameraOffset);

            base.PopulateXml(element, document);
        }

        public override void SetFromXml(System.Xml.XmlElement element)
        {
            if (element.HasAttribute("TileType"))
                this.TileType = (TileType)Enum.Parse(typeof(TileType), element.Attributes["TileType"].Value);

            if (element.HasAttribute("FloorTexture"))
                this.FloorTexture = element.Attributes["FloorTexture"].Value;

            if (element.HasAttribute("FloorTextureOffset"))
                FloorTextureOffset = XmlHelper.ParseVector2(element.Attributes["FloorTextureOffset"]);

            if (element.HasAttribute("FloorTextureSize"))
                FloorTextureSize = XmlHelper.ParseVector2(element.Attributes["FloorTextureSize"]);

            if (element.HasAttribute("FloorColor"))
                FloorColor = XmlHelper.ParseVector4(element.Attributes["FloorColor"]);

            if (element.HasAttribute("SpecialType"))
                SpecialType = (SpecialTileNodeType)Enum.Parse(typeof(SpecialTileNodeType), element.Attributes["SpecialType"].Value);

            if (element.HasAttribute("CameraOffset"))
                CameraOffset = XmlHelper.ParseVector3(element.Attributes["CameraOffset"]);
            

            base.SetFromXml(element);
        }

        private void Dirty()
        {
            dirtied = true;
        }

        public void LoadContent(Microsoft.Xna.Framework.Content.ContentManager content, Microsoft.Xna.Framework.Graphics.GraphicsDevice graphicsDevice)
        {
            device = graphicsDevice;
            this.content = content;

            tileSet = Tag as TilesetNode;
            if (tileSet == null)
                tileSet = TilesetManager.CurrentTileset;

            if (content == null || tileSet == null)
                return;

            content.Load(EngineFX.RenderGBuffer, out fxGBuffer);

            hasTexture = content.TryLoad(FloorTexture, out texture);
            if (!hasTexture)
            {
                texture = new Texture2D(graphicsDevice, 1, 1);
                texture.SetData(new Color[] { new Color(floorColor) });
            }

            floor = QuadNormal.GenerateXZ(graphicsDevice, tileSet.TileSize, 0, floorOffset, floorSize, floorColor);
            gdi.geometry = floor;
            gdi.texture = texture;
            gdi.textureAsset = FloorTexture;
        }

        public void UnloadContent()
        {
            Util.NullDispose(ref floor);
            if(!hasTexture)
                Util.NullDispose(ref texture);

        }

        public bool IsVisible
        {
            get
            {
                return visible;   
            }
            set
            {
                visible = value;
            }
        }

        public override void Update(Microsoft.Xna.Framework.GameTime time, DrawInfo drawInfo)
        {
            if (dirtied)
            {
                UnloadContent();
                LoadContent(content, device);
                dirtied = false;
            }
            base.Update(time, drawInfo);
        }

        public void Draw(GameTime dt, DrawInfo drawInfo)
        {
            if (floor != null)
            {
                if (FloorTexture != null)
                {
                    gdi.world = World;
                    drawInfo.Gdis.Add(gdi);
                }
/*                fxGBuffer.Parameters["World"].SetValue(World);

                fxGBuffer.Parameters["Texture"].SetValue(texture);

                floor.Draw(drawInfo.GraphicsDevice, fxGBuffer);*/
            }

           /* if (floor != null)
            {
                floorEffect.CurrentTechnique = floorEffect.Techniques[0];

                EngineFX.SetPolyWallSharedParams(floorEffect, drawInfo);

                floorEffect.Parameters["World"].SetValue(World);

                if (texture != null)
                    drawInfo.GraphicsDevice.Textures[0] = texture;

                floorEffect.Parameters["useTexture"].SetValue(texture != null);

                floor.Draw(drawInfo.GraphicsDevice, floorEffect);
            }*/
        }
    }




}
