using System;
using System.Collections.Generic;
using System.Text;
using System.Xml;
using Microsoft.Xna.Framework;

namespace GameEngine.Xml
{
    public class XmlHelper
    {
        public static XmlDocument CreateDocument(IEnumerable<SceneNode> node)
        {
            XmlDocument document = new XmlDocument();
            document.AppendChild(document.CreateXmlDeclaration("1.0", null, null));
            XmlElement root = document.CreateElement("Scene");
            foreach (SceneNode n in node)
            {
                root.AppendChild(CreateElement(n, document));
            }
            document.AppendChild(root);

            return document;
        }

        public static XmlElement CreateElement(SceneNode node, XmlDocument document)
        {
            XmlElement element = document.CreateElement(node.GetType().Name);
            node.PopulateXml(element, document);
            if(string.IsNullOrEmpty(node.Include))
                foreach (SceneNode childNode in node.ChildNodes)
                {
                    XmlElement childElement = CreateElement(childNode, document);
                    element.AppendChild(childElement);
                }
            return element;
        }

        public static XmlElement CreateElement(string name, Vector3 vector, XmlDocument document)
        {
            XmlElement element = document.CreateElement(name);
            SetAttribute(element, "Vector3", vector);
            return element;
        }
        public static Vector3 ParseVector3(XmlElement element)
        {
            return Vector3FromString(element.Attributes["Vector3"].Value);
        }
        public static Vector3 ParseVector3(XmlAttribute attribute)
        {
            return Vector3FromString(attribute.Value);
        }
        private static Vector3 Vector3FromString(string attribute)
        {
            Vector3 result;
            string[] parts = attribute.Split(';');
            result.X = Single.Parse(parts[0]);
            result.Y = Single.Parse(parts[1]);
            result.Z = Single.Parse(parts[2]);
            return result;
        }
        public static void SetAttribute(XmlElement element, string name, Vector3 vector)
        {
            element.SetAttribute(name, string.Format("{0}; {1}; {2}", vector.X, vector.Y, vector.Z));
        }

        public static XmlElement CreateElement(string name, Vector2 vector, XmlDocument document)
        {
            XmlElement element = document.CreateElement(name);
            SetAttribute(element, "Vector2", vector);
            return element;
        }
        public static Vector2 ParseVector2(XmlElement element)
        {
            return Vector2FromString(element.Attributes["Vector2"].Value);
        }
        public static Vector2 ParseVector2(XmlAttribute attribute)
        {
            return Vector2FromString(attribute.Value);
        }
        private static Vector2 Vector2FromString(string attribute)
        {
            Vector2 result;
            string[] parts = attribute.Split(';');
            result.X = Single.Parse(parts[0]);
            result.Y = Single.Parse(parts[1]);
            return result;
        }
        public static void SetAttribute(XmlElement element, string name, Vector2 vector)
        {
            element.SetAttribute(name, string.Format("{0}; {1}", vector.X, vector.Y));
        }

        public static XmlElement CreateElement(string name, Vector4 vector, XmlDocument document)
        {
            XmlElement element = document.CreateElement(name);
            SetAttribute(element, "Vector4", vector);
            return element;
        }
        public static Vector4 ParseVector4(XmlElement element)
        {
            return Vector4FromString(element.Attributes["Vector4"].Value);
        }
        public static Vector4 ParseVector4(XmlAttribute attribute)
        {
            return Vector4FromString(attribute.Value);
        }
        private static Vector4 Vector4FromString(string attribute)
        {
            Vector4 result;
            string[] parts = attribute.Split(';');
            result.X = Single.Parse(parts[0]);
            result.Y = Single.Parse(parts[1]);
            result.Z = Single.Parse(parts[2]);
            result.W = Single.Parse(parts[3]);
            return result;
        }
        public static void SetAttribute(XmlElement element, string name, Vector4 vector)
        {
            element.SetAttribute(name, string.Format("{0}; {1}; {2}; {3}", vector.X, vector.Y, vector.Z, vector.W));
        }

        public static XmlElement CreateElement(string name, Quaternion vector, XmlDocument document)
        {
            XmlElement element = document.CreateElement(name);
            SetAttribute(element, "Quaternion", vector);
            return element;
        }
        public static Quaternion ParseQuaternion(XmlElement element)
        {
            return QuaternionFromString(element.Attributes["Quaternion"].Value);
        }
        public static Quaternion ParseQuaternion(XmlAttribute attribute)
        {
            return QuaternionFromString(attribute.Value);
        }
        private static Quaternion QuaternionFromString(string attribute)
        {
            Quaternion result;
            string[] parts = attribute.Split(';');
            result.X = Single.Parse(parts[0]);
            result.Y = Single.Parse(parts[1]);
            result.Z = Single.Parse(parts[2]);
            result.W = Single.Parse(parts[3]);
            return result;
        }
        public static void SetAttribute(XmlElement element, string name, Quaternion quaternion)
        {
            element.SetAttribute(name, string.Format("{0}; {1}; {2}; {3}", quaternion.X, quaternion.Y, quaternion.Z, quaternion.W));
        }




        public static XmlElement CreateElement(string name, Matrix matrix, XmlDocument document)
        {
            XmlElement element = document.CreateElement(name);
            XmlElement row1 = CreateElement("R1", new Vector4(matrix.M11, matrix.M12, matrix.M13, matrix.M14), document);
            XmlElement row2 = CreateElement("R2", new Vector4(matrix.M21, matrix.M22, matrix.M23, matrix.M24), document);
            XmlElement row3 = CreateElement("R3", new Vector4(matrix.M31, matrix.M32, matrix.M33, matrix.M34), document);
            XmlElement row4 = CreateElement("R4", new Vector4(matrix.M41, matrix.M42, matrix.M43, matrix.M44), document);
            element.AppendChild(row1);
            element.AppendChild(row2);
            element.AppendChild(row3);
            element.AppendChild(row4);
            return element;
        }

        public static Matrix ParseMatrix(XmlNode element)
        {
            int i = 0;
            Vector4[] r = new Vector4[4];

            foreach (XmlElement cNode in element.ChildNodes)
            {
                if (element.Name == "Vector4")
                {
                    r[i] = ParseVector4(cNode);
                    if (++i == 4)
                        break;
                }
                
            }
            return new Matrix(r[0].X, r[0].Y, r[0].Z, r[0].W, r[1].X, r[1].Y, r[1].Z, r[1].W, r[2].X, r[2].Y, r[2].Z, r[2].W, r[3].X, r[3].Y, r[3].Z, r[3].W);
        }
    }

}
