﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Content;
using GameEngine.Cameras;
using GameEngine;
using GameEngine.Xml;
using Microsoft.Xna.Framework.Graphics;

namespace GamePrototype
{
    public class BaseLevel
    {
        public string LevelAsset { get; set; }
        public GraphicsDeviceManager Graphics { get; set; }
        public GraphicsDevice GraphicsDevice { get { return Graphics.GraphicsDevice; } }
        public Game Game { get; set; }
        public ContentManager Content { get; set; }
        public OrthoOrbitCamera Camera { get; set; }
        public ForestHandler Forest { get; set; }
        public AttentionSpan AttentionSpan { get; set; }

        protected DrawInfo DrawInfo = new DrawInfo();
        protected List<INodeUpdateable> updateRoots = new List<INodeUpdateable>();
        protected Queue<INodeRenderable> renderQueue = new Queue<INodeRenderable>();
        protected Dictionary<string, object> loadProperties = new Dictionary<string, object>();

        public BaseLevel(Game game, GraphicsDeviceManager graphics, string asset)
        {
            this.Game = game;
            this.Graphics = graphics;
            this.LevelAsset = asset;
            this.Content = new ContentManager(game.Services, "Content");
            this.Forest = new ForestHandler();
            this.Camera = new OrthoOrbitCamera(CameraMode.Perspective);
            //this.AttentionSpan = new AttentionSpan();
        }

        protected virtual void Forest_NodeRemoved(ForestHandler forest, string path, SceneNode node)
        {
            updateRoots.Clear();
            RebuildUpdateList();
        }

        protected virtual void Forest_NodeAdded(ForestHandler forest, string path, SceneNode node)
        {
            updateRoots.Clear();
            RebuildUpdateList();
        }

        protected virtual void RebuildUpdateList()
        {
            foreach (var node in Forest.NodesDepthFirst)
            {
                INodeUpdateable updatableNode = node as INodeUpdateable;
                if (updatableNode != null)
                {
                    if (!(node.Parent is INodeUpdateable))
                    {
                        updateRoots.Add(updatableNode);
                    }
                    foreach (GameEngine.Xml.SceneNode cNode in node.ChildNodes)
                    {
                        if (cNode is INodeUpdateable)
                            updatableNode.UpdateNodes.Add(cNode as INodeUpdateable);
                    }
                }
            }
        }

        public virtual void LoadContent()
        {
            this.Forest.AddRoots(XmlTest.LoadTest(LevelAsset, Content));
            this.Forest.NodeAdded += new ForestMapChange(Forest_NodeAdded);
            this.Forest.NodeRemoved += new ForestMapChange(Forest_NodeRemoved);

            foreach (var node in Forest.NodesBreadthFirst)
            {
                INodeLoadable nodeLoad = node as INodeLoadable;
                if (nodeLoad != null)
                    nodeLoad.LoadContent(Content, Graphics.GraphicsDevice, Forest.Map, loadProperties);
            }        

            RebuildUpdateList();
            
            DrawInfo.GraphicsDevice = Graphics.GraphicsDevice;
        }

        public virtual void Draw(GameTime dt)
        {
            Graphics.GraphicsDevice.Clear(Color.Black);

            while (renderQueue.Count > 0)
                renderQueue.Dequeue().Draw(dt, DrawInfo);
        }

        public virtual void Update(GameTime dt)
        {
            Camera.Viewport = Graphics.GraphicsDevice.Viewport;
            Camera.UpdateCamera(dt);
            DrawInfo.Camera = Camera;


            UpdateNodes(updateRoots, dt, DrawInfo);
        }

        protected virtual void UpdateNodes(IEnumerable<INodeUpdateable> nodes, GameTime dt, DrawInfo di)
        {
            foreach (INodeUpdateable node in nodes)
            {
                if (node.IsActive)
                {
                    node.Update(dt);
                    INodeRenderable renderable = node as INodeRenderable;
                    if (renderable != null && renderable.IsVisible)
                    {
                        renderQueue.Enqueue(renderable);
                    }
                    UpdateNodes(node.UpdateNodes, dt, di);
                }
            }
        }

        public virtual void UnloadContent()
        {
            Content.Unload();
        }

    }
}
