﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.ScreenSystem;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;

namespace GamePrototype.GUI.Screens
{
    public class DialogueScreen : GameScreen
    {
        public string DialogueResult { get; set; }
        public int CurrentIndex { get { return this.currentIndex; } }
        
        private string text;
        private string[] options;
        private int currentIndex = 0;
        private string currentScreenID;
        private Texture2D texture;
        
        private const float horizontalPadding = 25;
        private const float verticalPadding = 25;
        private Color backGroundColor = Color.Black;
        private Color textColor = Color.Gray;
        private Color optionsColor = Color.CornflowerBlue;
        private Color optionsSelectedColor = Color.White;
        private Color frameColor = Color.Gray;
        private const int frameSize = 3;
        private const float textScale = 1.6f;
        private const float optionsScale = 1.2f;
        private const float spaceFromTextToOptions = 20;
        private const float spaceBetweenOptions = 15;
                
        public DialogueScreen(string id, ScreenController controller, Game game)
            : base(id, controller, game)
        {
            this.DialogueResult = "";
        }

        public override void LoadContent(GraphicsDeviceManager graphics)
        {
            texture = new Texture2D(graphics.GraphicsDevice, 1, 1, 1, TextureUsage.None, SurfaceFormat.Vector4);
            texture.SetData(new Vector4[] { Vector4.One });
        }

        public override void Initialize(GraphicsDevice graphics)
        {
            
        }

        public override void Update(GameTime gameTime)
        {
            int menuIndex = currentIndex; 

            if (InputState.KeyPressed(Keys.Escape))
            {
                this.controller.SetHiddenScreen("DialogueScreen");
                this.controller.SetActiveScreen(currentScreenID);
            }
            else
            {
                if (options.Length == 0)
                {
                    Keys[] presssed = InputState.GetPressedKeys();

                    if (presssed.Length == 1 && InputState.KeyPressed(presssed[0]))
                    {
                        this.DialogueResult = currentScreenID + "|" + text + "|" + presssed[0].ToString();
                        this.controller.SetHiddenScreen("DialogueScreen");
                        this.controller.SetActiveScreen(currentScreenID);            
                    }
                }
                else
                {
                    if (InputState.KeyPressed(Keys.Up) || InputState.ButtonPressedOnAnyGamePad(Buttons.DPadUp))
                    {
                        currentIndex = (currentIndex == 0) ? options.Length - 1 : currentIndex - 1;
                    }
                    if (InputState.KeyPressed(Keys.Down) || InputState.ButtonPressedOnAnyGamePad(Buttons.DPadDown))
                    {
                        currentIndex = (currentIndex = currentIndex + 1) % options.Length;
                    }
                    if (InputState.KeyPressed(Keys.Enter) || InputState.ButtonPressedOnAnyGamePad(Buttons.A))
                    {
                        this.DialogueResult = currentScreenID + "|" + text + "|" + options[currentIndex];
                        this.controller.SetHiddenScreen("DialogueScreen");
                        this.controller.SetActiveScreen(currentScreenID);
                    }
                }
            }

            if (currentIndex != menuIndex)
            {
                AudioManager.PlaySoundEffect(Game1.MENU_CLICK, Game1.MENU_CLICK_VOLUME);
            }
        }

        public override void Draw(GameTime gameTime, GraphicsDevice graphics, SpriteBatch spriteBatch, SpriteFont spriteFont)
        {
            Rectangle target = CalculatePositionAndSize(spriteFont);

            graphics.Clear(ClearOptions.Target, backGroundColor, 0, 0, new Rectangle[]{ target } );
            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Deferred, SaveStateMode.SaveState);

            float textXpos = target.X + horizontalPadding;
            float textYpos = target.Y + verticalPadding;

            spriteBatch.DrawString(spriteFont, text, new Vector2(textXpos, textYpos), textColor, 0, Vector2.Zero, textScale, SpriteEffects.None, 0);

            float optionsYpos = textYpos + (spriteFont.MeasureString(text).Y * textScale);
            
            optionsYpos += spaceFromTextToOptions;
            
            float optionsHeight = (options.Length > 0)  ? spriteFont.MeasureString(options[0]).Y * optionsScale : 0;
            
            for (int i = 0; i < options.Length; i++)
            {
                Color itemColor = i == currentIndex ? optionsSelectedColor : optionsColor;
                                    
                float optionsXpos = GUIhelper.GetHorisontalCenter(options[i], spriteFont, Game, optionsScale);
                spriteBatch.DrawString(spriteFont, options[i], new Vector2(optionsXpos, optionsYpos), itemColor, 0, 
                    Vector2.Zero, optionsScale, SpriteEffects.None, 0);

                optionsYpos += spaceBetweenOptions + optionsHeight;
            }

            Rectangle left = new Rectangle(target.X - frameSize, target.Y - frameSize, frameSize, target.Height + (2 * frameSize));
            Rectangle right = new Rectangle(target.X + target.Width, target.Y - frameSize, frameSize, target.Height + (2 * frameSize));
            Rectangle up = new Rectangle(target.X - frameSize, target.Y - frameSize, target.Width + (2 * frameSize), frameSize);
            Rectangle down = new Rectangle(target.X - frameSize, target.Y + target.Height, target.Width + (2 * frameSize), frameSize);
            Rectangle right3d = new Rectangle(right.X + frameSize, right.Y + frameSize, right.Width, right.Height);
            Rectangle down3d = new Rectangle(down.X + frameSize, down.Y + frameSize, down.Width, down.Height);

            spriteBatch.Draw(texture, left, frameColor);
            spriteBatch.Draw(texture, right, frameColor);
            spriteBatch.Draw(texture, down, frameColor);
            spriteBatch.Draw(texture, up, frameColor);
            spriteBatch.Draw(texture, right3d, frameColor);
            spriteBatch.Draw(texture, down3d, frameColor);

            spriteBatch.End();
        }

        public void SetDialogueData(string text, string[] options, string currentScreenID, int startIndex)
        {
            this.text = text;
            this.options = options;
            this.currentScreenID = currentScreenID;
            this.DialogueResult = "";
            this.currentIndex = startIndex;
        }

        private Rectangle CalculatePositionAndSize(SpriteFont spriteFont)
        {
            float width = (spriteFont.MeasureString(text).X * textScale) + (2 * horizontalPadding);
            float x = (Game.Window.ClientBounds.Width - width) / 2;

            float optionsHeight = (options.Length != 0) ? ((options.Length - 1) * spaceBetweenOptions) +
                (options.Length * (spriteFont.MeasureString(options[0]).Y * optionsScale)) : 0;

            float height = (2 * verticalPadding) + optionsHeight + (spriteFont.MeasureString(text).Y * textScale) + spaceFromTextToOptions;

            float y = (Game.Window.ClientBounds.Height - height) / 2;

            return new Rectangle((int)x, (int)y, (int)width, (int)height);
        }
    }
}
