using System;
using System.Collections.Generic;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Audio;
using Microsoft.Xna.Framework.Content;
using Microsoft.Xna.Framework.GamerServices;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using Microsoft.Xna.Framework.Net;
using Microsoft.Xna.Framework.Storage;
using GameEngine.ScreenSystem;
using GamePrototype.GUI.Screens;

namespace GamePrototype
{
    public class Game1 : Game
    {
        private GraphicsDeviceManager graphics;
        private SpriteBatch spriteBatch;
        private ScreenController screenController;
        private SpriteFont spriteFont;
        
        private const int DEFAULT_RESOLUTION = 2;
        private const bool DEFAULT_FULLSCREEN = true;

        public static readonly string MENU_CLICK = "Audio/menuclick";
        public static readonly string MUSIC_MENU = "Audio/Music/menu-MRDC-Timecode";
        public static readonly string MUSIC_DEATH = "Audio/Music/death-Seismic Anamoly-Always on My Mind";
        public static readonly float MENU_CLICK_VOLUME = 0.2f;

        public bool FullScreen { get; set; }
        private int[] screenresWidths  = new int[] { 640, 800, 1024, 1280, 1280, 1400, 1650, 1920 };
        private int[] screenresHeights = new int[] { 480, 600, 768,  720,  1024, 1050, 1080, 1080 };
        private int screenWidth;
        private int screenHeight;
        private int screenResolution;
        public int ScreenResolution
        {
            get { return this.screenResolution; }
            set
            {
                this.screenResolution = value;
                this.screenWidth = this.screenresWidths[value];
                this.screenHeight = this.screenresHeights[value];
            }
        }

        public Game1(bool runDebug, string[] args)
        {
            graphics = new GraphicsDeviceManager(this);
            Content.RootDirectory = "Content";

            screenController = new ScreenController();

            if (runDebug)
            {
                switch (args[0])
                {
                    case "jan":
                        screenController.Screens.Add("JanDebugFraHelvede", new JanDebugFraHelvede("JanDebugFraHelvede", screenController, this));
                        break;
                    case "leinir":
                        screenController.Screens.Add("LeinirDebug", new LeninDebug("LeinirDebug", screenController, this));
                        break;
                    case "row":
                        // Screen
                        //screenController.Screens.Add(new RowDebug("RowDebug", screenController, this));
                        break;
                    default:
                        screenController.Screens.Add("DebuggingScreen", new DebuggingScreen("DebuggingScreen", screenController, this, args));
                        break;
                }
            }
            else
            {
                screenController.Screens.Add("MainMenuScreen", new MainMenuScreen("MainMenuScreen", screenController, this));
                screenController.Screens.Add("ConfigureControlsScreen", new ConfigureControlsScreen("ConfigureControlsScreen", screenController, this));
                screenController.Screens.Add("ConfigureGraphicsOptionsScreen", new ConfigureGraphicsOptionsScreen("ConfigureGraphicsOptionsScreen", screenController, this));
                screenController.Screens.Add("InGameScreen", new InGameScreen("InGameScreen", screenController, this));
                screenController.Screens.Add("SetupGameScreen", new SetupGameScreen("SetupGameScreen", screenController, this));
                screenController.Screens.Add("DialogueScreen", new DialogueScreen("DialogueScreen", screenController, this));
                screenController.Screens.Add("StatusScreen", new StatusScreen("StatusScreen", screenController, this));
                screenController.SetActiveScreen("MainMenuScreen");
            }

            // Set up graphics options
            FullScreen = DEFAULT_FULLSCREEN;
            ScreenResolution = DEFAULT_RESOLUTION;
            // End set up graphics options

            graphics.PreparingDeviceSettings += new EventHandler<PreparingDeviceSettingsEventArgs>(graphics_PreparingDeviceSettings);
            graphics.IsFullScreen = FullScreen;
            graphics.SynchronizeWithVerticalRetrace = true;
            this.Window.Title = "ORZAM";
            this.IsFixedTimeStep = false;
        }
        
        protected override void Initialize()
        {
            base.Initialize();
        }

        protected override void LoadContent()
        {
            spriteBatch = new SpriteBatch(GraphicsDevice);
            screenController.LoadAllScreens(graphics);
            spriteFont = Content.Load<SpriteFont>("HUD/Arial");

            AudioManager.LoadSoundEffect(MENU_CLICK, Content);

            AudioManager.LoadSongs(Content, MUSIC_MENU, MUSIC_DEATH);
            AudioManager.PlayMusic(MUSIC_MENU);
        }

        protected override void UnloadContent()
        {
            screenController.UnloadAllScreens();
        }

        protected override void Update(GameTime gameTime)
        {
            InputState.Update();
            
            if (InputState.IsKeyDown(Keys.LeftAlt) && InputState.KeyPressed(Keys.Enter))
            {
                graphics.ToggleFullScreen();
            }

            screenController.Update(gameTime);

            base.Update(gameTime);
        }

        protected override void Draw(GameTime gameTime)
        {
            screenController.Draw(gameTime, graphics.GraphicsDevice, spriteBatch, spriteFont);
            base.Draw(gameTime);
        }

        public void ReinitializeGraphics()
        {
            graphics.IsFullScreen = FullScreen;
            graphics.PreferredBackBufferWidth = screenWidth;
            graphics.PreferredBackBufferHeight = screenHeight;
            graphics.ApplyChanges();
        }

        void graphics_PreparingDeviceSettings(object sender, PreparingDeviceSettingsEventArgs e)
        {
            DisplayMode displayMode = GraphicsAdapter.DefaultAdapter.CurrentDisplayMode;
            e.GraphicsDeviceInformation.PresentationParameters.BackBufferFormat = displayMode.Format;
            e.GraphicsDeviceInformation.PresentationParameters.BackBufferWidth = screenresWidths[screenResolution]; //graphics.IsFullScreen ? displayMode.Width : screenresWidths[screenResolution];
            e.GraphicsDeviceInformation.PresentationParameters.BackBufferHeight = screenresHeights[screenResolution]; //graphics.IsFullScreen ? displayMode.Height : screenresHeights[screenResolution];
        }
    }
}
