﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.Cameras;
using Microsoft.Xna.Framework;
using GameEngine.Helpers;
using GameEngine.LevelGeneration;
using GamePrototype.GUI;
using Microsoft.Xna.Framework.Input;
using GameEngine.ScreenSystem;
using GameEngine.Xml;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Content;
using GameEngine;
using GameEngine.Physics;
using GameEngine.TextureGeneration;
using GamePrototype.GUI.Screens;

namespace GamePrototype.GameLogic
{
    public class GameWrapper : IDisposable
    {
        // Constants
        private readonly Color BACKGROUND_COLOR = Color.Black;
        public const int WORLD_MAX_COLLIDE_DIST = 2;
        
        // Properties and variables
        public CollisionSimulator CollisionSimulator { get; set; }
        public Player[] Players { get; set; }
        private ContentManager content;
        private bool horizontalSplit, verticalSplit;
        private LevelGenerator levelGenerator;
        private ChaseCamera[] cameras;
        public ZombieController enemyController;
        private Keys[,] playerControls;
        private int playerCount;
        private DrawInfo drawInfo = new DrawInfo();
        private LevelRenderer render;
        private Point[] playerAbsoluteTilePositions;
        private RenderTarget2D[] targets;
        private DeferredRenderer[] renderers;
        private int difficulty;
        private int currentLevel;
        public LevelTileNode goalNode;
        private ScreenController controller;
        private BearSoundsEmitter bearSoundsEmitter;
        private Game game;
        private GeometryDrawer geomDrawer;
        private Texture2D texture;
        
        public FragmentEmitter fragments = new FragmentEmitter(200);
        public TraceEmitter tracer = new TraceEmitter(100);
        public BloodEmitter blood = new BloodEmitter(200);

        private const int AMMO_RESPAWN_1 = 60;
        private const int AMMO_RESPAWN_2 = 90;
        private const int AMMO_RESPAWN_3 = -1;

        private const int HEALTH_RESPAWN_1 = 60;
        private const int HEALTH_RESPAWN_2 = 90;
        private const int HEALTH_RESPAWN_3 = -1;

        private const int WEAPON_RESPAWN_1 = 60;
        private const int WEAPON_RESPAWN_2 = 180;
        private const int WEAPON_RESPAWN_3 = -1;

        private float SPAWNPOINTRADIUS = 35f;
        private float SPAWNPOINTWEAPONSRADIUS = 80f;

        int AMMO_PER_SPAWN = 1;
        int HEALTH_PER_SPAWN = 20;

        private int cannonAmount;
        float halfTile;
        ParticleRenderer particles = new ParticleRenderer();
        private bool hasLost = false;
        private bool hasLostAndDrawedIt = false;

        ~GameWrapper()
        {
            Dispose();
        }

        public void Dispose()
        {
            if (targets != null)
            {
                for (int i = 0; i < targets.Length; i++)
                {
                    if (targets[i] != null)
                        targets[i].Dispose();
                    if (renderers[i] != null)
                        renderers[i].Dispose();

                    renderers[i] = null;
                    targets[i] = null;

                }
            }
            GC.SuppressFinalize(this);
        }

        SpawnPointNode.SpawnType[] possibleSpawnTypes;
        WeaponType[] possibleWeaponTypes;

        public ControllerType[] PlayerControllerTypes;

        public GameWrapper(ContentManager content, ControllerType[] joinedPlayers, int difficulty, Keys[,] playerControls, LevelGenerator levelGenerator, int currentLevel, ScreenController controller, Game game)
        {
            this.content = content;
            this.playerControls = playerControls;
            this.PlayerControllerTypes = joinedPlayers;
            this.playerCount = joinedPlayers.Length;
            this.difficulty = difficulty;
            this.levelGenerator = levelGenerator;
            this.currentLevel = currentLevel;
            this.controller = controller;
            this.game = game;

            Players = new Player[playerCount];
            cameras = new ChaseCamera[playerCount];
            targets = new RenderTarget2D[playerCount];
            playerAbsoluteTilePositions = new Point[playerCount];
            render = new LevelRenderer();
            renderers = new DeferredRenderer[playerCount];
            bearSoundsEmitter = new BearSoundsEmitter();
            WeaponFactory.LoadAllWeapons(content);

            possibleSpawnTypes = (SpawnPointNode.SpawnType[])Enum.GetValues(typeof(SpawnPointNode.SpawnType));
            possibleWeaponTypes = (WeaponType[])Enum.GetValues(typeof(WeaponType));
        }
        
        PrimitivesDrawer primDrawer;

        public void Initialize(GraphicsDevice graphics)
        {
            hasLost = false;
            wonGame = false;
            hasLostAndDrawedIt = false;
            graphics.DeviceReset += new EventHandler(graphics_DeviceReset);

            texture = new Texture2D(graphics, 1, 1, 1, TextureUsage.None, SurfaceFormat.Color);
            texture.SetData<Color>(new Color[] { Color.White });

            render.Initialize(graphics, content, levelGenerator);

            // Get goal tile
            goalNode = levelGenerator.PointsOfInterest[2].SceneNode;
 
            primDrawer = new PrimitivesDrawer(content.Load<Effect>("Hud/primitives"), graphics);
            AudioManager.LoadSoundEffect(content, "Audio/radarsweep");
            AudioManager.LoadSoundEffect(content, "Audio/teleport_activate");

            AudioManager.LoadSoundEffect(content, "Audio/floomp");
            AudioManager.LoadSoundEffect(content, "Audio/whack");
            foreach (var sound in ZombieBear.SuperSounds)
            {
                AudioManager.LoadSoundEffect(content, sound);
            }


            LevelTileNode startNode = (LevelTileNode)levelGenerator.PointsOfInterest[0].SceneNode;
            
            halfTile = (float)TilesetManager.CurrentTileset.TileSize / 2f;

            // Collision
            CollisionSimulator = new CollisionSimulator(render.Level);
            CollisionSimulator.OnCollision += new OnCollision(collisionSimulator_OnCollision);

            // Make sure we have enough spawnnodes
            for (int i = 0; i < render.Level.GetLength(0); i++)
            {
                for (int j = 0; j < render.Level.GetLength(1); j++)
                {
                    LevelTile tile = render.Level[i, j];
                    if(tile == null || tile.SceneNode == null)
                        continue;
                    MakeSpawnPointsInTile(tile.SceneNode as LevelTileNode, graphics);
                }
            }

            for (int i = 0; i < playerCount; i++)
            {
                Players[i] = new Player("Player " + i, this, content, graphics, playerControls, i);

                SpawnPointNode spawnPoint = startNode.GetGoodSpawnPoint(SpawnPointNode.SpawnType.Player);
                Players[i].Position = spawnPoint.Position;
                
                // Weapons
                Players[i].TryGiveWeapon(WeaponFactory.GetWeapon(WeaponType.Beretta));
                //Players[i].TryGiveWeapon(WeaponFactory.GetWeapon(WeaponType.Garand));
                

                // light
                PointLightNode pln = new PointLightNode("test") { Position = new Vector3(0, TilesetManager.CurrentTileset.Height, 0), LightRadius = 500, LightIntensity = 1 };
                Players[i].ChildNodes.Add(pln);

                Players[i].InitializeWeaponModels(content, graphics);

                startNode.AddDynamicNode(Players[i]);


                // camera
                cameras[i] = new ChaseCamera(MathHelper.PiOver2, Players[i].WorldPosition, Vector3.Zero, Vector3.Up, graphics.Viewport, 1, 2000, Vector3.Zero, Vector3.UnitZ);
                // See RunIntroAnimation function for the in-game desiredOffset value
                cameras[i].DesiredPositionOffset = new Vector3(0f, 140f, 40f); 
                cameras[i].Stiffness = 3000f;
                runIntroAnimation = true;
                cameras[i].Reset();
            }

            // Enemies
            FixZombies(graphics, startNode);
            enemyController = new ZombieController(difficulty, Players, CollisionSimulator, render.Level, render.LevelOffset, content, graphics, this);
            enemyController.Initialize();
            // Renderer
            drawInfo.GraphicsDevice = graphics;
            drawInfo.AllCameras = cameras;
            MakeTargets(graphics);
            render.Update(new GameTime(), drawInfo);

            GameStatistics.Reset(playerCount);

            // fix camera NaN bug
            for (int i = 0; i < cameras.Length; i++)
            {
                cameras[i].Reset();
                cameras[i].Position = Players[i].WorldPosition;
            }
/*            foreach (var camera in cameras)
            {
            }*/

            blood.Load(content, graphics);

            particles.Load(content, graphics);

            bearSoundsEmitter.LoadSounds(content);
            geomDrawer = new GeometryDrawer();
            geomDrawer.Load(content, graphics);
        }

        void graphics_DeviceReset(object sender, EventArgs e)
        {
            for (int i = 0; i < targets.Length; i++)
            {
                if(renderers[i] != null)
                    renderers[i].Dispose();
                if(targets[i] != null)
                    targets[i].Dispose();
            }
            MakeTargets(sender as GraphicsDevice);
        }

        private void FixZombies(GraphicsDevice graphics, LevelTileNode startNode)
        {
            Model zombieModel = content.Load<Model>(@"Models\zombiebear\deadbear");
            ZombieBear.FixZombieModel(zombieModel);
        }

        int updatesPrDraw = 0;

        bool wonGame = false;

        public void Update(GameTime gameTime)
        {
            bearSoundsEmitter.Update(gameTime);

            ModelNode.UpdatedAnimationBears.Clear(); ;
            UpdateInput(gameTime);

                for (int i = 0; i < playerCount; i++)
                {
                    LevelTile t = ((LevelTileNode)Players[i].Parent).Tile;
                    playerAbsoluteTilePositions[i] = new Point(t.X + render.LevelOffset.X, t.Y + render.LevelOffset.Y);

                    if (runIntroAnimation)
                    {
                        RunIntroAnimation(gameTime, cameras[i]);
                    }
                    
                    if ((InputState.KeyPressed(Keys.Enter) || InputState.ButtonPressedOnAnyGamePad(Buttons.Start)) && wonGame)
                    {
                        StatusScreen statusScreen = (StatusScreen)controller.GetScreen("StatusScreen");
                        statusScreen.CurrentLevel++;
                        statusScreen.Difficulty = this.difficulty;
                        statusScreen.PlayerCount = this.playerCount;
                        statusScreen.Prepare();

                        this.controller.SetHiddenScreen("InGameScreen");
                        this.controller.SetActiveScreen("StatusScreen");
                        AudioManager.PlayMusic(Game1.MUSIC_MENU);
                        this.Dispose();
                        break;
                    }
                    
                    if ((Players[i].Parent) == goalNode && !wonGame) // Did a player enter the goal node?
                    {
                        wonGame = true;    
                    }
                    
                    if (hasLost && hasLostAndDrawedIt && (InputState.KeyPressed(Keys.Enter) || InputState.ButtonPressedOnAnyGamePad(Buttons.Start)))
                    {
                        StatusScreen statusScreen = (StatusScreen)controller.GetScreen("StatusScreen");
                        statusScreen.CurrentLevel = currentLevel;
                        statusScreen.Difficulty = this.difficulty;
                        statusScreen.PlayerCount = this.playerCount;

                        this.controller.SetHiddenScreen("InGameScreen");
                        this.controller.SetActiveScreen("StatusScreen");
                        this.Dispose();
                    }

                    if (Players[i].CurrentHealth <= 0)
                    {
                        if (!hasLost)
                            AudioManager.PlayMusic(Game1.MUSIC_DEATH);

                        hasLost = true;
                    }
                }

                if (hasLost)
                    return;

                CollisionSimulator.Update(gameTime, playerAbsoluteTilePositions, WORLD_MAX_COLLIDE_DIST);

                render.Update(gameTime, drawInfo);

                for (int i = 0; i < playerCount; i++)
                {
                    if (!runIntroAnimation)
                    {
                        cameras[i].DesiredPositionOffset = ((LevelTileNode)Players[i].Parent).CameraOffset;
                    }
                    cameras[i].ChaseDirection = Players[i].Direction;
                    cameras[i].ChasePosition = Players[i].WorldPosition;
                    cameras[i].UpdateCamera(gameTime);
                }

                tracer.Update(gameTime);
                fragments.Update(gameTime);
                blood.Update(gameTime);


                enemyController.Update(gameTime);
                updatesPrDraw++;

            
        }

        float time = 0f;
        float frames = 0, fps = 0;

        public void Draw(GameTime gameTime, GraphicsDevice graphics, SpriteBatch spriteBatch, SpriteFont spriteFont)
        {
            if (wonGame)
            {
                GameStatistics.Draw(gameTime, graphics, spriteBatch, spriteFont, texture);
                return;
            }

            int modelCount = 0;
            
            if (!hasLost)
            {
                int drawnTiles = 0;

                for (int i = 0; i < playerCount; i++)
                {
                    drawInfo.Camera = cameras[i];

                    renderers[i].BeginRender(drawInfo);

                    Point start = playerAbsoluteTilePositions[i];

                    Point tileCoords;
                    int x = Math.Max(start.X - GameWrapper.WORLD_MAX_COLLIDE_DIST, 0);
                    int ex = Math.Min(start.X + GameWrapper.WORLD_MAX_COLLIDE_DIST, render.Level.GetLength(0) - 1);

                    int y = Math.Max(start.Y - GameWrapper.WORLD_MAX_COLLIDE_DIST, 0);
                    int ey = Math.Min(start.Y + GameWrapper.WORLD_MAX_COLLIDE_DIST, render.Level.GetLength(1) - 1);

                    for (tileCoords.X = x; tileCoords.X <= ex; tileCoords.X++)
                    {
                        for (tileCoords.Y = y; tileCoords.Y <= ey; tileCoords.Y++)
                        {
                            LevelTile tile = render.Level[tileCoords.X, tileCoords.Y];



                            if (tile != null)
                            {
                                BoundingBox bb = new BoundingBox(new Vector3(-halfTile, 0, -halfTile) + tile.SceneNode.Position, new Vector3(halfTile, TilesetManager.CurrentTileset.Height, halfTile) + tile.SceneNode.Position);
                                if (drawInfo.Camera.Frustrum.Intersects(bb))
                                {
                                    tile.SceneNode.Draw(gameTime, drawInfo);
                                    drawnTiles++;
                                }
                            }
                        }
                    }
                    modelCount = drawInfo.Models.Count;


                    geomDrawer.Draw(drawInfo);



                    particles.DrawAsLines(drawInfo, tracer, Color.Red, 50, 0.5f);
                    particles.DrawAsLines(drawInfo, fragments, Color.Wheat, 20, 0.5f);
                    particles.DrawAsPoints(drawInfo, blood, Color.DarkRed, 3f, 6f, 0f);

                    blood.DrawSplatter(drawInfo);


                    //renderers[i].TargetEndResult = DeferredRenderer.TargetOutput.Color;
                    renderers[i].EndRender(targets[i]);
                    drawInfo.PointLights.Clear();


                    Players[i].DrawPlayerHUD(gameTime, graphics, spriteBatch, spriteFont);
                
                }
                drawInfo.UpdatePointLights.Clear();

               
                graphics.SetRenderTarget(0, null);
                // set back to main render target

                spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);

                int width = graphics.Viewport.Width;
                int height = graphics.Viewport.Height;
                
                if(horizontalSplit)
                    width /= 2;
                
                if(verticalSplit)
                    height /= 2;

                for (int i = 0; i < targets.Length; i++)
                {
                    Texture2D t = targets[i].GetTexture();
                    spriteBatch.Draw(t, new Rectangle((i % 2) * width, i > 1 ? height : 0, width, height), Color.White);
                }
                
                frames++;
                time += (float)gameTime.ElapsedRealTime.TotalSeconds;
                if (time >= 1f)
                {
                    fps = frames / time;
                    time = frames = 0;
                }
                //spriteBatch.DrawString(spriteFont, Math.Round(fps) + "fps " + ModelNode.UpdatedAnimationBears.Count + " - " + modelCount + " - " + drawnTiles + " - " + updatesPrDraw, new Vector2(100, 100), Color.White);
                //spriteBatch.DrawString(spriteFont, Math.Round(fps) + "fps ", new Vector2(20, 20), Color.White);
                updatesPrDraw = 0;

                // Draw vertical screen splitter
                if (playerCount > 1)
                {
                    spriteBatch.Draw(texture, new Rectangle(graphics.Viewport.Width / 2 - 1, 0, 2, graphics.Viewport.Height), Color.CornflowerBlue);
                }

                spriteBatch.End();
            }
            else
            {
                graphics.Clear(Color.Black);
                spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Deferred, SaveStateMode.SaveState);

                string lostText = playerCount == 1 ? "Too bad! Try again. Press Enter or GamePad Start to replay current level." :
                                                     "Too bad! This is about teamwork! Press Enter or GamePad Start to replay current level.";
                float lostScale = 1.2f;

                float lostX = GUIhelper.GetHorisontalCenter(lostText, spriteFont, this.game, lostScale);
                float lostY = GUIhelper.GetVerticalCenter(lostText, spriteFont, this.game, lostScale);

                spriteBatch.DrawString(spriteFont, lostText, new Vector2(lostX, lostY), Color.CornflowerBlue, 0,
                Vector2.Zero, lostScale, SpriteEffects.None, 0);

                spriteBatch.End();
                hasLostAndDrawedIt = true;
            }
        }


        private float GetRespawnTime(SpawnPointNode sNode) 
        {
            switch (sNode.SpawnPointType)
            {
                case SpawnPointNode.SpawnType.Ammunition:
                    if (difficulty == 1)
                        return AMMO_RESPAWN_1;
                    else if (difficulty == 2)
                        return AMMO_RESPAWN_2;
                    else if (difficulty == 3)
                        return AMMO_RESPAWN_3;
                    break;
                case SpawnPointNode.SpawnType.Health:
                    if (difficulty == 1)
                        return HEALTH_RESPAWN_1;
                    else if (difficulty == 2)
                        return HEALTH_RESPAWN_2;
                    else if (difficulty == 3)
                        return HEALTH_RESPAWN_3;
                    break;
                case SpawnPointNode.SpawnType.Player:
                    return 0;
                case SpawnPointNode.SpawnType.Weapon:
                    if (difficulty == 1)
                        return WEAPON_RESPAWN_1;
                    else if (difficulty == 2)
                        return WEAPON_RESPAWN_2;
                    else if (difficulty == 3)
                        return WEAPON_RESPAWN_3;
                    break;
                case SpawnPointNode.SpawnType.Zombie:
                    return 0;
                default:
                    break;
            }
            return -1;
        }

        public void ThrowItemOnFloor(SpawnPointNode.SpawnType spawnType, Object item, DynamicCircleNode aroundNode)
        {
            BoundingCircle aroundCircle = aroundNode.GetCollisionCircle();

            CollisionRay testRay = new CollisionRay(aroundCircle.Center, Vector2.UnitX);

            float minDist = aroundNode.radius + SPAWNPOINTRADIUS * 2 + 10f;


            // Default to spawning 10 units away from the player along the positive X-axis
            Vector2 spawnPosition = Vector2.Zero;
            spawnPosition.X += 10f;

            // Shoot out some rays to find a more suitable startpoint
            for (int i = 0; i < 16; i++)
            {
                float angle = (float)(Math.PI * 2 * i / 16);
                testRay.Direction = new Vector2((float)Math.Cos(angle), (float)Math.Sin(angle));

                if (CollisionSimulator.GetClosestPolygonCollisionDistance(testRay, (aroundNode.Parent as LevelTileNode).Tile) > minDist)
                {
                    spawnPosition = testRay.Position + testRay.Direction * (minDist - SPAWNPOINTRADIUS);
                    break;
                }
            }

            SpawnPointNode newSpawn = new SpawnPointNode("thrownSpawnPoint" + item.ToString() + "From" + aroundNode.Name);
            newSpawn.Position = new Vector3(spawnPosition.X, 0, spawnPosition.Y);
            newSpawn.SpawnPointType = spawnType;
            newSpawn.radius = spawnType == SpawnPointNode.SpawnType.Weapon ? SPAWNPOINTWEAPONSRADIUS : SPAWNPOINTRADIUS;
            newSpawn.GiveObject(item);
            switch (spawnType)
            {
                case SpawnPointNode.SpawnType.Ammunition:
                case SpawnPointNode.SpawnType.BigAmmo:
                    newSpawn.ItemIntValue = 1;
                    break;
                case SpawnPointNode.SpawnType.Health:
                    break;
                case SpawnPointNode.SpawnType.Player:
                    break;
                case SpawnPointNode.SpawnType.Weapon:
                case SpawnPointNode.SpawnType.BigWeapon:
                    newSpawn.ItemIntValue = (int)(item as Weapon).Type;
                    break;
                case SpawnPointNode.SpawnType.Zombie:
                    break;
                default:
                    break;
            }
            SetSpawnPointModel(newSpawn, drawInfo.GraphicsDevice);
            (aroundNode.Parent as LevelTileNode).AddSpawnPointNode(newSpawn);
        }


        private bool GetRandomSpawnPointPosition(List<Vector2> list, LevelTileNode tileNode, out Vector2 result)
        {

            BoundingCircle circle = new BoundingCircle(Vector2.Zero, SPAWNPOINTRADIUS);
            Vector2 dummy;
            while (list.Count > 0)
            {
                int index = Util.Random.Next(list.Count);
                result = list[index];
                list.RemoveAt(index);
                circle.Center = result;

                bool ok = true;
                for (int i = 0; i < tileNode.DynamicCount; i++)
                {
                    if (tileNode[i].GetCollisionCircle().Intersects(circle, out dummy))
                    {
                        ok = false;
                        break;
                    }
                }
                if (ok)
                    return true;
            }

            result = Vector2.Zero;
            return false;
        }

        private void MakeSpawnPointsInTile(LevelTileNode tileNode, GraphicsDevice graphics)
        {
            for (int i = 0; i < tileNode.SpawnPoints.Count; i++)
            {
                InitializeSpawnPoint(tileNode.SpawnPoints[i], graphics);
            }
           
        }

        private void InitializeSpawnPoint(SpawnPointNode sNode, GraphicsDevice graphics)
        {
            
            
            sNode.RespawnTime = 0f;
            sNode.HasObject = true;
            if (sNode.SpawnPointType == SpawnPointNode.SpawnType.Random)
            {
                int randPercent = Util.Random.Next(100);
                if (randPercent < 1)
                    sNode.SpawnPointType = SpawnPointNode.SpawnType.BigZombie;
                else if (randPercent < 11)
                    sNode.SpawnPointType = SpawnPointNode.SpawnType.Health;
                else if (randPercent < 16)
                    sNode.SpawnPointType = SpawnPointNode.SpawnType.Zombie;
                else if (randPercent < 21)
                    sNode.SpawnPointType = SpawnPointNode.SpawnType.Ammunition;
                else
                    return;
                // sNode.SpawnPointType = SpawnPointNode.SpawnType.Zombie;
                //sNode.SpawnPointType = possibleSpawnTypes[1 + Util.Random.Next(possibleSpawnTypes.Length - 1)];
            }
            else
            {
                if (Util.Random.NextDouble() > sNode.SpawnLikelyhood)
                    return;
                
                if (sNode.SpawnPointType == SpawnPointNode.SpawnType.ZombieLauncher)
                {
                    cannonAmount++;
                    ZombieCannon cannon = new ZombieCannon("cannon1", 600, 0, this);
                    cannon.Position = sNode.Position;
                    cannon.ChildNodes.LoadRecursive(content, graphics);
                    ((LevelTileNode)sNode.Parent).AddDynamicNode(cannon);
                    return;
                }
            }
            switch (sNode.SpawnPointType)
            {
                case SpawnPointNode.SpawnType.Ammunition:
                    sNode.ItemIntValue = AMMO_PER_SPAWN;
                    break;
                case SpawnPointNode.SpawnType.Health:
                    sNode.ItemIntValue = HEALTH_PER_SPAWN;
                    break;
                case SpawnPointNode.SpawnType.Weapon:
                    WeaponType type;
                    int weapTypeRand = Util.Random.Next(3);
                    if(weapTypeRand == 0) 
                    {
                        type = WeaponType.Beretta;
                    }
                    else if (weapTypeRand == 1)
                    {
                        type = WeaponType.MiniShotgun;
                    }
                    else
                    {
                        type = WeaponType.Uzi;
                    }
                    sNode.ItemIntValue = (int)type;

                    sNode.radius = SPAWNPOINTWEAPONSRADIUS;
                    break;
                default:
                    break;
            }
            SetSpawnPointModel(sNode, graphics);
        }

        private void SetSpawnPointModel(SpawnPointNode spawnPoint, GraphicsDevice graphics)
        {
            ModelNode model = null;
            switch (spawnPoint.SpawnPointType)
            {
                case SpawnPointNode.SpawnType.Ammunition:
                    model = new ModelNode("AmmoModel");
                    model.Asset = @"models\pickups\ammoboxsmall_x";
                    break;
                case SpawnPointNode.SpawnType.BigAmmo:
                    model = new ModelNode("AmmoModelBig");
                    model.Asset = @"models\pickups\ammoboxbig_x";
                    break;
                case SpawnPointNode.SpawnType.Health:
                    model = new ModelNode("HealthModel");
                    model.Asset = @"models\pickups\medpack_x";
                    break;
                case SpawnPointNode.SpawnType.Weapon:
                case SpawnPointNode.SpawnType.BigWeapon:
                    WeaponType weaponType = (WeaponType)spawnPoint.ItemIntValue;
                    switch (weaponType)
                    {
                        case WeaponType.M4:
                            model = new ModelNode("M4Model");
                            model.Asset = @"models\weapons\weapon_m4_x";
                            break;
                        case WeaponType.Beretta:
                            model = new ModelNode("ColtModel");
                            model.Asset = @"models\weapons\weapon_colt_x";
                            break;
                        case WeaponType.MiniShotgun:
                            model = new ModelNode("WeaponModel");
                            model.Asset = @"models\weapons\minishotgun_x";
                            break;
                        case WeaponType.Garand:
                            model = new ModelNode("GarandModel");
                            model.Asset = @"models\weapons\rifle_x";
                            break;
                        case WeaponType.Uzi:
                            model = new ModelNode("WeaponModel");
                            model.Asset = @"models\weapons\weapon_uzi_x";
                            break;
                        case WeaponType.Minigun:
                            model = new ModelNode("MinigunModel");
                            model.Asset = @"models\weapons\minigun_x";
                            model.Scale = new Vector3(0.5f);
                            model.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitX, -MathHelper.PiOver2);
                            model.Position.Y = 40f;
                            break;
                        default:
                            model = new ModelNode("BrokenWeaponModel");
                            model.Asset = @"models\player\player";
                            break;
                    }
                    break;
                default:
                    break;
            }
            if (model != null)
            {
                //model.Scale = new Vector3(4);
                model.LoadContent(content, graphics);

                ModelNode happyModel = new ModelNode("HappyFace");
                happyModel.Asset = @"models\powerup_faces\powerup_happyface_x";
                happyModel.LoadContent(content, graphics);

                ModelNode waitingModel = new ModelNode("WaitingFace");
                waitingModel.Asset = @"models\powerup_faces\powerup_waitingface_x";
                waitingModel.LoadContent(content, graphics);

                ModelNode longtimeModel = new ModelNode("SadFace");
                longtimeModel.Asset = @"models\powerup_faces\powerup_sadface_x";
                longtimeModel.LoadContent(content, graphics);

                spawnPoint.SetModels(model, happyModel, waitingModel, longtimeModel);
            }

        }

        private void GiveFromSpawnPoint(SpawnPointNode fromNode, Player toPlayer)
        {
            switch (fromNode.SpawnPointType)
            {
                case SpawnPointNode.SpawnType.Ammunition:
                    //toPlayer.GiveAmmo(fromNode.ItemIntValue); TODO: Light og heavy ammo?
                    toPlayer.GiveAmmo(fromNode.ItemIntValue, AmmoType.Light);
                    return;
                case SpawnPointNode.SpawnType.BigAmmo:
                    toPlayer.GiveAmmo(fromNode.ItemIntValue, AmmoType.Heavy);
                    return;
                case SpawnPointNode.SpawnType.Health:
                    if (toPlayer.TryGiveHealth(fromNode.ItemIntValue))
                        return;
                    break;
                case SpawnPointNode.SpawnType.Weapon:
                case SpawnPointNode.SpawnType.BigWeapon:
                    WeaponType weaponType = (WeaponType)fromNode.ItemIntValue;
                    if (toPlayer.WantsWeapon(weaponType))
                    {
                        Weapon weapon = WeaponFactory.GetWeapon(weaponType);
                        if (!toPlayer.TryGiveWeapon(weapon))
                        {
                            Weapon oldWeapon = toPlayer.ExchangeWithCurrentWeapon(weapon);
                            if(oldWeapon.Type == WeaponType.Minigun || oldWeapon.Type == WeaponType.Garand || oldWeapon.Type == WeaponType.M4)
                                ThrowItemOnFloor(SpawnPointNode.SpawnType.BigWeapon, oldWeapon, toPlayer);
                            else
                                ThrowItemOnFloor(SpawnPointNode.SpawnType.Weapon, oldWeapon, toPlayer);
                        }
                        return;
                    }
                    else
                    break;
                default:
                    break;
            }

            // Default to giving item back to the spawnpoint
            fromNode.GiveBackObject();
        }

        public void GiveSpecificItem(SpawnPointNode fromNode, Player toPlayer, Object item)
        {
            switch (fromNode.SpawnPointType)
            {
                case SpawnPointNode.SpawnType.Ammunition:
                    toPlayer.GiveAmmo(fromNode.ItemIntValue, AmmoType.Light);
                    return;
                case SpawnPointNode.SpawnType.BigAmmo:
                    toPlayer.GiveAmmo(fromNode.ItemIntValue, AmmoType.Heavy);
                    return;
                case SpawnPointNode.SpawnType.Weapon:
                case SpawnPointNode.SpawnType.BigWeapon:
                    Weapon weapon = item as Weapon;
                    if (toPlayer.WantsWeapon(weapon.Type))
                    {
                        if (!toPlayer.TryGiveWeapon(weapon))
                        {
                            Weapon oldWeapon = toPlayer.ExchangeWithCurrentWeapon(weapon);
                            ThrowItemOnFloor(SpawnPointNode.SpawnType.Weapon, oldWeapon, toPlayer);
                        }
                        return;
                    }
                    break;
                default:
                    break;
            }
            // If nothing else happens, give back the object
            fromNode.GiveObject(item);
        }

        void collisionSimulator_OnCollision(ObjectNode a, ObjectNode b)
        {
            if (a is Player && b is SpawnPointNode)
            {
                SpawnPointNode sNode = b as SpawnPointNode;
                Object givenObject;
                if (sNode.TryTakeObject(GetRespawnTime(sNode), out givenObject))
                {
                    if (givenObject == null)
                        GiveFromSpawnPoint(sNode, a as Player);
                    else
                        GiveSpecificItem(sNode, a as Player, givenObject);
                }
            }
        }

        public void UpdateInput(GameTime gameTime)
        {
            for (int i = 0; i < playerCount; i++)
            {
                int controlType = (int)PlayerControllerTypes[i];
                if (controlType < 2)
                {
                    for (int j = 0; j < playerControls.GetLength(1); j++)
                    {
                        if (InputState.IsKeyDown(playerControls[controlType, j]))
                        {
                            switch (j)
                            {
                                case 0:
                                    Players[i].MoveForward(gameTime);
                                    break;
                                case 1:
                                    Players[i].RotateRight(gameTime);
                                    break;
                                case 2:
                                    Players[i].MoveBackward(gameTime);
                                    break;
                                case 3:
                                    Players[i].RotateLeft(gameTime);
                                    break;
                                case 4:
                                    bool justTriggered = InputState.KeyPressed(playerControls[controlType, j]);
                                    Players[i].Fire(gameTime, drawInfo, justTriggered);
                                    break;
                                case 5:
                                    if (InputState.KeyPressed(playerControls[controlType, j]))
                                    {
                                        Players[i].Reload();
                                    }
                                    break;
                                case 6:
                                    if (InputState.KeyPressed(playerControls[controlType, j]))
                                    {
                                        Players[i].ChangeWeapon();
                                    }
                                    break;
                                case 7:
                                    if (InputState.KeyPressed(playerControls[controlType, j]))
                                        Players[i].WantsToPickup = true;
                                    break;
                                default:
                                    break;
                            }
                        }
                        else
                        {
                            switch (j)
                            {
                                case 7:
                                    if (InputState.KeyReleased(playerControls[controlType, j]))
                                        Players[i].WantsToPickup = false;
                                    break;
                                default:
                                    break;
                            }
                        }
                    }
                }
                else
                {
                    PlayerIndex pIndex = (PlayerIndex)(controlType - 2);

                    if (InputState.IsButtonDown(pIndex, Buttons.LeftThumbstickUp))
                        Players[i].MoveForward(gameTime);
                    if (InputState.IsButtonDown(pIndex, Buttons.LeftThumbstickDown))
                        Players[i].MoveBackward(gameTime);


                    float turnAmount = GamePad.GetState(pIndex).ThumbSticks.Left.X;
                    if (Math.Abs(turnAmount) > 0.1)
                        Players[i].Rotate(gameTime, -turnAmount);
                    //if (InputState.IsButtonDown(pIndex, Buttons.LeftThumbstickLeft))
                    //    Players[i].RotateLeft(gameTime);
                    //if (InputState.IsButtonDown(pIndex, Buttons.LeftThumbstickRight))
                    //    Players[i].RotateRight(gameTime);
                    if (!hasLost && InputState.IsButtonDown(pIndex, Buttons.A))
                    {
                        bool justTriggered = InputState.ButtonPressed(pIndex, Buttons.A);
                        Players[i].Fire(gameTime, drawInfo, justTriggered);
                    }

                    if (InputState.ButtonPressed(pIndex, Buttons.X))
                        Players[i].Reload();
                    if (InputState.ButtonPressed(pIndex, Buttons.B))
                        Players[i].ChangeWeapon();
                    if (InputState.ButtonPressed(pIndex, Buttons.Y))
                        Players[i].WantsToPickup = true;
                    else if (InputState.ButtonReleased(pIndex, Buttons.Y))
                        Players[i].WantsToPickup = false;
                }
                Players[i].DoMovement(gameTime);
            }
        }


        private void MakeTargets(GraphicsDevice device)
        {

            Point size = new Point(device.Viewport.Width, device.Viewport.Height);

            if (cameras.Length > 2)
            {
                size.Y /= 2;
                verticalSplit = true;
            }

            if (cameras.Length > 1)
            {
                horizontalSplit = true;
                size.X /= 2;
            }

            for (int i = 0; i < cameras.Length; i++)
            {
                if (cameras[i] != null)
                {
                    Viewport p = device.Viewport;
                    p.Width = size.X;
                    p.Height = size.Y;

                    cameras[i].Viewport = p;
                    targets[i] = new RenderTarget2D(device, size.X, size.Y, 1, device.DisplayMode.Format);
                    renderers[i] = new DeferredRenderer(size.X, size.Y);
                    renderers[i].LoadContent(device, content);
                }
                else
                    targets[i] = null;
            }
        }

        bool runIntroAnimation = true;
        bool playingRadarSound = false;
        float introAnimationLength = 3500;
        float introAnimationStarted = 0;

        private void RunIntroAnimation(GameTime gameTime, ChaseCamera camera)
        {
            introAnimationStarted += gameTime.ElapsedGameTime.Milliseconds;
            
            if (!playingRadarSound && introAnimationStarted > 1500)
            {
                AudioManager.PlaySoundEffect("Audio/radarsweep", 2.0f);
                playingRadarSound = true;
            }

            if (introAnimationStarted > introAnimationLength)
            {
                camera.DesiredPositionOffset = new Vector3(0, 260f, 100f);
                //camera.Reset();
                runIntroAnimation = false;
            }
        }

    }
}
