﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.Xml;
using GameEngine.Helpers;
using Microsoft.Xna.Framework;
using Microsoft.Xna.Framework.Graphics;
using Microsoft.Xna.Framework.Input;
using GameEngine;
using Microsoft.Xna.Framework.Content;
using GamePrototype.GUI;
using GameEngine.LevelGeneration;

namespace GamePrototype.GameLogic
{
    public class Player : DynamicCircleNode
    {
        private const int STARTING_HEALTH = 100;

        // Movement control
        private const float PLAYER_MAX_SPEED = 300f;
        private const float PLAYER_ACCELERATION_FACTOR = 1200f;
        private const float PLAYER_DEACCELERATION_FACTOR = 1200f;
        private const float PLAYER_TURNING_SPEED = MathHelper.Pi;

        private const int MAX_WEAPONS = 2;
        private const int MAX_LIGHT_AMMO = 15;
        private const int MAX_HEAVY_AMMO = 10;

        // Player HUD
        private const int HUD_TOP_SPACING = 10;
        private const int HUD_LEFT_SPACING = 10;
        private const int HUD_BOTTOM_SPACING = 10;
        private const int HUD_RIGHT_SPACING = 10;
        private const int HUD_WEAPON_AMMO_SPACING = 3;
        private const float HUD_SECONDARY_WEAPON_SCALE = 0.6f;
        private const int HUD_HEAVY_AMMO_SIZE = 15;
        private const int HUD_LIGHT_AMMO_SIZE = 10;
        private const int HUD_HEALTHBAR_WIDTH = 150;
        private const int HUD_HEALTHBAR_HEIGHT = 30;
        private const int HUD_HEALTHHEART_SIZE = 30;

        private readonly Color HUD_HEALTH_COLOR = Color.Red;
        private readonly Color HUD_HEALTH_OUTLINE_COLOR = Color.Red;
        private readonly Color HUD_AMMO_COLOR = Color.CornflowerBlue;
        private readonly Color HUD_AMMO_RELOAD_COLOR = Color.Red;
        private readonly Color HUD_SEC_AMMO_COLOR = new Color(185, 185, 185, 64);
        private readonly Color HUD_SEC_AMMO_EMPTY_COLOR = new Color(235, 235, 235, 64);
        private readonly Color HUD_AMMO_OUTLINE = new Color(230, 230, 230, 64);
        private readonly Color HUD_AMMO_ALMOST_OUTLINE = new Color(255, 255, 255, 64);
    


        // Properties and variables

        public Weapon[] Weapons { get; set; }
        public Weapon CurrentWeapon { get { return Weapons[currentWeaponIndex] != null ? Weapons[currentWeaponIndex] : null; } private set { } }
        private int currentWeaponIndex = 0;
        private int currentHealth;
        private Texture2D texture;
        private Texture2D damageTexture;
        private ContentManager content;
        private List<ObjectNode> hits;
        private float currentSpeed = 0;
        private MuzzleFlashLight muzzleFlash = new MuzzleFlashLight() { Color = new Color(255,255,127), LightIntensity = 5f, LightRadius = 500 };
        private int[] ammoStats;

        private ModelNode[] weaponModels;

        public bool ShowPickupMessage { get; set; }
        public WeaponType PickupType { get; set; }

        public int CurrentHealth { get { return currentHealth; } private set { } }
        
        private GameWrapper gameWrapper;
        private ModelNode playerModel;
        private Texture2D heavyAmmo;
        private Texture2D lightAmmo;
        private Texture2D healthHeart;
        private Texture2D healthBar;

        private Texture2D endTile;
        private Texture2D cornTile;
        private Texture2D hwTile;
        private Texture2D tintTile;
        private Texture2D crossTile;
        private Dictionary<TileType, Color[]> tileData;
        private List<LevelTile> visitedTiles = new List<LevelTile>();
        private LevelTile currentTile;
        private bool minimapUnclean = false;
        private int minimapDiagonalSize;
        private Texture2D miniMap;

        public bool WantsToPickup { get; set; }
        private int damageTaken = 0;

        private Keys[,] controls;
        private int playerNumber;

        public float TurningMultiplier { get; set; }

        public Vector3 Direction;
        int traceCount = 0;
        public Player(string id, GameWrapper gameWrapper, ContentManager content, GraphicsDevice graphics, Keys[,] controls, int playerNumber) : base(id)
        {
            this.gameWrapper = gameWrapper;
            this.content = content;
            this.controls = controls;
            this.playerNumber = playerNumber;

            currentHealth = STARTING_HEALTH;
            this.radius = 25;

            texture = new Texture2D(graphics, 1, 1, 1, TextureUsage.None, SurfaceFormat.Color);
            texture.SetData(new Color[]{ Color.White });
            damageTexture = content.Load<Texture2D>("HUD/damage");
            healthHeart = content.Load<Texture2D>("HUD/healthheart");
            healthBar = content.Load<Texture2D>("HUD/healthbar");

            // Minimap data
            miniMap = texture;
            endTile = content.Load<Texture2D>(@"HUD\Minimap\End");
            cornTile = content.Load<Texture2D>(@"HUD\Minimap\Corner");
            hwTile = content.Load<Texture2D>(@"HUD\Minimap\Hallway");
            tintTile = content.Load<Texture2D>(@"HUD\Minimap\Tintersection");
            crossTile = content.Load<Texture2D>(@"HUD\Minimap\Crossing");

            tileData = new Dictionary<TileType, Color[]>();
            Color[] tempData = new Color[endTile.Height * endTile.Width];

            endTile.GetData(tempData);
            tileData.Add(TileType.End, (Color[])tempData.Clone());

            cornTile.GetData(tempData);
            tileData.Add(TileType.Corner, (Color[])tempData.Clone());

            hwTile.GetData(tempData);
            tileData.Add(TileType.Hallway, (Color[])tempData.Clone());

            tintTile.GetData(tempData);
            tileData.Add(TileType.Tintersection, (Color[])tempData.Clone());

            crossTile.GetData(tempData);
            tileData.Add(TileType.Crossing, (Color[])tempData.Clone());
            // End minimap data

            // Model
            playerModel = new ModelNode("PlayerModel_" + id)
            {
                Asset = @"Models\New_Monkey_sigway\MonkeyHighPoly",
                Rotation = baseRotation,
                Scale = new Vector3(.5f, .5f, .5f),
                Position = new Vector3(0,10,0)
            };
            playerModel.LoadContent(content, graphics);
            ChildNodes.Add(playerModel);

            Weapons = new Weapon[MAX_WEAPONS];

            ammoStats = new int[2]; // Index 0 is light ammo, index 1 is heavy ammo 

            heavyAmmo = content.Load<Texture2D>(@"Hud\Weapons\heavyammo");
            lightAmmo = content.Load<Texture2D>(@"Hud\Weapons\lightammo");

            // TODO: starting ammo?
            ammoStats[0] = MAX_LIGHT_AMMO;
            ammoStats[1] = MAX_HEAVY_AMMO;
        }

        public int GetAmmoCount(AmmoType ammo)
        {
            return ammo == AmmoType.Light ? ammoStats[0] : ammoStats[1];
        }

        public void DecreaseAmmoCount(AmmoType ammo)
        {
            if (ammo == AmmoType.Light)
            {
                ammoStats[0]--;
            }
            else
            {
                ammoStats[1]--;
            }
        }

        public void InitializeWeaponModels(ContentManager content, GraphicsDevice graphics)
        {
            int numWeaponTypes = Enum.GetValues(typeof(WeaponType)).Length;
            weaponModels = new ModelNode[numWeaponTypes];

            ModelNode model;

            model = new ModelNode("M4Model");
            model.Asset = @"models\weapons\weapon_m4_x";
            model.LoadContent(content, graphics);
            model.Position = new Vector3(-42, 17 + 78, 14);
            model.Scale = new Vector3(1.8f);
            model.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitX, MathHelper.PiOver2 * 0.85f);
            weaponModels[(int)WeaponType.M4] = model;


            model = new ModelNode("ColtModel");
            model.Asset = @"models\weapons\weapon_colt_x";
            model.LoadContent(content, graphics);
            model.Position = new Vector3(-41, 17 + 77, 42);
            model.Scale = new Vector3(1.0f);
            weaponModels[(int)WeaponType.Beretta] = model;

            model = new ModelNode("WeaponModel");
            model.Asset = @"models\weapons\minishotgun_x";
            model.LoadContent(content, graphics);
            model.Position = new Vector3(-41, 17 + 78, 28);
            model.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitY, -MathHelper.PiOver2) * Quaternion.CreateFromAxisAngle(Vector3.UnitZ, -MathHelper.PiOver2 * 0.75f);
            model.Scale = new Vector3(1.3f);
            weaponModels[(int)WeaponType.MiniShotgun] = model;

            
            model = new ModelNode("GarandModel");
            model.Asset = @"models\weapons\rifle_x";
            model.LoadContent(content, graphics);
            model.Position = new Vector3(-41, 17 + 81, 12);
            model.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitX, MathHelper.PiOver2) * Quaternion.CreateFromAxisAngle(Vector3.UnitY, MathHelper.PiOver2 * 2);
            model.Scale = new Vector3(1.3f);
            weaponModels[(int)WeaponType.Garand] = model;

            model = new ModelNode("WeaponModel");
            model.Asset = @"models\weapons\weapon_uzi_x";
            model.LoadContent(content, graphics);
            model.Position = new Vector3(-42, 17 + 71, 42);
            model.Rotation = Quaternion.CreateFromAxisAngle(Vector3.UnitY, MathHelper.PiOver2 * 2);
            model.Scale = new Vector3(1.3f);
            weaponModels[(int)WeaponType.Uzi] = model;

            model = new ModelNode("WeaponModel");
            model.Asset = @"models\weapons\minigun_x";
            model.LoadContent(content, graphics);
            model.Position = new Vector3(0, 17 + 18, 48);
            weaponModels[(int)WeaponType.Minigun] = model;

            for (int i = 0; i < numWeaponTypes; i++)
            {
                if (weaponModels[i] == null)
                {
                    // SanityCheck - all null values must be set to broken model
                    model = new ModelNode("BrokenWeaponModel");
                    model.Asset = @"models\player\player";
                    model.Position = new Vector3(-42, 71, 42);
                    model.LoadContent(content, graphics);
                    weaponModels[i] = model;
                }

                weaponModels[i].IsVisible = false;
                playerModel.ChildNodes.Add(weaponModels[i]);
            }

            UpdateVisibleWeapon();
        }

        private void UpdateVisibleWeapon()
        {
            if (weaponModels == null)
                return;

            for (int i = 0; i < weaponModels.Length; i++)
            {
                weaponModels[i].IsVisible = false;
            }

            WeaponType t = Weapons[currentWeaponIndex].Type;

            weaponModels[(int)t].IsVisible = true;

            // Try to change to secondary weapon if it is empty
            if (CurrentWeapon.AmmoInClip == 0 && !CurrentWeapon.IsReloading)
            {
                if (CurrentWeapon.AmmoType == AmmoType.Light)
                {
                    if (ammoStats[0] == 0)
                    {
                        ChangeWeapon();
                    }
                }
                else
                {
                    if (ammoStats[1] == 0)
                    {
                        ChangeWeapon();
                    }
                }
            }
        }


        public void Move(GameTime dt, Vector2 movement)
        {
            movement.Normalize();
            Direction = new Vector3(movement.X, 0, movement.Y);
            MoveForward(dt);
        }

        float acceleration = 0;

        public void MoveForward(GameTime gameTime)
        {
            acceleration += gameTime.ElapsedSeconds() * PLAYER_ACCELERATION_FACTOR;
        }
        
        public void RotateRight(GameTime gameTime)
        {
            Rotation *= Quaternion.CreateFromAxisAngle(Vector3.Up, -PLAYER_TURNING_SPEED * gameTime.ElapsedSeconds());
        }

        public void Rotate(GameTime gameTime, float turnAmount)
        {
            Rotation *= Quaternion.CreateFromAxisAngle(Vector3.Up, PLAYER_TURNING_SPEED * gameTime.ElapsedSeconds() * turnAmount);
        }
        
        public void MoveBackward(GameTime gameTime)
        {
            acceleration -= gameTime.ElapsedSeconds() * PLAYER_ACCELERATION_FACTOR;
        }


        public void DoMovement(GameTime gameTime)
        {
            Direction = Vector3.Normalize(Vector3.Transform(Vector3.UnitX, Rotation));

            if (acceleration != 0)
            {
                currentSpeed += acceleration;
            }
            else if(currentSpeed != 0)
            {
                float deacceleration = PLAYER_DEACCELERATION_FACTOR * gameTime.ElapsedSeconds();
                currentSpeed += currentSpeed > 0 ? Math.Max(-deacceleration, -currentSpeed) : Math.Min(deacceleration, -currentSpeed);
            }
            currentSpeed = MathHelper.Clamp(currentSpeed, -PLAYER_MAX_SPEED, PLAYER_MAX_SPEED);
            Position += Direction * currentSpeed * gameTime.ElapsedSeconds();
            acceleration = 0;
        }

        public Vector3 GetProjectedMovement()
        {
            return Direction * currentSpeed;
        }


        public void RotateLeft(GameTime gameTime)
        {
            Rotation *= Quaternion.CreateFromAxisAngle(Vector3.Up, PLAYER_TURNING_SPEED * gameTime.ElapsedSeconds());
        }
        
        public void Fire(GameTime gameTime, DrawInfo di, bool justTriggered)
        {
            if (CurrentWeapon != null)
            {
                if (Weapons[currentWeaponIndex].TryFire(this, justTriggered))
                {
                    GameStatistics.PlayerStats[playerNumber].BulletsFired++;
                    CollisionRay ray = CollisionRay.CreateFromModelInfo(this.Position, this.Rotation);
                    Vector2 endPolyNormal;
                    hits = gameWrapper.CollisionSimulator.ShootRay(ray, (Parent as LevelTileNode).Tile, out endPolyNormal);

                    ray.Position = new Vector2(this.WorldPosition.X, this.WorldPosition.Z);
                    muzzleFlash.LightPosition = this.WorldPosition + new Vector3(ray.Direction.X, 2, ray.Direction.Y) * 30f ;
                    di.UpdatePointLights.Add(muzzleFlash);
    
                    int index = hits.IndexOf(this);
                    if (index > -1)
                        gameWrapper.CollisionSimulator.shootRayResultDistances.RemoveAt(index);
                    
                    hits.Remove(this);
                    traceCount++;


                    if (hits.Count > 0)
                    {
                        if (hits[0] is Enemy)
                        {
                        }
                        else if (endPolyNormal != Vector2.Zero)
                        {
                            gameWrapper.fragments.DoFragments(ray.Position + ray.Direction * gameWrapper.CollisionSimulator.shootRayResultDistances[hits.Count - 1], -endPolyNormal, ray.Direction, 50, 15);
                        }
                        
                        int damage = CurrentWeapon.Damage;

                        for (int i = 0; i < hits.Count && i < CurrentWeapon.Penetration; i++)
                        {
                            if (hits[i] is ZombieCannon)
                            {
                                ((Enemy)hits[i]).TakeDamage(damage, this.playerNumber, true, ray.Direction);
                                gameWrapper.fragments.DoFragments(ray.Position + ray.Direction * gameWrapper.CollisionSimulator.shootRayResultDistances[i], -endPolyNormal, ray.Direction, 50, 15);
                            }
                            else if (hits[i] is Enemy)
                            {
                                ((Enemy)hits[i]).TakeDamage(damage, this.playerNumber, false, ray.Direction);
                                gameWrapper.blood.DoFragments(ray.Position + ray.Direction * gameWrapper.CollisionSimulator.shootRayResultDistances[i], ray.Direction, 10, 50);
                            }
                            else if (hits[i] is Player)
                            {
                                ((Player)hits[i]).TakeDamage(damage / 4, Vector2.Negate(ray.Direction));
                                gameWrapper.blood.DoFragments(ray.Position + ray.Direction * gameWrapper.CollisionSimulator.shootRayResultDistances[i], ray.Direction, 10, 50);
                            }

                            GameStatistics.PlayerStats[playerNumber].BulletsHit++;
                        }
                        
                        if (traceCount == 5)
                        {
                            gameWrapper.tracer.Trace(ray, 70, gameWrapper.CollisionSimulator.shootRayResultDistances[0]);
                            traceCount = 0;
                        }
                    }
                }
            }
        }

        


        Matrix parabolFix = Matrix.CreateRotationX(MathHelper.PiOver4 + MathHelper.PiOver4 / 2f) * Matrix.CreateTranslation(new Vector3(0, -5f, 5f));
        Quaternion baseRotation = Quaternion.CreateFromAxisAngle(Vector3.Up, MathHelper.PiOver2);
        public override void Update(GameTime gameTime, DrawInfo di)
        {
                        
            UpdateWeapons(gameTime);
            if (gameWrapper.goalNode != null)
            {
                Vector2 point = new Vector2(gameWrapper.goalNode.WorldPosition.X - WorldPosition.X, gameWrapper.goalNode.WorldPosition.Z - WorldPosition.Z);
                point.Normalize();
                float angle = (float)Math.Atan2(point.Y, point.X);

                Matrix rotation = Matrix.CreateFromQuaternion(Quaternion.Conjugate(Rotation) * Quaternion.CreateFromYawPitchRoll(-angle, 0, 0)); 
                playerModel.StaticManipulator["Parabol"] = parabolFix * rotation * playerModel.Model.Bones["Parabol"].Transform;
                playerModel.StaticManipulator["Parabol_Connector"] = rotation * playerModel.Model.Bones["Parabol_Connector"].Transform;
            }
            playerModel.Rotation = Quaternion.CreateFromYawPitchRoll(MathHelper.PiOver2, (currentSpeed / PLAYER_MAX_SPEED) * MathHelper.PiOver4 / 2f, 0);
            base.Update(gameTime, di);
        }
                
        public void Reload()
        {
            if (CurrentWeapon != null)
            {
                CurrentWeapon.Reload(this);    
            }            
        }

        private void UpdateWeapons(GameTime gameTime)
        {
            if (CurrentWeapon != null)
            {
                CurrentWeapon.Update(gameTime, this);    
            }
        }

        public void TakeDamage(int damage, Vector2 fromDirection)
        {
            GameStatistics.CommonStats.DamageFromZombies += damage;

            this.currentHealth -= damage;
            damageTaken += damage / 2;
        }

        
        public void ChangeWeapon()
        {
            if (CurrentWeapon != null)
            {
                CurrentWeapon.AbortReload();
                
                for (int i = 0; i < MAX_WEAPONS; i++)
                {
                    currentWeaponIndex = (currentWeaponIndex = currentWeaponIndex + 1) % MAX_WEAPONS;

                    if (Weapons[currentWeaponIndex] != null)
                    {
                        break;
                    }
                }                
            }
            UpdateVisibleWeapon();
        }

        public bool WantsWeapon(WeaponType type)
        {
            int numFreeSlots = 0;
            for (int i = 0; i < MAX_WEAPONS; i++)
            {
                if (Weapons[i] == null)
                    numFreeSlots++;
                else
                    if (Weapons[i].Type == type)
                        if (i == currentWeaponIndex)
                            if(WantsToPickup)
                                return true;
                            else 
                            {
                                ShowPickupMessage = true;
                                PickupType = type;
                                return false;
                            }
                        else
                            return false;
            }

            if (numFreeSlots > 0)
            {
                if (WeaponFactory.IsSmallWeapon(type))
                    if (WantsToPickup)
                        return true;
                    else
                    {
                        ShowPickupMessage = true;
                        PickupType = type;
                        return false;
                    }
                else
                    return true;
            }

            if (WantsToPickup)
                return true;
            else
            {
                ShowPickupMessage = true;
                PickupType = type;
                return false;
            }
        }

        private bool HasSmallWeapon()
        {
            for (int i = 0; i < MAX_WEAPONS; i++)
            {
                if (Weapons[i] != null && WeaponFactory.IsSmallWeapon(Weapons[i].Type))
                {
                    return true;
                }
            }

            return false;
        }

        public bool TryGiveWeapon(Weapon weapon)
        {
            if (WeaponFactory.IsSmallWeapon(weapon.Type) && HasSmallWeapon())
                return false;

            for (int i = 0; i < MAX_WEAPONS; i++)
            {
                if (Weapons[i] == null)
                {
                    Weapons[i] = weapon;
                    currentWeaponIndex = i;
                    UpdateVisibleWeapon();
                    return true;
                }
            }

            return false;
        }

        public Weapon ExchangeWithCurrentWeapon(Weapon weapon)
        {
            int otherIndex = 0;
            if (WeaponFactory.IsSmallWeapon(weapon.Type))
            {
                for (int i = 0; i < MAX_WEAPONS; i++)
                {
                    if (WeaponFactory.IsSmallWeapon(Weapons[i].Type))
                    {
                        otherIndex = i;
                        break;
                    }
                }
            }
            else
            {
                for (int i = 0; i < MAX_WEAPONS; i++)
                {
                    if (! WeaponFactory.IsSmallWeapon(Weapons[i].Type))
                    {
                        otherIndex = i;
                        break;
                    }
                }
            }


            Weapon oldWeapon = Weapons[otherIndex];
            Weapons[otherIndex] = weapon;

            if (otherIndex != currentWeaponIndex)
                ChangeWeapon();

            UpdateVisibleWeapon();

            WantsToPickup = false;

            return oldWeapon;
        }

        public void GiveAmmo(int count, AmmoType type)
        {
            if (type == AmmoType.Light)
            {
                ammoStats[0] += count;
            }
            else
            {
                ammoStats[1] += count;
            }
        }

        internal bool TryGiveHealth(int healthAmount)
        {
            if (currentHealth < STARTING_HEALTH)
            {
                currentHealth += healthAmount;
                if (currentHealth > STARTING_HEALTH)
                    currentHealth = STARTING_HEALTH;

                return true;
            }
            return false;
        }

        public void DrawPlayerHUD(GameTime gameTime, GraphicsDevice graphics, SpriteBatch spriteBatch, SpriteFont spriteFont)
        {
            spriteBatch.Begin(SpriteBlendMode.AlphaBlend, SpriteSortMode.Immediate, SaveStateMode.SaveState);
            DrawWeapons(graphics, spriteBatch);
            DrawHealth(graphics, spriteBatch);
            DrawPickupMessage(graphics, spriteBatch, spriteFont);
            DrawDamage(graphics, spriteBatch);
            DrawMinimap(graphics, spriteBatch);

            spriteBatch.End();
        }

        private void DrawDamage(GraphicsDevice graphics, SpriteBatch spriteBatch)
        {
            // Draw damage
            if (damageTaken > 0)
            {
                int width = graphics.Viewport.Width + Util.Random.Next(12) + 24;
                int height = graphics.Viewport.Height + Util.Random.Next(12) + 24;
                SpriteEffects effect = SpriteEffects.None;
                int n = Util.Random.Next(3);

                if (n == 1)
                {
                    effect = SpriteEffects.FlipHorizontally;
                }
                else if (n == 2)
                {
                    effect = SpriteEffects.FlipVertically;
                }

                float angle = (float)Util.Random.NextDouble() * 4;

                Rectangle damage = new Rectangle(-24, -24, width, height);

                spriteBatch.Draw(damageTexture, damage, null, Color.Red, 0, Vector2.Zero, effect, 0);

                damageTaken--;
            }
        }

        private void DrawPickupMessage(GraphicsDevice graphics, SpriteBatch spriteBatch, SpriteFont spriteFont)
        {

            // Draw pick up message
            if (ShowPickupMessage)
            {
                float pickUpStringScale = 1.6f;
                float weaponScale = 0.6f;
                Color pickUpStringColor = Color.White;
                Color pickUpWeaponColor = Color.White;
                Texture2D weaponTexture =  WeaponFactory.WhiteTextures[PickupType];
                StringBuilder pickUpString = new StringBuilder();
                pickUpString.Append("Press ");
                ControllerType t = gameWrapper.PlayerControllerTypes[playerNumber];
                if ((int)t < 2)
                    pickUpString.Append(controls[(int)t, 7].ToString());
                else
                    pickUpString.Append(Buttons.Y.ToString());
                pickUpString.Append(" to swap with ");

                float stringLength = spriteFont.MeasureString(pickUpString).X * pickUpStringScale;
                float totalLength = stringLength + weaponTexture.Width * weaponScale;
                float xPickUp = graphics.Viewport.Width / 2 - totalLength / 2;
                float yPickUp = graphics.Viewport.Height / 2 - (weaponTexture.Height * weaponScale) / 2;

                spriteBatch.DrawString(spriteFont, pickUpString, new Vector2(xPickUp, yPickUp), pickUpStringColor, 0, Vector2.Zero, pickUpStringScale, SpriteEffects.None, 0);
                spriteBatch.Draw(weaponTexture, new Rectangle((int)(xPickUp + stringLength), (int)(yPickUp - (weaponTexture.Height * weaponScale) / 2), (int)(weaponTexture.Width * weaponScale),
                    (int)(weaponTexture.Height * weaponScale)), pickUpWeaponColor);

                ShowPickupMessage = false;
            }
        }

        private void DrawHealth(GraphicsDevice graphics, SpriteBatch spriteBatch)
        {

            // Draw health
            spriteBatch.Draw(healthHeart, new Vector2(graphics.Viewport.Width - HUD_RIGHT_SPACING - HUD_HEALTHBAR_WIDTH - 5 - HUD_HEALTHHEART_SIZE,
                                                      graphics.Viewport.Height - HUD_BOTTOM_SPACING - HUD_HEALTHHEART_SIZE), Color.White);

            float healthWidth = (float)HUD_HEALTHBAR_WIDTH / (float)STARTING_HEALTH;

            spriteBatch.Draw(texture, new Rectangle(graphics.Viewport.Width - HUD_RIGHT_SPACING - HUD_HEALTHBAR_WIDTH,
                                                    graphics.Viewport.Height - HUD_BOTTOM_SPACING - HUD_HEALTHBAR_HEIGHT,
                                                    (int)(healthWidth * (float)currentHealth),
                                                    HUD_HEALTHBAR_HEIGHT),
                                                    HUD_HEALTH_COLOR);

            spriteBatch.Draw(healthBar, new Vector2(graphics.Viewport.Width - HUD_RIGHT_SPACING - HUD_HEALTHBAR_WIDTH,
                                                      graphics.Viewport.Height - HUD_BOTTOM_SPACING - HUD_HEALTHBAR_HEIGHT), HUD_HEALTH_OUTLINE_COLOR);
        }

        private void DrawWeapons(GraphicsDevice graphics, SpriteBatch spriteBatch)
        {

            if (CurrentWeapon != null)
            {
                float ammoWidth = (float)WeaponFactory.WhiteTextures[CurrentWeapon.Type].Width / (float)CurrentWeapon.ClipSize;

                spriteBatch.Draw(WeaponFactory.BlackTextures[CurrentWeapon.Type], new Vector2(HUD_LEFT_SPACING, graphics.Viewport.Height - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - HUD_BOTTOM_SPACING), Color.White);

                if (!CurrentWeapon.IsReloading)
                {
                    spriteBatch.Draw(WeaponFactory.WhiteTextures[CurrentWeapon.Type], new Vector2(HUD_LEFT_SPACING, graphics.Viewport.Height - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - HUD_BOTTOM_SPACING),
                        new Rectangle(0, 0, (int)(ammoWidth * (float)CurrentWeapon.AmmoInClip), WeaponFactory.WhiteTextures[CurrentWeapon.Type].Height), HUD_AMMO_COLOR);
                }
                else
                {
                    spriteBatch.Draw(WeaponFactory.WhiteTextures[CurrentWeapon.Type], new Vector2(HUD_LEFT_SPACING, graphics.Viewport.Height - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - HUD_BOTTOM_SPACING),
                        new Rectangle(0, 0, (int)((float)WeaponFactory.WhiteTextures[CurrentWeapon.Type].Width * CurrentWeapon.GetReloadProgress()), WeaponFactory.WhiteTextures[CurrentWeapon.Type].Height), HUD_AMMO_RELOAD_COLOR);
                }

                // Draw ammo
                if (CurrentWeapon != null)
                {

                    bool lightPrimary = CurrentWeapon.AmmoType == AmmoType.Light;

                    Color lightAmmoColor = lightPrimary ? Color.White : HUD_AMMO_ALMOST_OUTLINE;
                    Color heavyAmmoColor = !lightPrimary ? Color.White : HUD_AMMO_ALMOST_OUTLINE;

                    float heavyX = HUD_LEFT_SPACING;
                    float lightX = HUD_LEFT_SPACING;

                    float lightY = lightPrimary ? graphics.Viewport.Height - HUD_BOTTOM_SPACING - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - HUD_WEAPON_AMMO_SPACING - HUD_LIGHT_AMMO_SIZE
                                                : graphics.Viewport.Height - HUD_BOTTOM_SPACING - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - 2 * HUD_WEAPON_AMMO_SPACING - HUD_HEAVY_AMMO_SIZE - HUD_LIGHT_AMMO_SIZE;


                    float heavyY = !lightPrimary ? graphics.Viewport.Height - HUD_BOTTOM_SPACING - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - HUD_WEAPON_AMMO_SPACING - HUD_HEAVY_AMMO_SIZE
                                                 : graphics.Viewport.Height - HUD_BOTTOM_SPACING - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - 2 * HUD_WEAPON_AMMO_SPACING - HUD_HEAVY_AMMO_SIZE - HUD_LIGHT_AMMO_SIZE;

                    // Light ammo
                    for (int i = 0; i < MAX_LIGHT_AMMO; i++)
                    {
                        spriteBatch.Draw(lightAmmo, new Vector2(lightX, lightY), (i < ammoStats[0]) ? lightAmmoColor : lightPrimary ? HUD_AMMO_OUTLINE : Color.TransparentWhite);
                        lightX += HUD_LIGHT_AMMO_SIZE + 5;
                    }

                    // Heavy ammo
                    for (int i = 0; i < MAX_HEAVY_AMMO; i++)
                    {
                        spriteBatch.Draw(heavyAmmo, new Vector2(heavyX, heavyY), (i < ammoStats[1]) ? heavyAmmoColor : !lightPrimary ? HUD_AMMO_OUTLINE : Color.TransparentWhite);
                        heavyX += HUD_HEAVY_AMMO_SIZE + 5;
                    }
                }


                float ammoHeight = 3 * HUD_WEAPON_AMMO_SPACING + HUD_LIGHT_AMMO_SIZE + HUD_HEAVY_AMMO_SIZE;

                Weapon secondaryWeapon = TryGetSecondaryWeapon();

                if (secondaryWeapon != null)
                {
                    Texture2D black = WeaponFactory.BlackTextures[secondaryWeapon.Type];
                    Texture2D white = WeaponFactory.WhiteTextures[secondaryWeapon.Type];

                    float secondaryAmmoWidth = (float)white.Width / (float)secondaryWeapon.ClipSize;
                    int yPos = graphics.Viewport.Height - WeaponFactory.BlackTextures[CurrentWeapon.Type].Height - (int)ammoHeight - (int)(black.Height * HUD_SECONDARY_WEAPON_SCALE) - HUD_BOTTOM_SPACING;

                    spriteBatch.Draw(white,
                        new Rectangle(HUD_LEFT_SPACING, yPos, (int)(white.Width * HUD_SECONDARY_WEAPON_SCALE),
                            (int)(black.Height * HUD_SECONDARY_WEAPON_SCALE)), HUD_SEC_AMMO_EMPTY_COLOR);

                    spriteBatch.Draw(white, new Vector2(HUD_LEFT_SPACING, yPos),
                        new Rectangle(0, 0, (int)(secondaryAmmoWidth * (float)secondaryWeapon.AmmoInClip),
                            black.Height), HUD_SEC_AMMO_COLOR, 0, Vector2.Zero, HUD_SECONDARY_WEAPON_SCALE, SpriteEffects.None, 0);
                }
            }
        }

        private Weapon TryGetSecondaryWeapon()
        {
            // Assumption: MAX_WEAPONS = 2 (Otherwise it will blow up)

            return Weapons[currentWeaponIndex == 0 ? 1 : 0] == null ? null : Weapons[currentWeaponIndex == 0 ? 1 : 0];
        }

        private void DrawMinimap(GraphicsDevice graphics, SpriteBatch spriteBatch)
        {
            if (this.minimapUnclean)
                PaintMinimap(graphics);

            float scale = 3f;

            int move = (int)(scale * (float)minimapDiagonalSize);
            Vector2 position = new Vector2(graphics.Viewport.Width - (HUD_RIGHT_SPACING + move), HUD_TOP_SPACING + move);

            Vector2 origin = new Vector2(miniMap.Width / 2, miniMap.Height / 2);

            float rotation = (float)-Math.Atan2(Direction.Z, Direction.X) - MathHelper.PiOver2;

            spriteBatch.Draw(miniMap, position, null, Color.White, rotation, origin, scale, SpriteEffects.None, 0);
        }

        protected override void MoveToOtherNode(LevelTileNode node)
        {
            this.currentTile = node.Tile;
            if (!visitedTiles.Contains(node.Tile))
                visitedTiles.Add(node.Tile);

            this.minimapUnclean = true;

            base.MoveToOtherNode(node);
        }

        private void PaintMinimap(GraphicsDevice graphics)
        {
            if (this.visitedTiles.Count == 0)
                return;

            int minX = this.visitedTiles[0].X, minY = this.visitedTiles[0].Y;
            int maxX = this.visitedTiles[0].X, maxY = this.visitedTiles[0].Y;

            foreach (LevelTile aTile in this.visitedTiles)
            {
                minX = (minX > aTile.X) ? aTile.X : minX;
                maxX = (maxX < aTile.X) ? aTile.X : maxX;

                minY = (minY > aTile.Y) ? aTile.Y : minY;
                maxY = (maxY < aTile.Y) ? aTile.Y : maxY;
            }

            int height = maxY - minY + 1;
            int width = maxX - minX + 1;
            int scale = endTile.Height;

            int textureWidth = 2, textureHeight = 2;
            while (textureWidth < width)
                textureWidth *= 2;
            while (textureHeight < height)
                textureHeight *= 2;
            Texture2D miniMap = new Texture2D(graphics, textureWidth * scale, textureHeight * scale, 0, TextureUsage.None, SurfaceFormat.Color);
            Color[] minimapData = new Color[textureWidth * scale * textureHeight * scale];
            for (int i = 0; i < textureWidth * scale * textureHeight * scale; i++)
                minimapData[i] = Color.White;

            int halfScale = scale / 2;
            int quarterScale = halfScale / 2;
            Color[] currentTile = new Color[halfScale * halfScale];
            for (int i = 0; i < halfScale * halfScale; i++)
                currentTile[i] = Color.Blue;

            Color[] currentData = new Color[scale * scale];
            Color[,] rotatedData = new Color[scale, scale];
            int offsetTop, offsetLeft;
            int rotationamount;

            foreach (LevelTile aTile in this.visitedTiles)
            {
                currentData = (Color[])tileData[aTile.TileType].Clone();
                if (aTile.Rotation == GameEngine.LevelGeneration.Rotation.Up)
                {
                    // Nothing, already rotated correctly... Oh really?
                }
                else
                {
                    for (int i = 0; i < scale; i++)
                        for (int j = 0; j < scale; j++)
                            rotatedData[i, j] = currentData[i * scale + j];

                    if (aTile.Rotation == GameEngine.LevelGeneration.Rotation.Right)
                        rotationamount = 1;
                    else if (aTile.Rotation == GameEngine.LevelGeneration.Rotation.Down)
                        rotationamount = 2;
                    else
                        rotationamount = 3;

                    for (int k = 0; k < rotationamount; k++)
                        rotatedData = RotateMatrix(rotatedData, scale);

                    for (int i = 0; i < scale; i++)
                        for (int j = 0; j < scale; j++)
                            currentData[i * scale + j] = rotatedData[i, j];
                }

                offsetLeft = (width - 1 - (maxX - aTile.X)) * scale;
                offsetTop = (aTile.Y - minY) * scale;
                miniMap.SetData(0, new Rectangle(offsetLeft, offsetTop, scale, scale), currentData, 0, currentData.Length, SetDataOptions.None);

                if (aTile == this.currentTile)
                    miniMap.SetData(0, new Rectangle(offsetLeft + quarterScale, offsetTop + quarterScale, halfScale, halfScale), currentTile, 0, currentTile.Length, SetDataOptions.None);
            }

            this.minimapDiagonalSize = (int)Math.Sqrt(((double)miniMap.Width * (double)miniMap.Width) + ((double)miniMap.Height * (double)miniMap.Height));

//            if (this.miniMap != null)
  //              this.miniMap.Dispose();
            this.miniMap = miniMap;
            this.minimapUnclean = false;
        }

        private Color[,] RotateMatrix(Color[,] matrix, int n)
        {
            Color[,] ret = new Color[n, n];

            for (int i = 0; i < n; ++i)
                for (int j = 0; j < n; ++j)
                    ret[i, j] = matrix[n - j - 1, i];

            return ret;
        }
    }
}
