﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using GameEngine.Xml;
using GameEngine.Helpers;
using Microsoft.Xna.Framework;
using GameEngine;
using GameEngine.ScreenSystem;

namespace GamePrototype.GameLogic
{
    public class ZombieCannon : Enemy
    {
        public float FiringInterval = 3f;
        public float FiringVelocity = 1000f;
        public float TurningSpeed = MathHelper.Pi;
        
        private float coolDown = 0f;
        private GameWrapper game;
        private ModelNode Model;
        

        public ZombieCannon(string id, int hitpoint, float moveSpeed, GameWrapper game)
            : base(id, hitpoint, moveSpeed)
        {
            this.game = game;
            this.moveSpeed = 0;
            this.radius = 60;
            Model = new ModelNode("ZombieModel");
            Model.Asset = @"Models\Cannon\Cannon";
            Model.AnimationFactor = 2f;
            Model.Scale = new Vector3(0.15f);
            SetNextWaypoint(Position + Vector3.UnitX);
            this.currentDirection = targetDirection;
            
            this.ChildNodes.Add(Model);
        }

        float deathAnim = 2f;

        public override void Update(Microsoft.Xna.Framework.GameTime gameTime, GameEngine.DrawInfo di)
        {
            base.Update(gameTime, di);
            if (coolDown > 0)
                coolDown = Math.Max(0, coolDown - gameTime.ElapsedSeconds());

            if (!Dead && hitpoints <= 0)
            {
                Dead = true;
                PassThrough = true;
                AudioManager.PlaySoundEffect("Audio/teleport_activate", 1f, ((float)Util.Random.NextDouble() - 0.5f) * 0.1f);

            }

            if (Dead && deathAnim > 0)
            {
                deathAnim -= gameTime.ElapsedSeconds();

                Model.Scale = new Vector3(MathHelper.Clamp(deathAnim * 2f, 0.01f, 1f), Math.Max(deathAnim - .8f, 0.25f), MathHelper.Clamp(deathAnim * 2f, 0.01f, 1f)) * 0.15f;
                Model.Rotation = Quaternion.CreateFromAxisAngle(Vector3.Up, MathHelper.TwoPi * (2f - deathAnim) * (2f - deathAnim) * 2);
            }
            else if (deathAnim <= 0 && Model.IsVisible)
            {
                Model.IsVisible = false;
                if (Util.Random.Next(100) >= 50)
                {
                    Weapon item;
                    int rand = Util.Random.Next(10);
                    if (rand >= 8)
                        item = WeaponFactory.GetWeapon(WeaponType.Minigun);
                    else if (rand >= 4)
                        item = WeaponFactory.GetWeapon(WeaponType.Garand);
                    else
                        item = WeaponFactory.GetWeapon(WeaponType.M4);

                    game.ThrowItemOnFloor(SpawnPointNode.SpawnType.BigWeapon, item, this);
                } else
                    game.ThrowItemOnFloor(SpawnPointNode.SpawnType.BigAmmo, "BigAmmo", this);

            }
            

            if (!Dead)
            {
                TryFiring();

                if (target != null)
                {
                    float distance = Vector3.DistanceSquared(target.Position, Position);
                    float dt = distance / (FiringVelocity * FiringVelocity);

                    SetNextWaypoint(target.GetProjectedMovement() * dt + target.Position);
                }

                this.Position.Y = 20f + (float)Math.Sin(gameTime.TotalGameTime.TotalSeconds * 2.0) * 10f;
            }
        }

        private Player target = null;
        List<Player> potentialTargets = new List<Player>();


        protected void AquireTarget()
        {
            LevelTileNode tile = this.Parent as LevelTileNode;

            if (target != null && !tile.ContainsDynamic(target))
            {
                target = null;
            }

            if (target == null)
            {
                foreach (var player in game.Players)
                {
                    if(!tile.ContainsDynamic(player))
                        continue;

                    if (target == null || (target.Position - Position).LengthSquared() > (player.Position - Position).LengthSquared())
                    {
                        target = player;
                        coolDown = FiringInterval;
                    }
                }
            }

        }

        protected bool CanSeePlayer(Player p)
        {
            LevelTileNode tile = this.Parent as LevelTileNode;

            if (!tile.ContainsDynamic(p))
                return false;

            Vector2 pos = Position.ToVector2XZ();
            CollisionRay ray = new CollisionRay(pos, p.Position.ToVector2XZ() - pos);
            
            Vector2 endPolyNormal;
            var hits = game.CollisionSimulator.ShootRay(ray, tile.Tile, out endPolyNormal);
            return hits.Contains(p);
        }

        protected void TryFiring()
        {
            if (coolDown > 0)
                return;

            AquireTarget();

            if (coolDown > 0)
                return;

            if (target == null)
                return;

            if (!CanSeePlayer(target))
                return;

            Fire();
 
        }

        protected void Fire()
        {
            LevelTileNode tile = this.Parent as LevelTileNode;
            AudioManager.PlaySoundEffect("Audio/floomp", 2f, (float)Util.Random.NextDouble() - 0.5f);
            ZombieBear zb = game.enemyController.GetNextUnusedEnemy() as ZombieBear;
            if (zb == null)
                return;
            zb.Position = Position + moveDirection * (radius + zb.radius);
            zb.Position.Y = 0;
            zb.LastPosition = zb.Position;
            zb.IsInUse = true;
            tile.AddDynamicNode(zb);
            tile.SpawnCount--;
            zb.hitpoints = game.enemyController.HITPOINTS[game.enemyController.difficulty - 1];
            zb.CurrentTarget = target;
            zb.SetNextWaypoint(zb.CurrentTarget.Position);
            zb.TicCounter = Util.Random.Next(20);
            zb.IsProjectile = true;
            zb.SetSuper(0.4f);
            zb.DisableTic = true;
            zb.FlyingSpeed = FiringVelocity / 2f;
            zb.TicCounter = 0;


            if (moveDirection != Vector3.Zero)
            {
                zb.moveDirection = Vector3.Normalize(moveDirection);
            }
            
            coolDown = FiringInterval;
            // fire a zombie at the player.
        }

    }
}
